/* retawq/cursesx.c - a small xcurses implementation
   This file is part of retawq (<http://retawq.sourceforge.net/>), a network
   client created by Arne Thomassen; retawq is basically released under certain
   versions of the GNU General Public License and WITHOUT ANY WARRANTY.
   Read the file COPYING for license details, README for program information.
   Copyright (C) 2004 Arne Thomassen <arne@arne-thomassen.de>
*/

/* For the rest of the program, xcurses mode shall look like any normal curses
   mode as far as possible. */

#include "stuff.h"
#include "cursesx.h"
#include "resource.h"

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

__define_filestuff

static Display* xws_display;
static const_after_init Window xws_window;
static const_after_init GC xws_gc;
static const_after_init Colormap colormap;
static const_after_init unsigned long pixel_fg, pixel_bg, colors[COLOR_PAIRS];
static const_after_init XColor colors_storage[COLOR_PAIRS];
static const_after_init Atom WmProt, WmDelete;
static const_after_init Cursor cursor_arrow, cursor_textmark;
static const_after_init tBoolean use_cursor = falsE;

typedef struct
{ const XFontStruct* font;
  int width, height, ascent;
} tXwsFont;

static const tXwsFont *font_normal, *font_bold;

typedef struct tXwsConnection
{ struct tXwsConnection *prev, *next;
  int fd;
} tXwsConnection;

static tXwsConnection *conn = NULL, *current_conn = NULL;
static tBoolean must_remove_current_conn = falsE;

static attr_t attr = 0;

WINDOW* stdscr;
int COLS = 80, LINES = 25;

#define SAI(x, y) ( ((y) * COLS) + (x) ) /* screen array index */
#define A_CHARTEXT (255)

static mmask_t __mousemask = 0;

typedef struct tGetmouse
{ MEVENT data;
  struct tGetmouse* next;
} tGetmouse;

static tGetmouse* getmouse_list = NULL;

static struct
{ const char* text;
  size_t size;
  int x1, y1, x2, y2;
  int gx, gy; /* coordinates of button-down event */
  unsigned char state; /* "&1": user is currently selecting */
} sel;


/* Helper functions */

static unsigned char __must_redraw = 1 | 2 | 4;
static my_inline void must_redraw(unsigned char what)
/* <what>: "&1": cursor; "&2": contents; "&4": resized */
{ __must_redraw |= what;
}

static void connection_remove(const tXwsConnection* xc)
{ if (xc->prev != NULL) xc->prev->next = xc->next;
  if (xc->next != NULL) xc->next->prev = xc->prev;
  if (conn == xc) conn = xc->next;
  memory_deallocate(xc);
}

static void connection_do_watch(int fd)
{ tXwsConnection* xc;
  if (!my_try_FD_SET(fd, &fds_read)) fatal_tmofd(fd);
  xc = memory_allocate(sizeof(tXwsConnection));
  if (conn != NULL) conn->prev = xc;
  xc->fd = fd; xc->next = conn; conn = xc;
}

static void connection_watch(Display* display,
  __sunused XPointer client_data __cunused, int fd, Bool opening,
  __sunused XPointer* watch_data __cunused)
{ tXwsConnection* xc;
  if (display != xws_display) return; /* "can't happen"? */
  if (fd < 0) return; /* "can't happen"? */
  if (opening) connection_do_watch(fd);
  else /* remove the respective element from the list */
  { FD_CLR(fd, &fds_read); FD_CLR(fd, &fds_r); FD_CLR(fd, &fds_e);
    xc = conn;
    while (xc != NULL)
    { if (xc->fd == fd) /* found */
      { if (xc == current_conn) must_remove_current_conn = truE; /* defer */
        else connection_remove(xc);
        break;
      }
      xc = xc->next;
    }
  }
}

void xcurses_check_connections(int* _selectcount)
{ int selectcount = *_selectcount;
  current_conn = conn;
  while (current_conn != NULL)
  { tXwsConnection* next;
    const int fd = current_conn->fd;
    const tBoolean is_set =
      cond2boolean( (FD_ISSET(fd, &fds_r)) || (FD_ISSET(fd, &fds_e)) );
    if (is_set)
    { XProcessInternalConnection(xws_display, fd);
      if (--selectcount <= 0) break;
    }
    next = current_conn->next; /* (must be done exactly _here_!) */
    if (must_remove_current_conn)
    { connection_remove(current_conn); must_remove_current_conn = falsE; }
    current_conn = next;
  }
  *_selectcount = selectcount;
}

static tXwsFont* load_font(const char* which)
{ tXwsFont* retval;
  XFontStruct* font;
  char buf[1024], buf2[1024];
  int ascent;
  sprint_safe(buf, "-misc-fixed-%s-r-normal--13-120-*-*-c-70-iso8859-1",which);
  font = XLoadQueryFont(xws_display, buf);
  if (font == NULL)
  { sprint_safe(buf2, _("can't load font %s"), buf); fatal_error(0, buf2); }
  retval = memory_allocate(sizeof(tXwsFont));
  retval->font = font; retval->ascent = ascent = font->ascent;
  retval->height = ascent + font->descent;
  retval->width = XTextWidth(font, "M", 1);
  return(retval);
}

static void use_font(const tXwsFont* font)
{ (void) XSetFont(xws_display, xws_gc, font->font->fid);
}

static void set_fgbg(unsigned long fg, unsigned long bg)
{ XSetForeground(xws_display, xws_gc, fg);
  XSetBackground(xws_display, xws_gc, bg);
}

static void set_cursor(Cursor c)
{ if (use_cursor) (void) XDefineCursor(xws_display, xws_window, c);
}

static void coord_limit(int* _x, int* _y)
{ int x = *_x, y = *_y;
  if (x < 0) x = 0;
  else if (x > COLS - 1) x = COLS - 1;
  if (y < 0) y = 0;
  else if (y > LINES - 1) y = LINES - 1;
  *_x = x; *_y = y;
}

static tBoolean coord_pixel2char(int gx, int gy, /*@out@*/ int* _x,
  /*@out@*/ int* _y, tBoolean fail_if_bad)
{ int x = gx / font_normal->width, y = gy / font_normal->height;
  if (fail_if_bad)
  { if ( (x < 0) || (x >= COLS) || (y < 0) || (y >= LINES) ) return(falsE); }
  coord_limit(&x, &y);
  *_x = x; *_y = y;
  return(truE);
}

static void coord_reorder(int* x1, int* y1, int* x2, int* y2)
{ int a;
  if (*y1 > *y2)
  { a = *x1; *x1 = *x2; *x2 = a;
    a = *y1; *y1 = *y2; *y2 = a;
  }
  else if ( (*y1 == *y2) && (*x1 > *x2) ) { a = *x1; *x1 = *x2; *x2 = a; }
}

static void reinit_stdscr_data(void)
/* Call this whenever COLS or LINES changed. */
{ chtype* text;
  int prod = COLS * LINES, count;
  stdscr->x = stdscr->y = 0; __dealloc(stdscr->text);
  text = stdscr->text = __memory_allocate(prod * sizeof(chtype));
  for (count = 0; count < prod; count++) text[count] = ' ';
  __dealloc(stdscr->attr);
  stdscr->attr = memory_allocate(prod * sizeof(attr_t));
  must_redraw(1 | 2 | 4); sel.state &= ~1;
}


/* Initialization */

WINDOW* __init initscr(void)
{ int fd, screen, width, height, count;
  XSizeHints* size_hints = XAllocSizeHints();
  XWMHints* wm_hints = XAllocWMHints();
  XClassHint* class_hint = XAllocClassHint();
  unsigned long pixel_black, pixel_white;

  xws_display = XOpenDisplay(NULL);
  if (xws_display == NULL) { cant_open: fatal_error(0, _(strCantOpenXws)); }
  fd = ConnectionNumber(xws_display);
  if (fd < 0) goto cant_open; /* "can't happen"? */
  connection_do_watch(fd);
  if (!XAddConnectionWatch(xws_display, connection_watch, NULL))
    fatal_error(0, _("can't watch X display connections"));
  screen = DefaultScreen(xws_display);
  colormap = XDefaultColormap(xws_display, screen);
  pixel_black = BlackPixel(xws_display, screen);
  pixel_white = WhitePixel(xws_display, screen);
  if (config.flags & (cfColorsOff | cfColorsReverse))
  { pixel_fg = pixel_black; pixel_bg = pixel_white; }
  else { pixel_fg = pixel_white; pixel_bg = pixel_black; }

  font_normal = load_font("medium"); font_bold = load_font("bold");

  width = 80 * font_normal->width; height = 25 * font_normal->height;
  xws_window = XCreateSimpleWindow(xws_display, RootWindow(xws_display,
    screen), 0, 0, width, height, 2, pixel_fg, pixel_bg);
  if (xws_window == None)
  { fail_window: fatal_error(0, _("can't create window")); }

  size_hints->flags = PSize;
  size_hints->width = width; size_hints->height = height;
  wm_hints->flags = InputHint | StateHint | WindowGroupHint;
  wm_hints->input = True; wm_hints->initial_state = NormalState;
  wm_hints->window_group = xws_window;
  class_hint->res_name = unconstify(strRetawq);
  class_hint->res_class = unconstify("Xedit");
  XmbSetWMProperties(xws_display, xws_window, strVersion, strVersion, NULL, 0,
    size_hints, wm_hints, class_hint);
  WmProt = XInternAtom(xws_display, "WM_PROTOCOLS", False);
  WmDelete = XInternAtom(xws_display, "WM_DELETE_WINDOW", False);
  if (WmDelete != None) XSetWMProtocols(xws_display, xws_window, &WmDelete, 1);
  (void) XSelectInput(xws_display, xws_window, KeyPressMask | ButtonPressMask |
    ButtonReleaseMask | Button1MotionMask | ExposureMask |
    VisibilityChangeMask | StructureNotifyMask); /* CHECKME! */

  xws_gc = XCreateGC(xws_display, xws_window, 0, NULL);
  if (xws_gc == None) goto fail_window;
  for (count = 0; count < COLOR_PAIRS; count++) colors[count] = pixel_fg;
  set_fgbg(pixel_fg, pixel_bg); use_font(font_normal);

  (void) XMapWindow(xws_display, xws_window);
  cursor_arrow = XCreateFontCursor(xws_display, XC_left_ptr);
  cursor_textmark = XCreateFontCursor(xws_display, XC_xterm);
  if ( (cursor_arrow != None) && (cursor_textmark != None) ) use_cursor = truE;
  stdscr = (WINDOW*) memory_allocate(sizeof(WINDOW)); reinit_stdscr_data();
  set_cursor(cursor_textmark); (void) XFlush(xws_display); my_memclr_var(sel);
  return(stdscr);
}


/* Event handling */

static void redraw_char(int x, int y, attr_t a)
{ const tXwsFont* font = ( (a & A_BOLD) ? font_bold : font_normal );
  const int width = font->width, gx = x * width,
    gy = y * font->height + font->ascent;
  const chtype c = stdscr->text[SAI(x, y)];
  const tColorPairNumber cpn = ( (a & __A_COLORMARK) ?
    ((a & __A_COLORPAIRMASK) >> __A_COLORPAIRSHIFT) : cpnDefault );
  unsigned long fg = colors[cpn], bg = pixel_bg, xg;
  if (a & A_REVERSE) { xg = fg; fg = bg; bg = xg; }
  use_font(font); set_fgbg(fg, bg);
  (void) XDrawImageString(xws_display, xws_window, xws_gc, gx, gy,
    (char*) (&c), 1);
  if (a & A_UNDERLINE)
  { (void) XDrawLine(xws_display, xws_window, xws_gc, gx, gy + 1,
      gx + width - 1, gy + 1);
  }
}

static void redraw_stdscr(void)
{ const attr_t* a = stdscr->attr;
  int x, y;
  for (y = 0; y < LINES; y++)
  { for (x = 0; x < COLS; x++) redraw_char(x, y, a[SAI(x, y)]); }
  redraw_char(stdscr->x, stdscr->y, a[SAI(stdscr->x, stdscr->y)] ^ A_REVERSE);
  /* use_font(font_normal); */ set_fgbg(pixel_fg, pixel_bg); __must_redraw = 0;
}

static void write_key(tKey key)
{ (void) my_write(fd_xcurses2main_write, &key, sizeof(key));
}

static one_caller void selection_highlight(int x2, int y2)
{ int x1 = sel.x1, y1 = sel.y1, x = x1, y;
  const attr_t* a = stdscr->attr;
  redraw_stdscr(); /* FIXME: the flickering is too ugly! */
  coord_reorder(&x1, &y1, &x2, &y2);
  for (y = y1; y <= y2; y++)
  { int xend = ( (y == y2) ? x2 : COLS - 1 );
    while (x <= xend) { redraw_char(x, y, a[SAI(x, y)] ^ A_REVERSE); x++; }
    x = 0;
  }
}

static __my_inline void selection_unhighlight(void)
{ redraw_stdscr(); /* KISS */
}

static one_caller void selection_copy(/*const*/ XButtonEvent* be)
{ int x1 = sel.x1, y1 = sel.y1, x2 = sel.x2, y2 = sel.y2, x = x1, y;
  size_t size;
  char *text, *ptr;
  tBoolean need_newline = falsE;
  coord_reorder(&x1, &y1, &x2, &y2);
  size = (y2 - y1 + 1) * (COLS + 1) + 1; /* better too much than... */
  __dealloc(sel.text); sel.text = text = ptr = __memory_allocate(size);
  for (y = y1; y <= y2; y++)
  { int xend = ( (y == y2) ? x2 : COLS - 1 ), spacecount = 0;
    if (need_newline) { need_newline = falsE; *ptr++ = '\n'; }
    while (x <= xend)
    { const char ch = stdscr->text[SAI(x, y)];
      if (ch == ' ') spacecount++; /* (to ignore trailing space characters) */
      else
      { while (spacecount > 0) { *ptr++ = ' '; spacecount--; }
        *ptr++ = ch;
      }
      x++;
    }
    if (x >= COLS - 1) need_newline = truE;
    x = 0;
  }
  *ptr = '\0'; debugmsg("selection_copy(): *"); debugmsg(text);debugmsg("*\n");
  sel.size = ptr - text; /* exact, in contrast to <size> */
  (void) XSetSelectionOwner(xws_display, XA_PRIMARY, xws_window, be->time);
  if (XGetSelectionOwner(xws_display, XA_PRIMARY) != xws_window)
    selection_unhighlight();
}

static one_caller void selection_provide(const XSelectionRequestEvent* sre)
{ XSelectionEvent e;
  Window requestor;
  if (sel.text == NULL) return; /* "should not happen" */
  requestor = sre->requestor;
  my_memclr_var(e); e.type = SelectionNotify; e.display = sre->display;
  e.requestor = requestor; e.selection = sre->selection;
  e.target = sre->target; e.property = sre->property; e.time = sre->time;
  (void) XChangeProperty(xws_display, requestor, e.property, XA_STRING, 8,
    PropModeReplace, sel.text, sel.size);
  (void) XSendEvent(xws_display, requestor, False, 0, (XEvent*) (&e));
}

static one_caller void selection_paste(Time t)
{ static tBoolean did_init = falsE;
  static Atom dummy;
  if (XGetSelectionOwner(xws_display, XA_PRIMARY) == None) return;
  if (!did_init)
  { dummy = XInternAtom(xws_display, "RETAWQ_SELECTION", False);
    did_init = truE;
  }
  (void) XConvertSelection(xws_display, XA_PRIMARY, XA_STRING, dummy,
    xws_window, t);
}

static one_caller void selection_do_paste(const XSelectionEvent* se)
{ Atom prop = se->property, type;
  int format;
  unsigned long nitems, remaining, count;
  unsigned char* value;
  if (prop == None) return;
  if (XGetWindowProperty(xws_display, xws_window, prop, 0, 1024, True,
      AnyPropertyType, &type, &format, &nitems, &remaining, &value) != 0)
    return;
  if (format != 8) { debugmsg("do_paste(): bad format\n"); return; }
  for (count = 0; count < nitems; count++) /* IMPROVEME?! */
    write_key(value[count]);
  XFree(value);
  /* IMPLEMENTME: read the _whole_ value? paste only if main.c is in an
     "appropriate" mode, e.g. khmLineInput or editor? */
}

static one_caller void handle_event(/*const*/ XEvent* event)
{ int width, height;
  /*const*/ XKeyEvent* ke;
  /*const*/ XButtonEvent* be;
  unsigned int button;
  KeySym sym;
  char ch;
  switch (event->type)
  { case KeyPress:
      ke = (/*const*/ XKeyEvent*) event;
      (void) XLookupString(ke, &ch, 1, &sym, NULL);
      if (IsModifierKey(sym)) { /* nothing */ }
      else if (sym != NoSymbol)
      { if ( (ke->state & ControlMask) && (my_islower(sym)) )
          sym = (sym - 'a') + 1;
        write_key(sym);
      }
      break;
    case Expose:
      if ( ((XExposeEvent*) event)->count <= 0 ) redraw_stdscr();
      break;
    case ButtonPress:
      be = (/*const*/ XButtonEvent*) event;
      button = be->button;
      if (__mousemask != 0)
      { mmask_t m = 0;
        if (button == Button1) m |= BUTTON1_CLICKED;
        else if (button == Button2) m |= BUTTON2_CLICKED;
        else if (button == Button3) m |= BUTTON3_CLICKED;
        m &= __mousemask;
        if (m != 0)
        { tGetmouse *list, *entry;
          MEVENT* data;
          int x, y;
          if (!coord_pixel2char(be->x, be->y, &x, &y, truE)) return;
          list = getmouse_list; entry = memory_allocate(sizeof(tGetmouse));
          data = &(entry->data); data->x = x; data->y = y; data->bstate = m;
          if (list == NULL) getmouse_list = entry;
          else /* rare */
          { while (list->next != NULL) list = list->next;
            list->next = entry;
          }
          write_key(KEY_MOUSE);
        }
      }
      else if (button == Button1) /* user starts text selection */
      { int gx = be->x, gy = be->y, x, y;
        if (!coord_pixel2char(gx, gy, &x, &y, falsE)) return;
        sel.gx = gx; sel.gy = gy; sel.x1 = x; sel.y1 = y; sel.state |= 1;
      }
      else if ( (button == Button2) || (button == Button3) )
        selection_paste(be->time);
      break;
    case ButtonRelease:
      if (__mousemask != 0) goto not_sel;
      be = (/*const*/ XButtonEvent*) event;
      button = be->button;
      if ( (button == Button1) && (sel.state & 1) )
      { int x, y;
        if ( (be->x == sel.gx) && (be->y == sel.gy) )
        { selection_unhighlight(); goto not_sel; }
        if (!coord_pixel2char(be->x, be->y, &x, &y, falsE)) goto not_sel;
        sel.x2 = x; sel.y2 = y; selection_copy(be);
      }
      not_sel: sel.state &= ~1; break;
    case MotionNotify: /* highlight the currently selected text */
      if (sel.state & 1)
      { XMotionEvent* me = (XPointerMovedEvent*) event;
        int x2, y2;
        if (coord_pixel2char(me->x, me->y, &x2, &y2, falsE))
          selection_highlight(x2, y2);
      }
      break;
    case SelectionClear: selection_unhighlight(); break;
    case SelectionRequest:
      selection_provide((XSelectionRequestEvent*) event); break;
    case SelectionNotify:
      selection_do_paste((XSelectionEvent*) event); break;
    case DestroyNotify:
      do_finish: /* XCloseDisplay(xws_display); */
      do_quit(); /*@notreached@*/ break;
    case ConfigureNotify:
      width = event->xconfigure.width / font_normal->width;
      if (width < 10) width = 10;
      else if (width > 300) width = 300;
      height = event->xconfigure.height / font_normal->height;
      if (height < 5) height = 5;
      else if (height > 300) height = 300;
      COLS = width; LINES = height; reinit_stdscr_data(); redraw_stdscr();
      write_key(KEY_RESIZE);
      break;
    case ClientMessage:
      if ( (event->xclient.message_type == WmProt) &&
           ( ((Atom)(event->xclient.data.l[0])) == WmDelete) )
        goto do_finish;
      break;
  }
}


/* curses interface */

int addch(chtype c)
{ const int x = stdscr->x, idx = SAI(x, stdscr->y);
  stdscr->text[idx] = c & A_CHARTEXT;
  stdscr->attr[idx] = (c & ~A_CHARTEXT) | attr;
  if (x < COLS - 1) stdscr->x++;
  must_redraw(1 | 2);
  return(0);
}

int addnstr(const char* _str, int len)
{ const unsigned char* str = (const unsigned char*) _str; /* avoid sign ext. */
  while (len-- > 0)
  { const char ch = *str++;
    if (ch == '\0') break;
    (void) addch( ((chtype) ch) | attr );
  }
  return(0);
}

int attron(attr_t a)
{ if (a & __A_COLORMARK) attr &= ~__A_COLORPAIRMASK; /* caller changes color */
  attr |= a;
  return(0);
}

int attroff(attr_t a)
{ if (a & __A_COLORMARK) a |= __A_COLORPAIRMASK; /* caller turns color off */
  attr &= ~a;
  return(0);
}

int clrtoeol(void)
{ int x = stdscr->x;
  if (x < COLS)
  { const int y = stdscr->y;
    while (x < COLS)
    { const int idx = SAI(x, y);
      stdscr->text[idx] = ' '; stdscr->attr[idx] = 0; x++;
    }
    must_redraw(2);
  }
  return(0);
}

int getch(void)
{ tKey key;
  if (my_read(fd_keyboard_input, &key, sizeof(key)) != sizeof(key))
    fatal_error(0, "xcurses getch() failed");
  return(key);
}

int getmouse(MEVENT* e)
{ tGetmouse* entry = getmouse_list;
  if (entry == NULL) return(ERR); /* "should not happen" */
  my_memcpy(e, &(entry->data), sizeof(MEVENT));
  getmouse_list = entry->next; memory_deallocate(entry);
  return(OK);
}

int has_colors(void)
{ return( (XDisplayCells(xws_display, DefaultScreen(xws_display)) > 2)
    ? TRUE : FALSE );
}

chtype inch(void)
{ const int idx = SAI(stdscr->x, stdscr->y);
  return(stdscr->text[idx] | stdscr->attr[idx]);
}

int init_pair(short p, short fg, short bg)
{ unsigned long pix = pixel_fg, pix0;
  XColor dummy, *stor = &(colors_storage[p]);
  const char* name;
  switch (fg)
  { case ccRed: name = "red"; break;
    case ccBlue: name = "blue"; break;
    default: goto out; /*@notreached@*/ break;
  }
  if ( (XAllocNamedColor(xws_display, colormap, name, stor, &dummy) != 0) )
  { pix0 = stor->pixel;
    if (pix0 != pixel_bg) pix = pix0;
  }
  out:
  colors[p] = pix;
  return(0);
}

mmask_t mousemask(mmask_t new_mask, mmask_t* old_mask)
{ if (old_mask != NULL) *old_mask = __mousemask;
  __mousemask = new_mask;
  set_cursor(__mousemask ? cursor_arrow : cursor_textmark);
  return(__mousemask);
}

int move(int y, int x)
{ coord_limit(&x, &y);
  stdscr->x = x; stdscr->y = y; must_redraw(1);
  return(0);
}

int refresh(void)
{ loop:
  while (XPending(xws_display) > 0)
  { XEvent event;
    (void) XNextEvent(xws_display, &event); /* CHECKME: return value? */
    handle_event(&event);
  }
  if (__must_redraw) { redraw_stdscr(); goto loop; }
  return(0);
}
