/* retawq/resource.c - resource handling (network, cache, DNS, ...)
   This file is part of retawq (<http://retawq.sourceforge.net/>), a network
   client created by Arne Thomassen; retawq is basically released under certain
   versions of the GNU General Public License and WITHOUT ANY WARRANTY.
   Read the file COPYING for license details, README for program information.
   Copyright (C) 2001-2004 Arne Thomassen <arne@arne-thomassen.de>
*/

#include "stuff.h"
#include "resource.h"
#include "parser.h"

#include <time.h>

#if OPTION_THREADING == 2
#include <signal.h>
#endif

#if HAVE_DIRENT_H
#include <dirent.h>
#else
#define dirent direct
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#if HAVE_NETDB_H
#include <netdb.h>
#endif

#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

#if OPTION_TLS == TLS_OPENSSL
#include <openssl/err.h>
#include <openssl/rand.h>
#endif

declare_local_i18n_buffer
__define_filestuff

#if OPTION_IPV6 && defined(AF_INET6)
/* "user wants" && "library can" */
#define USE_IPv6 1
#else
#define USE_IPv6 0
#endif

#if USE_IPv6
typedef struct sockaddr_storage tSockaddr;
#define HINT_AF (AF_UNSPEC)
#else
typedef struct sockaddr tSockaddr;
#define HINT_AF (AF_INET)
#endif


/* Threading stuff */

#if OPTION_THREADING == 0

/* nothing */

#elif OPTION_THREADING == 1

/* use the pthreads library */
#include <pthread.h>
#define THREADRETTYPE void*
#define THREADRETVAL NULL

#elif OPTION_THREADING == 2

/* use the Linux syscall clone() */
#include <sched.h>
#define THREADRETTYPE int
#define THREADRETVAL 0

#else /* #if OPTION_THREADING... */

#error Bad multi-threading configuration; check option "OPTION_THREADING"!

#endif /* #if OPTION_THREADING... */


/* Resource protocols */

#define RPCFG(cfg, rp) (((cfg) * (rp)) + ((1 - (cfg)) * rpDisabled)) /* :-) */

const tRpData rp_data[rpMax + 1] =
{ { strUnknown, rpUnknown, rpfUsesAuthority },
  { strBracedDisabled, rpDisabled, rpfNone },
  { strHttp, rpHttp, rpfIsHttplike | rpfUsesAuthority | rpfIsPathHierarchical |
    rpfUiToupper },
  { strFtp, RPCFG(CONFIG_FTP, __rpFtp), rpfIsFtplike | rpfUsesAuthority |
    rpfIsPathHierarchical | rpfUiToupper },
  { strLocal, rpLocal, rpfIsLocallike | rpfIsPathHierarchical },
  { strAbout, rpAbout, rpfNone },
  { strFinger, RPCFG(CONFIG_FINGER, __rpFinger), rpfUsesAuthority },
  { strLocalCgi, RPCFG(OPTION_LOCAL_CGI, __rpLocalCgi), rpfIsLocallike |
    rpfIsPathHierarchical },
  { strNews, RPCFG(OPTION_NEWS, __rpNntp), rpfUsesAuthority | rpfUiAlternate |
    rpfUiToupper },
  { strPop, RPCFG(OPTION_POP, __rpPop), rpfUsesAuthority | rpfUiAlternate |
    rpfUiToupper | rpfIsPoplike },
  { strPops, RPCFG((OPTION_TLS ? 1 : 0) * OPTION_POP, __rpPops),
    rpfUsesAuthority | rpfIsPoplike | rpfIsTlslike },
  { strCvs, rpCvs, rpfUsesAuthority | rpfIsPathHierarchical | rpfUiToupper },
  { strGopher, rpGopher, rpfUsesAuthority },
  { strInfo, rpInfo, rpfIsPathHierarchical },
  { strMailto, RPCFG(OPTION_EDITOR, __rpMailto), rpfNone },
  { strJavascript, RPCFG(CONFIG_JAVASCRIPT, __rpJavascript), rpfNone },
  { strHttps, RPCFG((OPTION_TLS ? 1 : 0), __rpHttps), rpfIsHttplike |
    rpfIsTlslike | rpfUsesAuthority | rpfIsPathHierarchical },
  { strFtps, RPCFG(((OPTION_TLS ? 1 : 0) * CONFIG_FTP), __rpFtps),
    rpfIsFtplike | rpfIsTlslike | rpfUsesAuthority | rpfIsPathHierarchical },
  { strExecextShell, rpDisabled, rpfNone }
};

#undef RPCFG


/* Strings */

/* String buffers for the resource handler: */
static char strbuf[STRBUF_SIZE], strbuf2[STRBUF_SIZE];
#if CONFIG_DEBUG
static char debugstrbuf[STRBUF_SIZE];
#endif

static /*@observer@*/ const char strTcp[] = "tcp", strPop3[] = "pop3",
  strGet[] = "GET", strPost[] = "POST", strPercsColonPercs[] = "%s:%s",
  strCrlf[] = "\r\n", strNewlineBr[] ="\n<br>",
  strEndpNewlineEndBodyHtml[] = "</p>\n</body></html>",
  strAxwfu[] = "application/x-www-form-urlencoded";
#define strBr (strNewlineBr + 1)                              /* ug */
#define strNewlineEndBodyHtml (strEndpNewlineEndBodyHtml + 4) /* ly */
#define strEndBodyHtml (strNewlineEndBodyHtml + 1)            /* :-) */

#if CONFIG_FTP || OPTION_POP
static const char strNetcmdUser[] = "USER %s\r\n",
  strNetcmdPass[] = "PASS %s\r\n";
#endif

static const char strStopped[] = N_("Stopped"), strError[] = N_("error"),
  strNoError[] = N_("(no error)"), strInitialization[] = N_("initialization"),
  strByte[] = N_("byte"), strBytes[] = N_("bytes");
const char strRedirection[] = N_("redirection");

const char* const strResourceState[MAX_RESOURCE_STATE + 1] =
{ N_("(resource state unknown - internal error)"), strError,
  N_("Connecting"), N_("Receiving"), N_("Document complete"), strStopped
};

const char* const strResourceError[MAX_RESOURCE_ERROR + 1] =
{ strNoError, N_("Bad request kind (internal error)"),
  N_("Can't create socket"), N_("Connection failed"),
  N_("Connection refused"), N_("Specific network unreachable"),
  N_("Can't lookup host name"), N_("Can't read/write file/directory"),
  N_("Bad URL"), N_("Server closed connection"), N_("Internal handshake bug"),
  N_("Port number invalid"), N_("Unknown protocol"),
  N_("Time-out (server busy)"), N_("Server response invalid"),
  strLoginFailed, N_("Redirection problem"),
  N_("Can't calculate random numbers"), N_("TLS-related error"),
  N_("Proxy authentication required"), N_("Can't create internal pipe"),
  N_("Can't create new process"), N_("Forbidden by configuration"),
  N_("Can't execute program"), N_("Protocol disabled"), N_("Hostname missing"),
  N_("Too many open file descriptors")
};

#if OPTION_TLS
const char* const strTlsError[MAX_TLS_ERROR + 1] =
{ strNoError, strUnknown, N_("session expired"), N_("fatal alert"),
  N_("file"), N_("application data"), N_("memory"), N_("ciphers"),
  N_("(de-)compression"), N_("(de-)cryption"), "SRP", N_("public key"),
  N_("packet"), N_("library feature unimplemented"), N_("handshake"),
  strInitialization
};
#endif

const char* const strResourceRequestState[MAX_RESOURCE_REQUEST_STATE + 1] =
{ N_("(request state unknown - internal error)"), strError,
  N_("Preparing"), N_("(Attached resource)"), strStopped,
  N_("Looking up host name")
};


/* Protocols, ports 'n' pipes (aka PPP:-) */

static const_after_init int ipprotocolnumber_tcp = IPPROTO_TCP; /* default */

static const_after_init tPortnumber portnumber_http, portnumber_ftp,
  portnumber_finger, portnumber_cvs, portnumber_gopher, portnumber_nntp,
  portnumber_pop3, portnumber_https;
  /* all these port numbers are stored in network byte order */

const_after_init int fd_any2main_read, fd_any2main_write;
int highest_select_fd = 0;
fd_set fds_read, fds_write, fds_r, fds_w, fds_e;
#if OPTION_THREADING
static const_after_init int fd_dns2resource_read, fd_dns2resource_write,
  fd_resource2dns_read, fd_resource2dns_write;
#endif
#if CONFIG_TG == TG_XCURSES
const_after_init int fd_xcurses2main_read, fd_xcurses2main_write;
#endif

#if OPTION_THREADING == 2
static pid_t pid_main = 0, pid_dns = 0; /* PIDs of our threads */
#endif


/* DNS lookups */

enum { dlfNone = 0, dlfFailed = 0x01, dlfTryIpv6 = 0x02 };
typedef unsigned char tDnsLookupFlags;

typedef struct
{ tSockaddr addrinfo;
  struct tCachedHostInformation* hostinfo;
    /* the hostinfo for which the lookup is done; not for use by DNS handler */
  const char* hostname;
  size_t addrinfolen;
  int address_family;
  tDnsLookupFlags flags;
} tDnsLookup;

#define my_h_addr h_addr
  /* (currently; we'll use h_addr_list[0] in a later version if available) */

static one_caller void my_getaddrinfo(tDnsLookup* dns_lookup)
/* performs a DNS hostname lookup; the monstrous size of the source code of
   this function is only a result of the attempt to make it as portable as
   possible; you could stop reading after the getaddrinfo() stuff... */
{ const char* hostname = dns_lookup->hostname;
#if HAVE_GETADDRINFO || HAVE_GETIPNODEBYNAME
  int err;
#endif
#if HAVE_GETADDRINFO
  static tBoolean did_init_hints = falsE;
  static struct addrinfo hints;
  struct addrinfo* result;
  if (!did_init_hints)
  { did_init_hints = truE; my_memclr_var(hints); hints.ai_family = HINT_AF;
    hints.ai_socktype = SOCK_STREAM; hints.ai_protocol = ipprotocolnumber_tcp;
  }
#else
  struct hostent* result;
#endif

#if HAVE_GETADDRINFO

  result = NULL; /* (extra care for buggy libraries) */
  err = getaddrinfo(hostname, NULL, &hints, &result);
  if ( (err == 0) && (result != NULL) && (result->ai_addr != NULL) )
  { size_t len = result->ai_addrlen;
    if ( (len <= 0) || (len > sizeof(tSockaddr)) ) /* "should not happen" */
      goto failed;
    my_memcpy(&(dns_lookup->addrinfo), result->ai_addr, len);
    dns_lookup->addrinfolen = len;
    dns_lookup->address_family = result->ai_family;
  }
  else
  { failed:
    dns_lookup->flags |= dlfFailed;
    debugmsg(gai_strerror(err)); debugmsg(strNewline);
  }
  if (result != NULL) freeaddrinfo(result);

#elif HAVE_GETIPNODEBYNAME

  result = getipnodebyname(hostname, AF_INET, 0, &err);
  if ( (result != NULL) && (result->my_h_addr != NULL) &&
       (result->h_addrtype == AF_INET) && (result->h_length == 4) )
  { /* (The above h_addrtype and h_length tests "should" be unnecessary.) */
    struct sockaddr_in* addr =
      (struct sockaddr_in*) (&(dns_lookup->addrinfo));
    dns_lookup->address_family = addr->sin_family = AF_INET;
    my_memcpy(&(addr->sin_addr.s_addr), result->my_h_addr, 4);
    dns_lookup->addrinfolen = sizeof(struct sockaddr_in);
    freehostent(result);
    return;
  }
  if (result != NULL) freehostent(result);
#if USE_IPv6
  result = getipnodebyname(hostname, AF_INET6, 0, &err);
  if ( (result != NULL) && (result->my_h_addr != NULL) &&
       (result->h_addrtype == AF_INET6) && (result->h_length == 16) )
  { /* (The above h_addrtype and h_length tests "should" be unnecessary.) */
    struct sockaddr_in6* addr =
      (struct sockaddr_in6*) (&(dns_lookup->addrinfo));
    dns_lookup->address_family = addr->sin6_family = AF_INET6;
    my_memcpy(&(addr->sin6_addr.s6_addr), result->my_h_addr, 16);
    dns_lookup->addrinfolen = sizeof(struct sockaddr_in6);
    freehostent(result);
    return;
  }
  if (result != NULL) freehostent(result);
#endif /* #if USE_IPv6 */
  dns_lookup->flags |= dlfFailed;

#else /* #if HAVE_GETADDRINFO / #elif HAVE_GETIPNODEBYNAME */

  /* The nice getaddrinfo() and the ugly getipnodebyname() aren't available, so
     we have to work harder. We can't simply rely on gethostbyname() because
     "the behavior of gethostbyname() when passed a numeric address string is
     unspecified" (SUSv3); so we must try several other functions in advance.
     (The word "unspecified" could at least in theory mean that gethostbyname()
     returns a non-NULL value pointing to rubbish. And there are so many buggy
     libraries around...) */

  /* Check whether <hostname> is a numerical host address: */
#if HAVE_INET_PTON
#if USE_IPv6
  if (dns_lookup->flags & dlfTryIpv6)
  { char resbuf[16];
    int result = inet_pton(AF_INET6, hostname, resbuf);
    if (result == 1) /* that worked */
    { struct sockaddr_in6* addr =
        (struct sockaddr_in6*) (&(dns_lookup->addrinfo));
      dns_lookup->address_family = addr->sin6_family = AF_INET6;
      my_memcpy(&(addr->sin6_addr.s6_addr), resbuf, 16);
      dns_lookup->addrinfolen = sizeof(struct sockaddr_in6);
      return;
    }
  }
#endif /* #if USE_IPv6 */
  { char resbuf[16]; /* (unnecessarily large, just for buggy libraries) */
    int result = inet_pton(AF_INET, hostname, resbuf);
    if (result == 1) /* that worked */
    { struct sockaddr_in* addr =
        (struct sockaddr_in*) (&(dns_lookup->addrinfo));
      dns_lookup->address_family = addr->sin_family = AF_INET;
      my_memcpy(&(addr->sin_addr.s_addr), resbuf, 4);
      dns_lookup->addrinfolen = sizeof(struct sockaddr_in);
      return;
    }
  }
#elif HAVE_INET_ADDR /* #if HAVE_INET_PTON */
#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long int) (-1))
#endif
  { unsigned long int res = (unsigned long int) inet_addr(hostname);
      /* (The correct formal type would be in_addr_t, but glibc before
          2000-04-01 (and probably other libraries) doesn't have it...) */
    if (res != INADDR_NONE) /* that worked */
    { struct sockaddr_in* addr =
        (struct sockaddr_in*) (&(dns_lookup->addrinfo));
      dns_lookup->address_family = addr->sin_family = AF_INET;
      addr->sin_addr.s_addr = res;
      dns_lookup->addrinfolen = sizeof(struct sockaddr_in);
      return;
    }
  }
#endif /* #if HAVE_INET_PTON / #elif HAVE_INET_ADDR */

  /* Seems that <hostname> isn't a numerical host address... */
  result = gethostbyname(hostname);
  if ( (result != NULL) && (result->my_h_addr != NULL) &&
       (result->h_addrtype == AF_INET) && (result->h_length == 4) )
  { /* (The above h_addrtype and h_length tests "should" be unnecessary.) */
    struct sockaddr_in* addr =
      (struct sockaddr_in*) (&(dns_lookup->addrinfo));
    dns_lookup->address_family = addr->sin_family = AF_INET;
    my_memcpy(&(addr->sin_addr.s_addr), result->my_h_addr, 4);
    dns_lookup->addrinfolen = sizeof(struct sockaddr_in);
  }
  else
  { dns_lookup->flags |= dlfFailed;
  }

#endif /* #if HAVE_GETADDRINFO / #elif HAVE_GETIPNODEBYNAME */

}

#if OPTION_THREADING

static THREADRETTYPE dns_handler_thread(__sunused void* dummydata __cunused)
/* This function is the DNS handler thread; it checks for DNS lookup demands
   from the resource handler forever. Apart from its local variables, it only
   accesses the tDnsLookup structure given by the resource handler (which won't
   touch it while the lookup is running), so there aren't any multi-threading
   race conditions possible. */
{ while (1)
  { static tDnsLookup* dns_lookup;
    static unsigned char count = 0;
    int err = my_read(fd_resource2dns_read, ((char*)(&dns_lookup)) + count,
      sizeof(dns_lookup) - count);
    if (err <= 0) fatal_error((err == -1) ? errno : 0, _("read(DNS) failed"));
    count += err;
    if (count >= sizeof(dns_lookup)) /* got a complete pointer */
    { count = 0; my_getaddrinfo(dns_lookup);
      my_write_crucial(fd_dns2resource_write, &dns_lookup, sizeof(dns_lookup));
    }
  }
  /*@notreached@*/
  return(THREADRETVAL); /* avoid compiler warning for compilers which are too
    stupid to recognize simple infinite loops... */
}

#endif /* #if OPTION_THREADING */


/* Helper functions I */

static tBoolean is_time_valid = falsE;

static time_t my_time(void)
{ static time_t current_time = 0;
  if (!is_time_valid)
  { time_t t = time(NULL);
    if (t != -1) { current_time = t; is_time_valid = truE; }
  }
  return(current_time);
}

#if OPTION_LOCAL_CGI || OPTION_TLS || OPTION_EXECEXT
static tBoolean may_use_fd2(void)
{ static tBoolean did_calc = falsE, retval;
  if (!did_calc)
  { did_calc = truE;
    if ( (lfdmbs) &&
         (
#if TGC_IS_CURSES
           (!is_environed) || /* don't ruin the curses screen handling... */
#endif
           (!isatty(2)) ) )
      retval = truE;
    else retval = falsE;
  }
  return(retval);
}
#endif

#if OPTION_LOCAL_CGI || OPTION_EXECEXT
static tBoolean my_dup2(int fd1, int fd2)
/* returns whether it worked; IMPORTANT: this function is only intended for use
   in child processes after a fork(); otherwise we could get a race condition
   with retawq's other threads here because there's a window between close()
   and fcntl() - even within the C library's dup2(), unless e.g. an atomic
   dup2() in the OS is used */
{ unsigned char loopcount = 0;
  int err;
#if HAVE_DUP2
  do
  { err = dup2(fd1, fd2);
  } while ( (err == -1) && ( (errno == EINTR) || (errno == EBUSY) ) &&
            (++loopcount < 100) ); /* (EBUSY is for Linux) */
#else
  my_close(fd2);
  do
  { err = fcntl(fd1, F_DUPFD, fd2);
  } while ( (err == -1) && (errno == EINTR) && (++loopcount < 100) );
#endif
  return(cond2boolean(err >= 0));
}
#endif

static my_inline void resource_set_error(tResource* resource,
  tResourceError error)
{ if (resource->state != rsError)
  { resource->state = rsError; resource->error = error; }
}

static void resource_request_set_error(tResourceRequest* request,
  tResourceError error)
{ tResource* resource = request->resource;
  if (request->state != rrsError)
  { request->state = rrsError; request->error = error; }
  if (resource != NULL) resource_set_error(resource, error);
}

void fatal_tmofd(int fd)
{ sprint_safe(strbuf,
    _("%s (given value is %d, highest select()-able value is %d)"),
    _(strResourceError[reTmofd]), fd, FD_SETSIZE - 1);
  fatal_error(0, strbuf);
}

my_inline void my_FD_SET(int fd, fd_set* fds_addr)
{ if (highest_select_fd < fd)
  {
#if CONFIG_DEBUG
    sprint_safe(debugstrbuf, "highest_select_fd: %d,%d\n",
      highest_select_fd, fd);
    debugmsg(debugstrbuf);
#endif
    highest_select_fd = fd;
  }
  FD_SET(fd, fds_addr);
}

my_inline tBoolean my_try_FD_SET(int fd, fd_set* fds_addr)
/* returns whether it worked */
{ tBoolean retval;
  if (is_fd_selectable(fd)) { my_FD_SET(fd, fds_addr); retval = truE; }
  else retval = falsE; /* too high, can't select() it */
  return(retval);
}

static __sallocator one_caller char* __callocator
  base64_encode(const unsigned char* src)
/* converts an arbitrary string to base64 encoding (RFC3548, 3.) */
{ static /*@observer@*/ const unsigned char conv[] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  size_t len = strlen((const char*) src), destlen = ((len + 2) / 3) * 4 + 1;
  unsigned char ch, *dest = __memory_allocate(destlen), *ptr = dest, count = 0;
  unsigned int convval = 0;
#define PUT(offset) *ptr++ = conv[(convval >> offset) & 63]
  while ((ch = *src++) != '\0')
  { convval |= ((unsigned int) ch);
    if (++count < 3) convval <<= 8;
    else { PUT(18); PUT(12); PUT(6); PUT(0); convval = 0; count = 0; }
  }
  if (count > 0) /* something left to do */
  { if (count == 1) convval <<= 8;
    PUT(18); PUT(12);
    if (count == 2) PUT(6);
    else *ptr++ = '=';
    *ptr++ = '=';
  }
#undef PUT
  *ptr = '\0';
  return((char*) dest);
}

#if OPTION_POP || OPTION_TRAP || CONFIG_DEBUG

static one_caller void md5_decode(/*@out@*/ tUint32* dest, const tUint8* src,
  size_t count)
{ while (count-- > 0)
  { *dest++ = ((tUint32) src[0]) | (((tUint32) src[1]) << 8) |
     (((tUint32) src[2]) << 16) | (((tUint32) src[3]) << 24);
    src += 4;
  }
}

static void md5_encode(/*@out@*/ tUint8* dest, const tUint32* src,
  size_t count)
{ while (count-- > 0)
  { unsigned char cnt;
    tUint32 value = *src++;
    for (cnt = 0; cnt <= 3; cnt++) { *dest++ = value & 255; value >>= 8; }
  }
}

#define md5_prepare(roundnum) do { currfunc = func[roundnum]; currshift = shift[roundnum]; xidxval = xidxvals[roundnum]; xidxoff = xidxoffs[roundnum]; } while (0)
#define md5_rotate(val, cnt) (((val) << (cnt)) | ((val) >> (32-(cnt))))
#define md5_operate(a, b, c, d) do { (a) += (currfunc(b, c, d)) + (x[xidxval & 15]) + sineval[stepcount]; (a) = md5_rotate((a), currshift[stepcount & 3]); (a) += (b); stepcount++; si--; xidxval += xidxoff; } while (0)

typedef tUint32 (*tMd5Function)(tUint32, tUint32, tUint32);
static tUint32 md5_F(tUint32 a, tUint32 b, tUint32 c)
{ return(((a) & (b)) | ((~a) & (c))); }
static tUint32 md5_G(tUint32 a, tUint32 b, tUint32 c)
{ return(((a) & (c)) | ((b) & (~c))); }
static tUint32 md5_H(tUint32 a, tUint32 b, tUint32 c)
{ return((a) ^ (b) ^ (c)); }
static tUint32 md5_I(tUint32 a, tUint32 b, tUint32 c)
{ return((b) ^ ((a) | (~c))); }

static void md5_digest_block(tUint32* state, const tUint8* data)
/* Maybe slow, but small; speed doesn't matter much here because this is rarely
   used, if ever, and only with rather small amounts of data. */
{ static const tUint32 sineval[64] =
  { 0xd76aa478, 0xe8c7b756, 0x242070db, 0xc1bdceee, 0xf57c0faf, 0x4787c62a,
    0xa8304613, 0xfd469501, 0x698098d8, 0x8b44f7af, 0xffff5bb1, 0x895cd7be,
    0x6b901122, 0xfd987193, 0xa679438e, 0x49b40821, 0xf61e2562, 0xc040b340,
    0x265e5a51, 0xe9b6c7aa, 0xd62f105d,  0x2441453, 0xd8a1e681, 0xe7d3fbc8,
    0x21e1cde6, 0xc33707d6, 0xf4d50d87, 0x455a14ed, 0xa9e3e905, 0xfcefa3f8,
    0x676f02d9, 0x8d2a4c8a, 0xfffa3942, 0x8771f681, 0x6d9d6122, 0xfde5380c,
    0xa4beea44, 0x4bdecfa9, 0xf6bb4b60, 0xbebfbc70, 0x289b7ec6, 0xeaa127fa,
    0xd4ef3085,  0x4881d05, 0xd9d4d039, 0xe6db99e5, 0x1fa27cf8, 0xc4ac5665,
    0xf4292244, 0x432aff97, 0xab9423a7, 0xfc93a039, 0x655b59c3, 0x8f0ccc92,
    0xffeff47d, 0x85845dd1, 0x6fa87e4f, 0xfe2ce6e0, 0xa3014314, 0x4e0811a1,
    0xf7537e82, 0xbd3af235, 0x2ad7d2bb, 0xeb86d391
  };
  static const unsigned char shift1[4] = {7,12,17,22}, shift2[4] = {5,9,14,20},
    shift3[4] = {4,11,16,23}, shift4[4] = {6,10,15,21}, xidxvals[4] ={0,1,5,0},
    xidxoffs[4] = {1,5,3,7}, *shift[4] = { shift1, shift2, shift3, shift4 };
  static const tMd5Function func[4] = { md5_F, md5_G, md5_H, md5_I };
  unsigned char stepcount = 0, si = 128, roundnum, count, xidxval, xidxoff;
  const unsigned char* currshift;
  tMd5Function currfunc;
  tUint32 x[16], cs[4];

  md5_decode(x, data, 16); my_memcpy(cs, state, sizeof(cs));
  for (roundnum = 0; roundnum <= 3; roundnum++)
  { md5_prepare(roundnum);
    for (count = 0; count <= 15; count++)
      md5_operate(cs[si & 3], cs[(si+1) & 3], cs[(si+2) & 3], cs[(si+3) & 3]);
  }
  for (count = 0; count <= 3; count++) state[count] += cs[count];
}

static void md5_digest(const tUint8* data, size_t size,
  /*@out@*/ tUint8* result)
{ tUint32 state[4] = { 0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476 };
  size_t numobits = size * 8, tsize;
  tUint8 tailbuf[128], *tail = tailbuf;

  while (size >= 64) { md5_digest_block(state, data); data += 64; size -= 64; }

  if (size > 0) my_memcpy(tailbuf, data, size);
  tsize = ( (size >= 56) ? 128 : 64 );
  tailbuf[size++] = 0x80; my_memclr(tailbuf + size, tsize - size);
  md5_encode(tailbuf + tsize - 8, &numobits, 1);
  while (tsize >= 64) { md5_digest_block(state, tail); tail += 64; tsize-=64; }

  md5_encode(result, state, 4);
}

#endif /* need-md5 */


/* Host information cache; address lookups */

my_enum1 enum
{ hppifNone = 0, hppifUserQueryRunning = 0x01, hppifCannotHttp11 = 0x02,
  __hppifFtpCannotEpsv = 0x04, __hppifNntpCannotPostArticles = 0x08
} my_enum2(unsigned char) tHostPortProtInfoFlags;
#if CONFIG_FTP
#define hppifFtpCannotEpsv (__hppifFtpCannotEpsv)
#endif
#if OPTION_NEWS
#define hppifNntpCannotPostArticles (__hppifNntpCannotPostArticles)
#endif

#define hppifAddrRel \
  (hppifCannotHttp11 | __hppifFtpCannotEpsv | __hppifNntpCannotPostArticles)
  /* stuff which might change (e.g. because we get a different address at which
     a different server program is running) when a (re-)lookup is performed */

typedef struct tLoginData
{ struct tLoginData* next;
  const char *username, *password /* , *account */;
} tLoginData;

typedef struct tHostPortProtInfo
{ struct tHostPortProtInfo* next;
  struct tCachedHostInformation* host;
  tLoginData* login_data;
  const char* software_id;
  tPortnumber portnumber; /* (in network byte order) */
  tResourceProtocol protocol;
  tHostPortProtInfoFlags flags;
} tHostPortProtInfo;

my_enum1 enum
{ chifNone = 0, chifAddressValid = 0x01, chifAddressLookupFailed = 0x02,
  __chifAddressLookupRunning = 0x04
} my_enum2(unsigned char) tCachedHostInformationFlags;

#define chifAddrRel \
  (chifAddressValid | chifAddressLookupFailed | __chifAddressLookupRunning)

#if CONFIG_ASYNC_DNS
#define chifAddressLookupRunning (__chifAddressLookupRunning)
#endif

#if OPTION_COOKIES
struct tCookie;
#endif

#if OPTION_NEWS
struct tNewsGroupInformation;
#endif

typedef struct tCachedHostInformation
{ tSockaddr addrinfo;
  struct tCachedHostInformation* next;
  struct tHostPortProtInfo* hppi;
  const char* hostname;
#if CONFIG_ASYNC_DNS
  tResourceRequest** waiting; /* requests which "wait" for a lookup result */
#endif
#if OPTION_COOKIES
  struct tCookie* cookies; /* cookies for this host */
#endif
#if OPTION_NEWS
  struct tNewsGroupInformation* ngi;
#endif
  size_t addrinfolen;
  tReqlistNum waiting_num, waiting_maxnum;
  int address_family;
  time_t lookupfailuretime;
  tCachedHostInformationFlags flags;
#if OPTION_COOKIES
  unsigned char cookiecount; /* how many cookies we've stored for this host */
#endif
} tCachedHostInformation;

#define resource2actual_host(resource)  ((resource)->actual_hppi->host)
#define resource2textual_host(resource) ((resource)->textual_hppi->host)

static tHostPortProtInfo* hppi_lookup(tCachedHostInformation* hostinfo,
  tPortnumber portnumber, tResourceProtocol protocol, tBoolean create_if_null)
{ tHostPortProtInfo* hppi = hostinfo->hppi;
  while (hppi != NULL)
  { if ( (hppi->portnumber == portnumber) && (hppi->protocol == protocol) )
      goto out; /* found */
    hppi = hppi->next;
  }
  if ( (hppi == NULL) && (create_if_null) )
  { hppi = (tHostPortProtInfo*) memory_allocate(sizeof(tHostPortProtInfo));
    hppi->next = hostinfo->hppi; hostinfo->hppi = hppi; hppi->host = hostinfo;
    hppi->portnumber = portnumber; hppi->protocol = protocol;
  }
  out:
  return(hppi);
}

static void hppi_set_software_id(tHostPortProtInfo* hppi, const char* text)
{ const char* r = my_strchr(text, '\r');
  __dealloc(hppi->software_id);
  hppi->software_id = ( (r != NULL) ? my_strndup(text, r - text) :
    my_strdup(text) );
}

static void hppi_login_set(tHostPortProtInfo* hppi, const char* username,
  const char* password)
{ tLoginData* ld;
  if ( (username == NULL) || (password == NULL) ) /* "should not happen" */
    return;
  ld = hppi->login_data;
  while (ld != NULL)
  { const char* u = ld->username;
    if ( (u != NULL) && (!strcmp(u, username)) )
    { const char* p = ld->password;
      if ( (p == NULL) || (strcmp(p, password)) )
        my_strdedup(ld->password, password);
      return;
    }
    ld = ld->next;
  }
  ld = memory_allocate(sizeof(tLoginData));
  ld->username = my_strdup(username); ld->password = my_strdup(password);
  ld->next = hppi->login_data; hppi->login_data = ld;
}

static void hppi_login_get(const tHostPortProtInfo* hppi,
  /*@out@*/ const char** username, /*@out@*/ const char** password)
{
}

#define HASHTABLESIZE (101)
typedef unsigned short tHashIndex;

#define HASHTABLESIZE_CHI (HASHTABLESIZE)
static tCachedHostInformation* chi_head[HASHTABLESIZE_CHI];

static tHashIndex hostinfo_hostname2hashindex(const char* hostname)
/* calculates a hash table index from the given hostname; we take at most the
   first ten characters of the name into account - this "should be enough" for
   good hashing and we need not do "%" after each partial operation (because
   the maximum possible value, 10 * 255, fits into a tHashIndex variable). */
{ tHashIndex retval = 0;
  unsigned char count = 10;
  while (count-- > 0)
  { unsigned char ch = *((const unsigned char*) (hostname));
    if (ch == '\0') break; /* reached end of string */
    retval += ((tHashIndex) ch);
    hostname++;
  }
  return(retval % HASHTABLESIZE_CHI);
}

static /* __sallocator -- not an "only" reference... */
  tCachedHostInformation* __callocator hostinfo_create(const char* hostname)
/* creates a new record for the host information cache */
{ tHashIndex i = hostinfo_hostname2hashindex(hostname);
  tCachedHostInformation* retval = memory_allocate(sizeof(*retval));
  retval->hostname = my_strdup(hostname);
  retval->next = chi_head[i];
  chi_head[i] = retval;
  return(retval);
}

static /*@observer@*/ tCachedHostInformation* hostinfo_lookup(const char*
  hostname)
/* tries to find known information about the host with the given <hostname> in
   the host information cache */
{ tHashIndex i = hostinfo_hostname2hashindex(hostname);
  tCachedHostInformation* retval = chi_head[i];
  while (retval != NULL)
  { if (streqcase(hostname, retval->hostname)) break; /* found */
        /* IMPROVEME: "!strcmp()" should be enough here. */
    retval = retval->next;
  }
  return(retval);
}

static void reqlist_add(tResourceRequest*** _r, tReqlistNum* _num,
  tReqlistNum* _maxnum, tResourceRequest* request)
{ tResourceRequest** reqlist = *_r;
  tReqlistNum num = *_num, maxnum = *_maxnum;
  if (num >= maxnum)
  { if (maxnum < 2) maxnum = 2; /* (most times only 1 is needed) */
    else maxnum += 3;
    reqlist = memory_reallocate(reqlist, maxnum * sizeof(tResourceRequest*));
    *_r = reqlist; *_maxnum = maxnum;
  }
  reqlist[num++] = request;
  *_num = num;
}

static void reqlist_remove(tResourceRequest*** _r, tReqlistNum* _num,
  tResourceRequest* request)
{ tResourceRequest** reqlist = *_r;
  tReqlistNum num = *_num, count;
  for (count = 0; count < num; count++)
  { if (reqlist[count] == request)
    { if (count < num - 1) reqlist[count] = reqlist[num - 1];
      num--; *_num = num;
      break;
    }
  }
}

static void reqlist_reset(tResourceRequest*** _r, tReqlistNum* _num,
  tReqlistNum* _maxnum)
{ tResourceRequest** reqlist= *_r;
  if (reqlist != NULL)
  { memory_deallocate(reqlist); *_r = NULL; *_num = *_maxnum = 0;
  }
}

enum
{ halrFine = 0, halrLookupFailed = 1, halrNotInCache = 2
#if CONFIG_ASYNC_DNS
  , halrLookupRunning = 3
#endif
};
typedef unsigned char tHostAddressLookupResult;

enum { halfNone = 0, halfInCacheOnly = 0x01, halfEnforcedReload = 0x02 };
typedef unsigned char tHostAddressLookupFlags;

static void check_proxies(tResourceRequest* request, const tConfigProxy* proxy,
  const char** _hostname)
{ tPortnumber reqport = request->textual_portnumber;
  const char* hostname = *_hostname;
  while (proxy != NULL)
  { tPortnumber hp = proxy->hosts_portnumber;
    if ( ( (hp == 0) || (hp == reqport) ) &&
         (my_pattern_matcher(proxy->hosts_pattern, hostname)) )
    { const char* h = proxy->proxy_hostname;
      if (h != NULL) { *_hostname = h; request->proxy = proxy; }
      /* "else": configuration explicitly said: "Don't use a proxy here!" */
      break;
    }
    proxy = proxy->next;
  }
}

static tHostAddressLookupResult lookup_hostaddress(tResourceRequest* request,
  /*@out@*/ tCachedHostInformation** hostinforet, tHostAddressLookupFlags half)
/* tries to lookup the address information for the given hostname in the host
   information cache etc.; if that fails, it tries to start a new DNS lookup */
{ const char* hostname = request->hostname;
  tResourceProtocol protocol = request->protocol;
  tBoolean in_cache_only = cond2boolean(half & halfInCacheOnly),
    enforced_reload = cond2boolean(half & halfEnforcedReload);
  tCachedHostInformation* hostinfo;
  tDnsLookup* dns_lookup;

  /* Check whether we shall use a proxy */

  if (protocol == rpHttp)
    check_proxies(request, config.http_proxies, &hostname);
#if OPTION_TLS
  else if (protocol == rpHttps)
    check_proxies(request, config.https_proxies, &hostname);
#endif

#if CONFIG_DEBUG
  { char* spfbuf;
    my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf,
      "lookup_hostaddress(): %p,%d,%s,%s\n", request, half, request->hostname,
      hostname);
    debugmsg(spfbuf);
    my_spf_cleanup(debugstrbuf, spfbuf);
  }
#endif

  /* Try to find the address for the <hostname> in the cache */

  hostinfo = hostinfo_lookup(hostname);
  *hostinforet = hostinfo; /* (may be NULL) */
  if (hostinfo != NULL)
  { tCachedHostInformationFlags flags = hostinfo->flags;
    if (flags & chifAddressValid) return(halrFine); /* already calculated */
#if CONFIG_ASYNC_DNS
    else if (flags & chifAddressLookupRunning) /* started, not yet finished */
      return(halrLookupRunning);
#endif
    else if (flags & chifAddressLookupFailed)
    { if (enforced_reload)
        goto start_dns_lookup; /* enforce (re-)lookup */
      /* IMPLEMENTME: if ( (!in_cache_only) && (->lookupfailuretime is
         more than e.g. one minute in the past) ) goto start_dns_lookup; */
    }
    return(halrLookupFailed); /* lookup failed or bug occurred */
  }
  if (in_cache_only) return(halrNotInCache);

  /* Seems we have to start a DNS lookup request; we do this by setting up a
     tDnsLookup structure and writing its address to the DNS handler's pipe. */

  start_dns_lookup:
  if (hostinfo == NULL) *hostinforet = hostinfo = hostinfo_create(hostname);
  else
  { tHostPortProtInfo* hppi = hostinfo->hppi;
    while (hppi != NULL) { hppi->flags &= ~hppifAddrRel; hppi = hppi->next; }
    hostinfo->flags &= ~chifAddrRel;
  }
  dns_lookup = (tDnsLookup*) memory_allocate(sizeof(tDnsLookup));
  dns_lookup->hostinfo = hostinfo;
  dns_lookup->hostname = my_strdup(hostname);
  if (request->flags & rrfTryIpv6) dns_lookup->flags |= dlfTryIpv6;
#if CONFIG_DEBUG
  { char* spfbuf;
    my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf,
      "starting DNS lookup at %p for hostinfo %p, hostname %s\n",
      dns_lookup, hostinfo, hostname);
    debugmsg(spfbuf);
    my_spf_cleanup(debugstrbuf, spfbuf);
  }
#endif
#if OPTION_THREADING
  my_write_crucial(fd_resource2dns_write, &dns_lookup, sizeof(dns_lookup));
  hostinfo->flags |= chifAddressLookupRunning;
  return(halrLookupRunning);
#else
  my_getaddrinfo(dns_lookup); /* this call might block the program */
  is_time_valid = falsE; /* might have taken a long time */
  if (dns_lookup->flags & dlfFailed)
  { hostinfo->flags |= chifAddressLookupFailed;
    hostinfo->lookupfailuretime = my_time();
    return(halrLookupFailed);
  }
  else
  { hostinfo->addrinfo = dns_lookup->addrinfo;
    hostinfo->addrinfolen = dns_lookup->addrinfolen;
    hostinfo->address_family = dns_lookup->address_family;
    hostinfo->flags |= chifAddressValid;
    return(halrFine);
  }
#endif
}


/* Sockets and connections I */

static tBoolean make_fd_nonblocking(int fd)
/* tries to make <fd> non-blocking; returns whether it worked */
{ int flags = fcntl(fd, F_GETFL, 0);
  if ( (flags == -1) || (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1) )
    return(falsE);
  else return(truE);
}

static int create_socket(int address_family)
/* creates a new TCP socket and sets it to non-blocking mode; returns -2 if the
   fd wouldn't be select()able, -1 for OS errors or a valid fd if it worked */
{ int retval = socket(address_family, SOCK_STREAM, ipprotocolnumber_tcp);
  if (retval >= 0)
  { if (!is_fd_selectable(retval)) { my_close(retval); retval = -2; }
    else if (!make_fd_nonblocking(retval)) { my_close(retval); retval = -1; }
    else make_fd_cloexec(retval);
  }
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "create_socket(%d): %d\n", address_family, retval);
  debugmsg(debugstrbuf);
#endif
  return(retval);
}

static void set_portnumber(tSockaddr* addr, int address_family,
  tPortnumber portnumber)
{ if (address_family == AF_INET)
    ((struct sockaddr_in*)(addr))->sin_port = portnumber;
#if USE_IPv6
  else if (address_family == AF_INET6)
    ((struct sockaddr_in6*)(addr))->sin6_port = portnumber;
#endif
  /* "else": we don't know the AF, so we don't override default ports. */
}

static int my_do_connect(int fd, const tSockaddr* addr, size_t addrlen)
{ int err;
  unsigned char loopcount = 0;
  loop:
  err = connect(fd, (const struct sockaddr*) addr, (socklen_t) addrlen);
  if (err != 0)
  { if ( (errno == EINTR) && (++loopcount < 100) ) goto loop;
    else if ( (errno == EALREADY) && (loopcount > 0) ) errno = EINPROGRESS;
    else if (errno != EINPROGRESS) { int e = errno; my_close(fd); errno = e; }
  }
  return(err);
}

static one_caller int my_connect(int fd, const tCachedHostInformation*
  hostinfo, tPortnumber portnumber)
{ tSockaddr addr = hostinfo->addrinfo; /* (yes, we copy a structure here) */
  set_portnumber(&addr, hostinfo->address_family, portnumber);
  return(my_do_connect(fd, &addr, hostinfo->addrinfolen));
}

#define connect_err2error conn_err2error /* (currently the same) */
#define reConnectionFailureDefault (reConnect)

static /*@null@*/ tConnection** conn_table = NULL;
typedef unsigned int tConnTableIndex;
static tConnTableIndex conn_table_num = 0, conn_table_maxnum = 0;
static int conn_err; /* temporary storage for connection-related errors */

#if CONFIG_DEBUG
static void debug_conn_callback(const tConnection* conn, tConnCbEventKind ccek)
{ sprint_safe(debugstrbuf, "conn_callback(): fd=%d, ccek=%d, data=%p, x=%d\n",
    conn->fd, ccek, conn->data, conn->x);
  debugmsg(debugstrbuf);
}
static void conn_bug(const tConnection* conn, tConnCbEventKind ccek)
{ sprint_safe(debugstrbuf, "conn_bug(): fd #%d, ccek=%d\n", conn->fd, ccek);
  debugmsg(debugstrbuf);
}
#else
#define debug_conn_callback(conn, ccek) do { } while (0)
#define conn_bug(conn, ccek) do { } while (0)
#endif

#define conn_callback(conn, ccek) \
  do \
  { debug_conn_callback((conn), (ccek)); \
    ((conn)->callback)((conn), (ccek)); \
  } while (0)

#define conn2resource(conn) \
  ( (conn->flags & cnfDataIsResource) ? ((tResource*) (conn->data)) : NULL )

static tConnection* conn_create(int fd, tConnCbFunc callback, void* data,
  tBoolean data_is_resource, const tHostPortProtInfo* hppi,
  tResourceProtocol protocol)
{ tConnection* retval = (tConnection*) memory_allocate(sizeof(tConnection));
  if (conn_table_num >= conn_table_maxnum)
  { conn_table_maxnum += 10;
    conn_table = memory_reallocate(conn_table,
      conn_table_maxnum * sizeof(tConnection*));
  }
  conn_table[conn_table_num++] = retval;

  retval->fd = fd; retval->callback = callback; retval->data = data;
  if (data_is_resource) retval->flags |= cnfDataIsResource;
  retval->hppi = hppi; retval->protocol = protocol;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "conn_create(): fd #%d\n", fd);
  debugmsg(debugstrbuf);
#endif
  return(retval);
}

static my_inline void conn_change_data(tConnection* conn, void* data,
  tBoolean data_is_resource)
{ conn->data = data;
  if (data_is_resource) conn->flags |= cnfDataIsResource;
  else conn->flags &= ~cnfDataIsResource;
}

static my_inline void conn_change_callback(tConnection* conn,
  tConnCbFunc callback, void* data, tBoolean data_is_resource)
{ conn->callback = callback;
  conn_change_data(conn, data, data_is_resource);
}

static tBoolean did_conn_remove; /* basically: "did conn_table_num decrease?"*/
static void conn_remove(tConnection** _conn)
{ tConnection* conn = *_conn;
  tConnectionFlags flags;
  tConnTableIndex cti;
  int fd;

#if NEED_DISSOLVING
  { tConnDissolver dissolver = conn->dissolver;
    if (dissolver != NULL)
    { tBoolean did_dissolve_now = ((dissolver)(conn));
#if CONFIG_DEBUG
      sprint_safe(debugstrbuf, "dissolve(%d,%d,%p,%p)\n", conn->fd,
        did_dissolve_now, conn, dissolver);
      debugmsg(debugstrbuf);
#endif
      if (!did_dissolve_now) { *_conn = NULL; /* detach */ return; }
    }
  }
#endif

  for (cti = 0; cti < conn_table_num; cti++)
  { if (conn == conn_table[cti])
    { if (cti < conn_table_num - 1)
        conn_table[cti] = conn_table[conn_table_num - 1];
      conn_table_num--;
      break;
    }
  }

#if OPTION_TLS
  { tTlsSession session = conn->tls_session;
    if (session != NULL)
    {
#if OPTION_TLS == TLS_GNUTLS
      (void) gnutls_deinit(session);
#elif OPTION_TLS == TLS_OPENSSL
      SSL_free(session);
#endif
    }
  }
#endif

  flags = conn->flags; fd = conn->fd;
  if (flags & cnfReading)
  { FD_CLR(fd, &fds_read); FD_CLR(fd, &fds_r); FD_CLR(fd, &fds_e); }
  if (flags & cnfWriting) { FD_CLR(fd, &fds_write); FD_CLR(fd, &fds_w); }
  my_close(fd); __dealloc(conn->writedata); memory_deallocate(conn);
  *_conn = NULL; did_conn_remove = truE;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "conn_remove(): fd #%d\n", fd);
  debugmsg(debugstrbuf);
#endif
}

#if NEED_DISSOLVING
static my_inline void conn_set_dissolver(tConnection* conn,
  tConnDissolver dissolver)
{ conn->dissolver = dissolver;
}
static my_inline void conn_do_dissolve(tConnection* conn)
{ conn_set_dissolver(conn, NULL); conn_remove(&conn);
}
#endif

#define __conn_set(name, countername, flag, counterflag, which_fds) \
static void name(tConnection* conn, unsigned char mode) \
/* <mode>: 0=clear, 1=set, 2=exclusive-set */ \
{ int fd = conn->fd; \
  if (mode == 0) { conn->flags &= ~flag; FD_CLR(fd, &which_fds); } \
  else \
  { conn->flags |= flag; my_FD_SET(fd, &which_fds); \
    if ( (mode == 2) && (conn->flags & counterflag) ) \
      countername(conn, 0); \
  } \
}

static void conn_set_write(tConnection*, unsigned char); /* prototype */
__conn_set(conn_set_read, conn_set_write, cnfReading, cnfWriting, fds_read)
__conn_set(conn_set_write, conn_set_read, cnfWriting, cnfReading, fds_write)
#undef __conn_set

#if CONFIG_USER_QUERY
static void conn_set_suspend(tConnection* conn, tBoolean do_suspend)
{ tConnectionFlags flags = conn->flags;
  tBoolean is_suspended = cond2boolean(flags & cnfSuspended);
  if (do_suspend != is_suspended) /* must do something */
  { int fd = conn->fd;
    if (do_suspend)
    { if (flags & cnfReading) FD_CLR(fd, &fds_read);
      if (flags & cnfWriting) FD_CLR(fd, &fds_write);
      conn->flags |= cnfSuspended;
    }
    else /* unsuspend */
    { if (flags & cnfReading) my_FD_SET(fd, &fds_read);
      if (flags & cnfWriting) my_FD_SET(fd, &fds_write);
      conn->flags &= ~cnfSuspended;
    }
  }
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "conn_set_suspend(): fd #%d; %d,%d\n", conn->fd,
    do_suspend, is_suspended);
  debugmsg(debugstrbuf);
#endif
}
#endif

static tResourceError conn_err2error(int err)
/* translates error codes; use this when a general connection-related (read,
   write, connect) error occurs */
{ tResourceError re;
  switch (err)
  { case ECONNREFUSED: re = reRefused; break;
    case EADDRNOTAVAIL: re = reNetwork; break; /* CHECKME! */
    case ENETUNREACH: re = reNetwork; break;
    case ETIMEDOUT: re = reTimeout; break;
    case EPIPE: case EAGAIN:
#ifdef ECONNRESET
    case ECONNRESET:
#endif
      re = reServerClosed; break;
    default: re = reConnectionFailureDefault; break;
  }
  return(re);
}


/* TLS */

#if OPTION_TLS

static tBoolean tls_is_usable = falsE;

#define conn_using_tls(conn) ((conn)->flags & cnfTlsDedication)

static tTlsError tls_err2te(int err)
/* converts TLS library error codes to tTlsError codes */
{ tTlsError retval;
  switch (err)
  {
#if OPTION_TLS == TLS_GNUTLS
    case GNUTLS_E_EXPIRED: retval = teSessionExpired; break;
    case GNUTLS_E_FATAL_ALERT_RECEIVED: retval = teFatalAlert; break;
    case GNUTLS_E_FILE_ERROR: retval = teFile; break;
    case GNUTLS_E_GOT_APPLICATION_DATA: retval = teData; break;
    case GNUTLS_E_MEMORY_ERROR: retval = teOom; break;
    case GNUTLS_E_NO_CIPHER_SUITES: case GNUTLS_E_UNKNOWN_CIPHER_SUITE:
      retval = teCipher; break;
    case GNUTLS_E_COMPRESSION_FAILED: case GNUTLS_E_DECOMPRESSION_FAILED:
    case GNUTLS_E_NO_COMPRESSION_ALGORITHMS:
    case GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM:
      retval = teCompression; break;
    case GNUTLS_E_DECRYPTION_FAILED: case GNUTLS_E_ENCRYPTION_FAILED:
      retval = teCrypt; break;
    case GNUTLS_E_ILLEGAL_SRP_USERNAME: case GNUTLS_E_SRP_PWD_ERROR:
    case GNUTLS_E_SRP_PWD_PARSING_ERROR:
      retval = teSrp; break;
    case GNUTLS_E_PKCS1_WRONG_PAD: case GNUTLS_E_PK_DECRYPTION_FAILED:
    case GNUTLS_E_PK_ENCRYPTION_FAILED: case GNUTLS_E_PK_SIGN_FAILED:
    case GNUTLS_E_PK_SIG_VERIFY_FAILED: case GNUTLS_E_UNKNOWN_PK_ALGORITHM:
      retval = tePk; break;
    case GNUTLS_E_LARGE_PACKET: case GNUTLS_E_RECORD_LIMIT_REACHED:
    case GNUTLS_E_TOO_MANY_EMPTY_PACKETS: case GNUTLS_E_UNEXPECTED_PACKET:
    case GNUTLS_E_UNEXPECTED_PACKET_LENGTH:
    case GNUTLS_E_UNSUPPORTED_VERSION_PACKET:
      retval = tePacket; break;
    case GNUTLS_E_UNIMPLEMENTED_FEATURE: retval = teUnimplemented; break;
    case GNUTLS_E_ERROR_IN_FINISHED_PACKET:
    case GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET:
      retval = teHandshake; break;
#elif OPTION_TLS == TLS_OPENSSL
    /* IMPLEMENTME? */
#endif
    default: retval = teUnknown; break;
  }
  return(retval);
}

static void tls_set_error(tConnection* conn, tTlsError te)
{ tResource* resource = conn2resource(conn);
  if (resource != NULL)
  { resource_set_error(resource, reTls);
    if (resource->tls_error == teFine) resource->tls_error = te;
  }
}

#if OPTION_TLS == TLS_GNUTLS

#define __tls_gnutls_intr(err) \
  ( ((err) == GNUTLS_E_INTERRUPTED) || ((err) == GNUTLS_E_AGAIN) )
#define __tls_gnutls_warn(err) ((err) == GNUTLS_E_WARNING_ALERT_RECEIVED)

static gnutls_certificate_client_credentials tls_xcred;

#if CONFIG_DEBUG

static void tls_gnutls_debug(const char* operation, int err)
{ tBoolean is_error = cond2boolean(err < 0);
  const char* errstr;
  char* spfbuf;
  if (!is_error) errstr = strEmpty;
  else { errstr = gnutls_strerror(err); if (errstr == NULL) errstr=strEmpty; }

  my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf, "TLS%d: %s (%d%s%s)\n", OPTION_TLS,
    operation, err, ( (errstr[0] != '\0') ? ", " : strEmpty ), errstr);
  debugmsg(spfbuf);
  my_spf_cleanup(debugstrbuf, spfbuf);
  /* Don't take the identifier "err" too seriously. The great GnuTLS
     documentation often does not say anything about the meaning of the return
     value of a specific function, so it might not actually be related to error
     handling... :-( */
}

#define tls_gnutls_debug_errvalue tls_gnutls_debug

#else /* #if CONFIG_DEBUG */

#define tls_gnutls_debug(operation, err) do { } while (0)

static void tls_gnutls_debug_errvalue(const char* operation, int err)
/* Call this function only if it is _known_ (e.g. from GnuTLS documentation)
   that <err> really means an error indicator. */
{ if ( (err < 0) && (!__tls_gnutls_intr(err)) && (!__tls_gnutls_warn(err)) )
  { tTlsError te = tls_err2te(err);
    tBoolean show_te = cond2boolean(is_tls_error_expressive(te));
    const char* errstr = gnutls_strerror(err);
    char* spfbuf;
    my_spf(strbuf, STRBUF_SIZE, &spfbuf,
      _("retawq: TLS library error during operation \"%s\": #%d%s%s - %s\n"),
      operation, err, (show_te ? strSpacedDash : strEmpty), (show_te ?
      _(strTlsError[te]) : strEmpty), null2empty(errstr));
    if (may_use_fd2()) my_write_str(2, spfbuf);
    debugmsg(spfbuf);
    my_spf_cleanup(strbuf, spfbuf);
  }
}

#endif /* #if CONFIG_DEBUG */

static void tls_gnutls_check_direction(tConnection* conn)
{ if (gnutls_record_get_direction(conn->tls_session) == 1)
    conn_set_write(conn, 2);
  else conn_set_read(conn, 2);
}

#elif OPTION_TLS == TLS_OPENSSL

static SSL_CTX* tls_context = NULL;

#if CONFIG_DEBUG
static int tls_openssl_print_errtext(const char* str, size_t len,
  __sunused void* data __cunused)
{ return(my_write(debugfd, str, len));
}
static void tls_openssl_debug_errstuff(const char* funcname,
  const tConnection* conn, int err, int sslge)
{ char* spfbuf;
  my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf,
    "TLS%d: SSL_%s(): fd=%d, err=%d, sslge=%d, errno=%d%s%s\n", OPTION_TLS,
    funcname, conn->fd, err, sslge, errno, ( (errno != 0) ? strSpacedDash :
    strEmpty ), ( (errno != 0) ? my_strerror(errno) : strEmpty ));
  debugmsg(spfbuf);
  my_spf_cleanup(debugstrbuf, spfbuf);
  ERR_print_errors_cb(tls_openssl_print_errtext, NULL);
    /* (undocumented, but "hopefully not vanishing" - it is OpenSSL's sole
        non-ridiculous error printing interface for file descriptors (like
        debugfd)) */
}
#else
#define tls_openssl_debug_errstuff(funcname, conn, err, sslge) do { } while (0)
#endif

static void tls_openssl_handle_error(tConnection* conn, int sslge)
/* Just one more OpenSSL silliness... */
{ if (sslge == SSL_ERROR_SYSCALL)
  { tResource* resource = conn2resource(conn);
    if (resource != NULL)
    { tResourceError re = conn_err2error(errno);
      if (re != reConnectionFailureDefault)
      { resource_set_error(resource, re); return; }
    }
  }
  tls_set_error(conn, teUnknown);
}

#define tls_openssl_operate(operation, funcname) \
  do \
  { ERR_clear_error(); err = operation; sslge = SSL_get_error(session, err); \
    tls_openssl_debug_errstuff(funcname, conn, err, sslge); \
  } while (0)

static tBoolean tls_openssl_shutdown(tConnection* conn)
/* returns whether it "finished somehow" */
{ tBoolean retval;
  tTlsSession session = conn->tls_session;
  int err, sslge;
  tls_openssl_operate(SSL_shutdown(session), "shutdown");
  if ( (sslge == SSL_ERROR_SYSCALL) && (err == 0) )
  { /* Just one more OpenSSL silliness. SSL_shutdown() documentation says that
       sslge can be wrong if SSL_shutdown() returned 0: "The output of
       SSL_get_error(3) may be misleading, as an erroneous SSL_ERROR_SYSCALL
       may be flagged even though no error occurred." */
    /* An additional silliness: SSL_want() documentation says: "Error
       conditions are not handled and must be treated using SSL_get_error(3)."
       But since SSL_get_error() "may be misleading" - what to do instead?! */
    switch (SSL_want(session))
    { case SSL_READING: sslge = SSL_ERROR_WANT_READ; break;
      case SSL_WRITING: sslge = SSL_ERROR_WANT_WRITE; break;
    }
  }
  switch (sslge)
  { case SSL_ERROR_NONE: retval = truE; break;
    case SSL_ERROR_WANT_READ: conn_set_read(conn, 2); retval = falsE; break;
    case SSL_ERROR_WANT_WRITE: conn_set_write(conn, 2); retval = falsE; break;
    default: tls_openssl_handle_error(conn, sslge); retval = truE; break;
  }
  return(retval);
}

#endif

static one_caller tBoolean tls_initialize(void)
{ static tBoolean tls_did_init = falsE;
  if (tls_did_init) goto out;
  tls_did_init = truE;

#if OPTION_TLS == TLS_GNUTLS

  { char* spfbuf;
    int err;
    err = gnutls_global_init();
    tls_gnutls_debug_errvalue(_(strInitialization), err);
    if (err != 0) goto out;
    err = gnutls_certificate_allocate_credentials(&tls_xcred);
    tls_gnutls_debug_errvalue("allocate credentials", err);
    my_spf(strbuf, STRBUF_SIZE, &spfbuf, strPercsPercs, config.path,
      "gnutls-ca.pem");
    err = gnutls_certificate_set_x509_trust_file(tls_xcred, spfbuf,
      GNUTLS_X509_FMT_PEM);
    tls_gnutls_debug("trustfile", err);
    my_spf_cleanup(strbuf, spfbuf);
    tls_is_usable = truE;
  }

#elif OPTION_TLS == TLS_OPENSSL

  { SSL_METHOD* m;
    char seedfilenamebuf[512];
    const char* seedfilename;
    time_t seedval;
    unsigned short randcount;
    SSL_load_error_strings();  /* Do something before... */
    (void) SSL_library_init(); /* ...initializing; what a sane concept! :-( */

    /* Handle this random rubbish which a sane library would do itself... */
    debugmsg("TLS: random A\n");
    if (RAND_status()) goto rand_done; /* nothing to do, nice */
    debugmsg("TLS: random B\n");
    seedfilename = RAND_file_name(seedfilenamebuf, sizeof(seedfilenamebuf));
    if (seedfilename != NULL)
    { debugmsg("TLS: random C\n");
#if CONFIG_DEBUG
      { char* spfbuf;
        my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf, "TLS%d: seedfilename=%s\n",
          OPTION_TLS, seedfilename);
        debugmsg(spfbuf);
        my_spf_cleanup(debugstrbuf, spfbuf);
      }
#endif
      if (RAND_egd(seedfilename) >= 0) goto rand_finish;
      debugmsg("TLS: random D\n");
      (void) RAND_load_file(seedfilename, -1);
        /* ("-1" is allowed for OpenSSL >= 0.9.5) */
      if (RAND_status()) goto rand_finish;
    }
    debugmsg("TLS: random E\n");
    seedval = my_time() ^ ((~getpid()) * (~getppid())); /* good enough? :-) */
    RAND_seed((const void*) (&seedval), sizeof(seedval));
    if (RAND_status()) goto rand_finish;
    debugmsg("TLS: random F\n");
    srand48(seedval);
    randcount = 10000; /* wild guess, thanks to lack of useful documentation,
      just to avoid infinite loops */
    while (randcount-- > 0)
    { long l = lrand48();
      RAND_seed((const void*) (&l), sizeof(l));
      if (RAND_status()) goto rand_counted;
    }
    debugmsg("TLS: random G\n");
    /* CHECKME: what to do here? Set reRandom? Just hope the best?? */
    rand_counted: {}
#if CONFIG_DEBUG
    sprint_safe(debugstrbuf, "TLS%d: remaining randcount: %d\n", OPTION_TLS,
      randcount);
    debugmsg(debugstrbuf);
#endif
    rand_finish:
    debugmsg("TLS: random H\n");
    if (seedfilename != NULL) RAND_write_file(seedfilename);
    rand_done:
    debugmsg("TLS: random Z\n");

    m = SSLv23_client_method();
    if ( (m == NULL) || ( (tls_context = SSL_CTX_new(m)) == NULL ) )
    { if (may_use_fd2())
        my_write_str(2, _("retawq: OpenSSL initialization failed\n"));
      goto out;
    }
    (void) SSL_CTX_set_options(tls_context, SSL_OP_ALL);
      /* (OpenSSL documentation says this is safe; stupid-me doubts it somehow,
          esp. for SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS, but...) */
    (void) SSL_CTX_set_mode(tls_context, SSL_MODE_ENABLE_PARTIAL_WRITE);
    tls_is_usable = truE;
  }

#endif

  out:
  return(tls_is_usable);
}

static one_caller void tls_deinitialize(void)
{
#if OPTION_TLS == TLS_GNUTLS
  if (tls_is_usable) gnutls_global_deinit();
#elif OPTION_TLS == TLS_OPENSSL
  if (tls_context != NULL) SSL_CTX_free(tls_context);
#endif
}

static unsigned char tls_do_handshaking(tConnection* conn)
{ unsigned char retval; /* 0=error, 1=proceeding, 2=done */
  tTlsSession session = conn->tls_session;

#if OPTION_TLS == TLS_GNUTLS

  int err = gnutls_handshake(session);
  tls_gnutls_debug_errvalue("handshake", err);
  if ( (__tls_gnutls_intr(err)) || (__tls_gnutls_warn(err)) )
  { tls_gnutls_check_direction(conn); retval = 1; }
  else if (err < 0) { tls_set_error(conn, tls_err2te(err)); retval = 0; }
  else retval = 2;

#elif OPTION_TLS == TLS_OPENSSL

  int err, sslge;
  tls_openssl_operate(SSL_connect(session), "connect");
  switch (sslge)
  { case SSL_ERROR_NONE: retval = 2; break;
    case SSL_ERROR_WANT_READ: conn_set_read(conn, 2); retval = 1; break;
    case SSL_ERROR_WANT_WRITE: conn_set_write(conn, 2); retval = 1; break;
    default: tls_openssl_handle_error(conn, sslge); retval = 0; break;
  }

#endif

  if (retval == 1)
  {
#if CONFIG_DEBUG
    if (!(conn->flags & cnfTlsHandshaking))
    { sprint_safe(debugstrbuf, "TLS%d: starting handshake, fd #%d\n",
        OPTION_TLS, conn->fd);
      debugmsg(debugstrbuf);
    }
#endif
    conn->flags |= cnfTlsHandshaking;
  }
  else
  { conn->flags &= ~cnfTlsHandshaking;
    if (retval == 2)
    { if (conn->flags & cnfWantToWrite) conn_set_write(conn, 2);
      else conn_set_read(conn, 2);
    }
  }
  return(retval);
}

static void tls_session_dissolver_callback(tConnection* conn,
  tConnCbEventKind ccek)
{
#if OPTION_TLS == TLS_GNUTLS
  tTlsSession session = conn->tls_session;
  int err;
#endif

  switch (ccek)
  {
#if OPTION_TLS == TLS_GNUTLS
    case ccekRead:
      err = gnutls_record_recv(session, NULL, 0);
      handle_err:
#if CONFIG_DEBUG
      sprint_safe(debugstrbuf, "TLS%d: dis - %d,%d,%d\n", OPTION_TLS, conn->fd,
        ccek, err);
      debugmsg(debugstrbuf);
#endif
      if (__tls_gnutls_intr(err) || __tls_gnutls_warn(err))
        tls_gnutls_check_direction(conn);
      else goto do_dissolve; /* error or done */
      break;
    case ccekWrite:
      err = gnutls_record_send(session, NULL, 0); goto handle_err;
      /*@notreached@*/ break;
#elif OPTION_TLS == TLS_OPENSSL
    case ccekRead: case ccekWrite:
      if (tls_openssl_shutdown(conn)) goto do_dissolve;
      break;
#endif
    default: conn_bug(conn, ccek); do_dissolve: conn_do_dissolve(conn); break;
  }
}

static tBoolean tls_session_dissolver(tConnection* conn)
/* returns whether it actually dissolved the session right now */
{ tBoolean retval;
#if OPTION_TLS == TLS_GNUTLS
  int err;
  conn->flags &= ~cnfDataIsResource;
  err = gnutls_alert_send(conn->tls_session, GNUTLS_AL_WARNING,
    GNUTLS_A_CLOSE_NOTIFY);
  tls_gnutls_debug("alert_send", err);
  if ( (__tls_gnutls_intr(err)) || (__tls_gnutls_warn(err)) )
  { tls_gnutls_check_direction(conn); retval = falsE; }
  else retval = truE; /* done (unlikely since non-blocking) or error */
#elif OPTION_TLS == TLS_OPENSSL
  conn->flags &= ~cnfDataIsResource;
  retval = tls_openssl_shutdown(conn);
#endif
  if (!retval)
  { conn_set_dissolver(conn, NULL); conn->flags |= cnfDissolving;
    conn_change_callback(conn, tls_session_dissolver_callback, NULL, falsE);
  }
  return(retval);
}

static tBoolean tls_session_init(tConnection* conn, const char* hostname)
{ tBoolean retval = falsE;
  tTlsSession session;
  tTlsError te = teFine;
#if OPTION_TLS == TLS_GNUTLS
  static const int cert_type_priority[] =
  { GNUTLS_CRT_X509, GNUTLS_CRT_OPENPGP, 0 };
  int err;
#endif

  if (!tls_initialize())
  {
#if OPTION_TLS == TLS_OPENSSL
    init_failed:
#endif
    te = teInit; goto out;
  }

#if OPTION_TLS == TLS_GNUTLS

  err = gnutls_init(&session, GNUTLS_CLIENT);
  conn->tls_session = session;
  tls_gnutls_debug_errvalue("initialize session", err);
  if (err != 0) { te = tls_err2te(err); goto out; }
  err = gnutls_set_default_priority(session);
  tls_gnutls_debug("set default priority", err);
  err = gnutls_certificate_type_set_priority(session, cert_type_priority);
  tls_gnutls_debug("set certificate type priority", err);
  err = gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, tls_xcred);
  tls_gnutls_debug("set credentials", err);
  gnutls_transport_set_ptr(session, (gnutls_transport_ptr) conn->fd);
  if (hostname != NULL)
  { err = gnutls_server_name_set(session, GNUTLS_NAME_DNS, hostname,
      strlen(hostname));
    tls_gnutls_debug("nameset", err);
  }
  /* FIXME: RFC2818, 3.1! */

#elif OPTION_TLS == TLS_OPENSSL

  session = conn->tls_session = SSL_new(tls_context);
  if (session == NULL) goto init_failed;
  (void) SSL_set_fd(session, conn->fd);

#endif

  /* try to start the handshaking */
  if (tls_do_handshaking(conn) == 0) goto out;

  /* 't worked */
  conn->flags |= cnfTlsDedication;
  conn_set_dissolver(conn, tls_session_dissolver);
  retval = truE;

  out:
  if (te != teFine) tls_set_error(conn, te);
  return(retval);
}

static ssize_t tls_record_read(tConnection* conn, /*@out@*/ void* buf,
  size_t size, /*@out@*/ tBoolean* _was_interrupted)
{ ssize_t retval;
  tTlsSession session = conn->tls_session;

#if OPTION_TLS == TLS_GNUTLS

  *_was_interrupted = falsE;
  retval = gnutls_record_recv(session, buf, size);
  tls_gnutls_debug_errvalue(_("read"), retval);
  if (__tls_gnutls_intr(retval) || __tls_gnutls_warn(retval))
  { tls_gnutls_check_direction(conn);
    zeroy: retval = 0; *_was_interrupted = truE;
  }
  else if (retval == GNUTLS_E_REHANDSHAKE)
  { if (tls_do_handshaking(conn) != 0) goto zeroy;
  }
  else if (retval < 0) tls_set_error(conn, tls_err2te(retval));

#elif OPTION_TLS == TLS_OPENSSL

  int err, sslge;
  *_was_interrupted = falsE;
  tls_openssl_operate(SSL_read(session, buf, size), "read");
  switch (sslge)
  { case SSL_ERROR_NONE: retval = err; break;
    case SSL_ERROR_WANT_READ:
      conn_set_read(conn, 2);
      zeroy: retval = 0; *_was_interrupted = truE;
      break;
    case SSL_ERROR_WANT_WRITE: conn_set_write(conn, 2); goto zeroy; break;
    default: tls_openssl_handle_error(conn, sslge); retval = 0; break;
  }

#endif

  return(retval);
}

static ssize_t tls_record_write(tConnection* conn, const void* buf,
  size_t size, /*@out@*/ tBoolean* _was_interrupted)
{ ssize_t retval;
  tTlsSession session = conn->tls_session;

#if OPTION_TLS == TLS_GNUTLS

  retval = gnutls_record_send(session, buf, size);
  tls_gnutls_debug_errvalue(_("write"), retval);
  if (__tls_gnutls_intr(retval) || __tls_gnutls_warn(retval))
  { tls_gnutls_check_direction(conn); *_was_interrupted = truE; retval = 0; }
  else
  { *_was_interrupted = falsE;
    if (retval < 0) tls_set_error(conn, tls_err2te(retval));
  }

#elif OPTION_TLS == TLS_OPENSSL

  int err, sslge;
  tls_openssl_operate(SSL_write(session, buf, size), "write");
  *_was_interrupted = falsE;
  switch (sslge)
  { case SSL_ERROR_NONE: retval = err; break;
    case SSL_ERROR_WANT_READ:
      conn_set_read(conn, 2);
      zeroy: *_was_interrupted = truE; retval = 0;
      break;
    case SSL_ERROR_WANT_WRITE: conn_set_write(conn, 2); goto zeroy; break;
    default: tls_openssl_handle_error(conn, sslge); retval = 0; break;
  }

#endif

  return(retval);
}

#define tls_record_write_str(conn, str, wi) \
  tls_record_write(conn, str, strlen(str), wi)

static one_caller tBoolean tls_version_is_old(const char* version)
/* returns whether the TLS/libgcrypt library version is known to be "old" */
{ static const int minpart[3] = /* "most recent" version number */
#if OPTION_TLS == TLS_GNUTLS
    { 1, 0, 8 }; /* GnuTLS */
#elif OPTION_TLS == TLS_OPENSSL
    { 0, 9, 7 }; /* OpenSSL */
#define minpart_ch ('c')
#elif OPTION_TLS == TLS_BUILTIN
    { 1, 1, 92 }; /* libgcrypt */
#endif
  tBoolean retval = falsE;
  const char* temp = version;
  unsigned char count;
  int part[3];

  /* try to parse the version string; since the format of this string might
     change in the future etc., let's be careful */
  for (count = 0; count <= 2; count++)
  { if (!my_isdigit(*temp)) goto out;
    my_atoi(temp, &(part[count]), &temp, 999);
    if (count < 2)
    { if (*temp != '.') goto out;
      else temp++;
    }
  }

  /* understood the version string, now check the version */
  for (count = 0; count <= 2; count++)
  { int p = part[count], m = minpart[count];
    if (p != m)
    { if (p < m) retval = truE;
      goto out;
    }
  }
#if OPTION_TLS == TLS_OPENSSL
  { char ch = *temp;
    if ( (my_islower(ch)) && (*(temp + 1) == ' ') && (ch < minpart_ch) )
      retval = truE;
  }
#undef minpart_ch
#endif

  out:
  return(retval);
}

static one_caller const char* tls_version_warning(const char* version)
/* returns an HTML warning text if the TLS/libgcrypt library version is known
   to be "old" */
{ const char* retval = (tls_version_is_old(version) ?
    _(" <b>Warning</b>: this version is out of date!") : strEmpty);
  return(retval);
}

#endif /* #if OPTION_TLS */


/* Communication with main.c */

static my_inline void __push_to_main(void* data, tMainCallbackKind mck)
{ main_callback(data, mck);
}

static void push_to_main_res(tResource* resource, unsigned char flags)
{ if (!(flags & 1)) resource->flags |= rfFinal;
  __push_to_main((void*) resource, mckResource);
}

static void push_to_main_req(tResourceRequest* request, unsigned char flags)
{ if (!(flags & 1))
  { tResource* resource = request->resource;
    if (resource != NULL) resource->flags |= rfFinal;
  }
  __push_to_main((void*) request, mckRequest);
}

#if CONFIG_USER_QUERY

static void resource_suspend(const tResource* resource)
{ tConnection* conn = resource->cconn;
  if (conn != NULL) conn_set_suspend(conn, truE);
}

static void resource_unsuspend(const tResource* resource)
{ tConnection* conn = resource->cconn;
  if (conn != NULL) conn_set_suspend(conn, falsE);
}

static my_inline void user_query_deallocate(const tUserQuery* query)
{ memory_deallocate(query);
}

static tBoolean resource_ask_anything(tResource* resource,
  tUserQueryCallback callback, tMissingInformationFlags mif,
  unsigned char flags)
/* flags: "&1": prior login attempt failed */
{ tBoolean retval = falsE;
  tHostPortProtInfo* hppi;
  if (!is_promptable) goto out; /* no way of prompting a user */
  if (flags & 1) mif |= mifPriorLoginAttemptFailed;
  hppi = ( (mif & mifProxyrelated) ? resource->actual_hppi :
    resource->textual_hppi );
  if (hppi->flags & hppifUserQueryRunning) /* already a query running */
  { goto out; /* IMPLEMENTME: "wait"! */
  }
  else /* start a query */
  { tUserQuery* query = (tUserQuery*) memory_allocate(sizeof(tUserQuery));
    query->callback = callback; query->resource = resource; query->hppi = hppi;
    query->id = resource->id; query->mif = mif;
    query->hostname = hppi->host->hostname;
    query->portnumber = hppi->portnumber;
    main_callback(query, mckUserQuery);
    if (query->mif & mifQueryFailed) /* main.c knows in advance it won't go */
    { user_query_deallocate(query); goto out; }
    resource_suspend(resource);
    hppi->flags |= hppifUserQueryRunning;
    retval = truE;
  }
  out:
  return(retval);
}

static void resource_ask_finish(tUserQuery* query)
{ tHostPortProtInfo* hppi = query->hppi;
  /* IMPLEMENTME: store login information in hppi! */
  hppi->flags &= ~hppifUserQueryRunning;
  /* IMPLEMENTME: kick any resources which were "waiting" on this hppi! */
  user_query_deallocate(query);
}

#endif /* #if CONFIG_USER_QUERY */


/* Helper functions II */

tPortnumber rp2portnumber(tResourceProtocol rp)
/* returns the default portnumber for the given protocol */
{ tPortnumber retval;
  switch (rp)
  { case rpHttp: retval = portnumber_http; break;
    case __rpFtp: case __rpFtps: retval = portnumber_ftp; break;
    case __rpFinger: retval = portnumber_finger; break;
    case rpCvs: retval = portnumber_cvs; break;
    case rpGopher: retval = portnumber_gopher; break;
    case __rpNntp: retval = portnumber_nntp; break;
    case __rpPop: case __rpPops: retval = portnumber_pop3; break;
    case __rpHttps: retval = portnumber_https; break;
    default: retval = 0; break; /* might happen (for unknown schemes) */
  }
  return(retval);
}

void rp2scheme(tResourceProtocol rp, char* dest)
/* intended for user interface texts */
{ const char* str = NULL;
  tBoolean use_uc = cond2boolean(__has_rpflag(rp, rpfUiToupper)),
    use_alt = cond2boolean(__has_rpflag(rp, rpfUiAlternate));
  if (use_alt)
  { if (rp == __rpNntp) str = strNntp;
    else if (rp == __rpPop) str = strPop3;
  }
  if (str == NULL) str = rp_data[rp].scheme;
  if (use_uc) my_strcpy_toupper(dest, str);
  else strcpy(dest, str);
}

#if CONFIG_BLOAT & 1
const char* ssc2info(tResourceProtocol rp, tServerStatusCode ssc)
/* tries to convert a server status code to a nice, human-readable string; this
   produces mostly very general texts right now - IMPLEMENTME further! */
{ const char* retval = strEmpty; /* default */
  static const char strIntermediary[] = N_("intermediary"),
    strServerError[] = N_("server error"), strOkay[] = N_("okay");
#if OPTION_NEWS || CONFIG_FTP
  static const char strServerReady[] = N_("server ready");
#endif
#if CONFIG_FTP
  static const char strAccountRequired[] = N_("account required");
#endif

  if ( (is_httplike(rp))
#if OPTION_LOCAL_CGI
       || (rp == rpLocalCgi)
#endif
     )
  { switch (ssc / 100)
    { case 1: retval = _(strIntermediary); break;
      case 2: retval = _(strOkay); break;
      case 3: retval = _(strRedirection); break;
      case 4: retval = _("request error"); break;
      case 5: retval = _(strServerError); break;
    }
  }
#if CONFIG_FTP
  else if (is_ftplike(rp))
  { switch (ssc / 100)
    { case 1: retval = _(strIntermediary); break;
      case 2:
        if (ssc == 200) retval = _(strOkay);
        else if (ssc == 220) retval = _(strServerReady);
        else if (ssc == 221) retval = _("server closed");
        else if (ssc == 226) retval = _("transfer successful");
        else if ( (ssc == 227) || (ssc == 229) )
          retval = _("entering passive mode");
        else if ( (ssc == 230)
#if OPTION_TLS
                  || ( (rp == rpFtps) && (ssc == 232) )
#endif
                )
        { retval = _("logged in"); }
        break;
      case 3:
        if (ssc == 331) retval = _("password required");
        else if (ssc == 332) retval = _(strAccountRequired);
        else if (ssc == 350) retval = _(strIntermediary);
        break;
      case 4:
        if (ssc == 421) retval = _("service not available");
        else retval = _("transient server problem");
        break;
      case 5:
        if (ssc == 530) retval = _(strResourceError[reLogin]);
        else if (ssc == 532) retval = _(strAccountRequired);
        else if (ssc == 550) retval = _("file/directory not found");
        else retval = _(strServerError);
        break;
      case 6: retval = _("protected reply"); break;
    }
  }
#endif
#if OPTION_NEWS
  else if (rp == rpNntp)
  { switch (ssc / 100)
    { case 2:
        if ( (ssc == 200) || (ssc == 201) )
          retval = _(strServerReady);
        else if (ssc == 211) retval = _("group selected");
        else if ( (ssc == 215) || (ssc == 231) )
          retval = _("news groups follow");
        else if ( (ssc >= 220) && (ssc <= 222) ) retval = _("content follows");
        break;
      case 4:
        if (ssc == 411) retval = _("no such news group");
        else if ( ( (ssc >= 421) && (ssc <= 423) ) || (ssc == 430) )
          retval = _("no such article");
        else if (ssc == 480) retval = _("authentication required");
        break;
    }
  }
#endif
  return(retval);
}
#endif /* #if CONFIG_BLOAT & 1 */

static tBoolean is_filesuffix(const char* filename, const char* suff)
/* returns whether the <filename> has the suffix <suff>; the latter must be
   given in lower-case! */
{ size_t filenamelen = strlen(filename), sufflen = strlen(suff);
  const char* temp;
  if (filenamelen < sufflen + 1 /* dot */ + 1 /* at least one other char */)
    return(falsE);
  temp = filename + filenamelen - sufflen;
  if ( (*(temp - 1)) != '.' ) return(falsE);
  if (!streqcase(temp, suff)) return(falsE);
  return(truE);
}

static tResourceContentKind filesuffix2rck(const char* name)
/* calculates a resource content kind based on the suffix of <name> */
{ if ( (is_filesuffix(name, strHtml)) || (is_filesuffix(name, "htm")) ||
       (is_filesuffix(name, strShtml)) || (is_filesuffix(name, "phtml")) )
    return(rckHtml);
  return(rckUnknown);
}

static tBoolean calc_parentpath(const char* path, char* staticbuf, size_t size)
{ tBoolean retval = falsE;
  size_t pathlen = strlen(path);
  if (pathlen > 1)
  { const char* p = path + pathlen - 1;
    if (*p == '/') p--;
    while (p >= path)
    { if (*p == '/')
      { size_t len = p - path + 1;
        if (len + 3 < size)
        { my_memcpy(staticbuf, path, len); staticbuf[len] = '\0';
          retval = truE;
        }
        break;
      }
      p--;
    }
  }
  return(retval);
}

static const char* htmlify(const char* src)
/* converts special HTML characters in <src> */
{ enum { convnum = 3 };
  static const struct
  { const char* replacement;
    unsigned short len; /* == strlen(replacement) */
    char ch;
  } conv[convnum] =
  { /* { "&quot;", 6, '"'  }, */ /* #34 */
       { "&amp;",  5, '&'  },    /* #38 */
    /* { "&#39;",  5, '\'' }, */ /* #39 */
       { "&lt;",   4, '<'  },    /* #60 */
       { "&gt;",   4, '>'  }     /* #62 */
  };
  size_t destlen;
  const char* srctmp;
  char ch, *retval, *desttmp;
  unsigned short count;
  tBoolean need_conversion = falsE;

  /* Calculate the length of the resulting string: */
  destlen = 1;
  srctmp = src;
  while ( (ch = *srctmp++) != '\0' )
  { for (count = 0; count < convnum; count++)
    { if (ch == conv[count].ch)
      { destlen += conv[count].len;
        need_conversion = truE;
        goto next_char_1;
      }
    }
    destlen++;
    next_char_1: {}
  }
  if (!need_conversion) return(src); /* the most likely case */

  /* Build the resulting string: */
  retval = desttmp = __memory_allocate(destlen);
  srctmp = src;
  while ( (ch = *srctmp++) != '\0' )
  { for (count = 0; count < convnum; count++)
    { if (ch == conv[count].ch)
      { const char* tmp = conv[count].replacement;
        while ( (ch = *tmp++) != '\0' ) *desttmp++ = ch;
        goto next_char_2;
      }
    }
    *desttmp++ = ch; /* no replacement */
    next_char_2: {}
  }
  *desttmp = '\0';
  return(retval);
}

#define htmlify_cleanup(orig, htmlified) \
  do { if (htmlified != orig) memory_deallocate(htmlified); } while (0)


/* Connections II */

#if OPTION_TLS

static ssize_t __conn_read(tConnection* conn, /*@out@*/ void* buf,
  size_t bufsize, /*@out@*/ tBoolean* _was_interrupted)
{ ssize_t retval;
  if (conn_using_tls(conn))
  { retval = tls_record_read(conn, buf, bufsize, _was_interrupted);
    if (retval < 0) tls_set_error(conn, tls_err2te(retval));
  }
  else { retval = my_read(conn->fd, buf, bufsize); *_was_interrupted = falsE; }
  return(retval);
}

#define conn_read(resultvar, conn, buf, bufsize, action_if_interrupted) \
  do \
  { tBoolean was_interrupted; \
    resultvar = __conn_read(conn, buf, bufsize, &was_interrupted); \
    if (was_interrupted) { action_if_interrupted; } \
  } while (0)

#else

#define conn_read(resultvar, conn, buf, bufsize, action_if_interrupted) \
  resultvar = my_read(conn->fd, buf, bufsize)

#endif

static void conn_set_writedata(tConnection* conn, const void* data,
  size_t size)
/* sets a new message to be sent over the network for this connection */
{ __dealloc(conn->writedata); conn->writedata = data;
  conn->writedata_todo = size; conn->writedata_done = 0;
  conn->flags |= cnfWantToWrite;
#if CONFIG_DEBUG
  conn->flags &= ~cnfDebugNodata;
#endif
}

static my_inline void conn_set_writestr(tConnection* conn, const char* str)
{ conn_set_writedata(conn, str, strlen(str));
}

static unsigned char conn_write_writedata(tConnection* conn)
/* tries to write as much of the current message as possible; return value:
   0=error, 1=proceeding, 2=done */
{ const char *data = conn->writedata, *src;
  size_t todo, done, writelen;
  ssize_t write_err;
  int fd;

  if (data == NULL) /* "should not happen" */
  { failed: conn->flags &= ~cnfWantToWrite; return(0); }
  todo = conn->writedata_todo; done = conn->writedata_done;

  src = data + done; writelen = todo - done; fd = conn->fd;
#if OPTION_TLS
  if (conn_using_tls(conn))
  { tBoolean was_interrupted;
    write_err = tls_record_write(conn, src, writelen, &was_interrupted);
    if (was_interrupted) return(1);
    if (write_err < 0) tls_set_error(conn, tls_err2te(write_err));
  }
  else
#endif
  { write_err = __my_write(fd, src, writelen); }

  if (write_err <= 0) goto failed;
  done += write_err; conn->writedata_done = done;
#if CONFIG_DEBUG
  { sprint_safe(debugstrbuf, "cww: %d wrote %d/%d:\n", fd, done, todo);
    debugmsg(debugstrbuf);
    if (!(conn->flags & cnfDebugNodata))
    { __debugmsg(src, write_err);
      if (src[write_err - 1] != '\n') debugmsg(strNewline);
    }
  }
#endif
  if (done >= todo)
  { conn->flags &= ~cnfWantToWrite; dealloc(conn->writedata); return(2); }
  else return(1);
}

#if NEED_QUITCMD_DISSOLVING

static void quitcmd_dissolver_callback(tConnection* conn,
  tConnCbEventKind ccek)
/* I like spaghetti. Hey, this function is almost readable... :-) */
{ switch (ccek)
  {
#if OPTION_TLS
    case ccekRead:
      if (!conn_using_tls(conn)) goto handle_bug;
      /* "else": */ /*@fallthrough@*/
#endif
    case ccekWrite:
      switch (conn_write_writedata(conn))
      { case 0: do_dissolve: conn_do_dissolve(conn); break;
        case 1: break;
        case 2:
#if OPTION_TLS
          if (conn_using_tls(conn))
          { /* We handled the "high-level" quitcmd, now we have to handle the
               "low-level" TLS close_notify alert. */
            conn_set_dissolver(conn, tls_session_dissolver);
            conn_remove(&conn);
          }
          else
#endif
          { goto do_dissolve; /*@notreached@*/ }
          break;
      }
      break;
    default:
#if OPTION_TLS
      handle_bug:
#endif
      conn_bug(conn, ccek); goto do_dissolve; /*@notreached@*/ break;
  }
}

static tBoolean quitcmd_dissolver(tConnection* conn)
/* returns whether it actually dissolved the session right now */
{ tBoolean retval;
  if ( (conn->writedata_done <= 0) || (!(conn->flags & cnfWantToWrite)) )
  { /* The dissolving attempt doesn't happen in the middle of a write (which
       could happen e.g. if we dissolve due to a write() call failure); so we
       can safely try to send a nice "quit" command to the server. */
    static const char strNetcmdQuit[] = "QUIT\r\n";
    conn_set_dissolver(conn, NULL); conn->flags |= cnfDissolving;
    conn_change_callback(conn, quitcmd_dissolver_callback, NULL, falsE);
    conn_set_writestr(conn, my_strdup(strNetcmdQuit));
    conn_set_write(conn, 2);
    retval = falsE;
  }
#if OPTION_TLS
  else if (conn_using_tls(conn))
  { /* We can't send the "high-level" quitcmd, but maybe we can at least send
       the "low-level" TLS close_notify alert. */
    retval = tls_session_dissolver(conn);
  }
#endif
  else retval = truE; /* can't do anything */
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "quitcmd_dissolver(): %d,%d\n", conn->fd, retval);
  debugmsg(debugstrbuf);
#endif
  return(retval);
}

#endif /* #if NEED_QUITCMD_DISSOLVING */


/* Content blocks */

static __sallocator char* __callocator contentblock_create(size_t size,
  tBoolean is_indirect)
/* creates a single block for content data */
{ char* retval;
  if (is_indirect)
  { retval = __memory_allocate(2 * sizeof(char**) + 2 * sizeof(size_t));
    __CONTENTBLOCK_USABLESIZE(retval) = size | __CONTENTBLOCK_INDIRECTIONBIT;
    __CONTENTBLOCK_DATAPTR(retval) = NULL;
    /* IMPORTANT: the _caller_ must set _DATAPTR to the appropriate value! */
  }
  else
  { retval = __memory_allocate(size + sizeof(char**) + 2 * sizeof(size_t*));
    __CONTENTBLOCK_USABLESIZE(retval) = size;
  }
  CONTENTBLOCK_USEDSIZE(retval) = 0;
  CONTENTBLOCK_NEXT(retval) = NULL;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "created contentblock (%p,%d)\n", retval, size);
  debugmsg(debugstrbuf);
#endif
  return(retval);
}

static void contentblocklist_deallocate(const char* headblock)
{ while (headblock != NULL)
  { const char* nextblock = CONTENTBLOCK_NEXT(headblock);
#if CONFIG_DEBUG
    sprint_safe(debugstrbuf, "contentblocklist_deallocate(): %p\n", headblock);
    debugmsg(debugstrbuf);
#endif
    if (CONTENTBLOCK_DATA_IS_INDIRECT(headblock))
    { const char* data = __CONTENTBLOCK_DATAPTR(headblock);
      if (data != NULL) /* (necessary check; see comment in drop_resource()) */
        memory_deallocate(data);
    }
    memory_deallocate(headblock);
    headblock = nextblock;
  }
}

#define HIGHEST_OCBS (131072)

static size_t optimal_contentblocksize(size_t count)
/* calculates the optimal :-) size for a new content block of a resource,
   depending on how many bytes we already got for it; the assumption behind
   this is that e.g. if we already received 100,000 bytes there might be
   100,000 more to come, so we allocate bigger blocks, read in larger chunks
   (as far as the socket buffer of the operating system allows), reserve memory
   less frequently and thus hopefully improve performance... */
{ size_t retval;
  if (count <= 10000) retval = 4096;
  else if (count <= 40000) retval = 8192;
  else if (count <= 200000) retval = 32768;
  else retval = HIGHEST_OCBS; /* Downloading Linux kernel sources, eh? :-) */
  return(retval);
}

static void deallocate_resource_tree(tResource* resource)
{ tHtmlNode* node = resource->tree;
  while (node != NULL)
  { tHtmlNode* next = node->next;
    deallocate_html_node(node);
    node = next;
  }
  resource->tree = NULL;
}

static void resource_set_firstcontent(tResource* resource, char* content)
{ if (resource->tree != NULL) deallocate_resource_tree(resource);
  resource->content = resource->lastcontent = resource->lhpp_content = content;
  resource->lhpp_byte = 0;
}

static char* resource_append_new_contentblock(tResource* resource,
  size_t usablesize)
{ char* retval = contentblock_create(usablesize, falsE);
  if (resource->content == NULL) resource_set_firstcontent(resource, retval);
  else
  { if (resource->lastcontent != NULL)
    { CONTENTBLOCK_NEXT(resource->lastcontent) = retval; }
    resource->lastcontent = retval;
  }
  return(retval);
}


/* Low-level resource handling */

typedef struct tResourceCacheEntry
{ struct tResourceCacheEntry *prev, *next;
  tResource* resource;
  size_t refcount;
} tResourceCacheEntry;

#if 0
static size_t ramcachesize_used = 0, ramcachesize_freeable = 0,
  diskcachesize_used = 0;
#endif

#define HASHTABLESIZE_RC (HASHTABLESIZE)
static tResourceCacheEntry* rc_head[HASHTABLESIZE_RC];

static tHashIndex resource_uri2hashindex(const char* uri)
/* calculates a hash table index from the given URI; we take at most the
   first 50 characters of the URI into account - this "should be enough" for
   good hashing and we need not do "%" after each partial operation (because
   the maximum possible value, 50 * 255, fits into a tHashIndex variable). */
{ tHashIndex retval = 0;
  unsigned char count = 50;
  while (count-- > 0)
  { unsigned char ch = *((const unsigned char*) uri);
    if (ch == '\0') break; /* reached end of string */
    retval += ((tHashIndex) ch);
    uri++;
  }
  return(retval % HASHTABLESIZE_RC);
}

static void attach_resource_to_request(tResourceRequest* request,
  tResource* resource)
{ request->resource = resource;
  request->state = rrsAttachedResource;
  request->flags |= rrfResourceChanged;
  reqlist_add(&(resource->reqlist), &(resource->reqlist_num),
    &(resource->reqlist_maxnum), request);
}

static /* __sallocator -- not an "only" reference */ tResource* __callocator
  resource_create(tResourceRequest* request, /*const*/ char* content,
  size_t nominal_contentlength, tResourceContentKind kind,tResourceState state)
{ const char* uri = request->uri;
  tResourceProtocol protocol = request->protocol;
  tHashIndex i = resource_uri2hashindex(uri);
  tResourceCacheEntry* entry = memory_allocate(sizeof(tResourceCacheEntry));
  tResource* retval = (tResource*) memory_allocate(sizeof(tResource));

#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "created resource at %p\n", retval);
  debugmsg(debugstrbuf);
#endif

  attach_resource_to_request(request, retval);

  /* Setup the cache entry for the resource */
  entry->resource = retval;
  entry->refcount = 1;
  if (rc_head[i] != NULL) { entry->next = rc_head[i]; rc_head[i]->prev=entry; }
  rc_head[i] = entry;

  /* Setup the resource */
  retval->id = object_id_create();
  retval->uri = my_strdup(uri);
  if (request->path != NULL) retval->path = my_strdup(request->path);
  if (request->flags & rrfDownload)
  { retval->flags |= rfDownload | rfActivityWatched;
    retval->download_fd = request->download_fd;
  }
  retval->protocol = protocol;
  resource_set_firstcontent(retval, content);
  retval->nominal_contentlength = nominal_contentlength;
  retval->kind = kind;
  retval->state = state;
  retval->cache_entry = entry;
  return(retval);
}

static tResource* resource_lookup(tResourceRequest* request)
/* tries to find a resource with the given URI (from <request>) in the internal
   resource cache and attaches it to <request> if found; make sure to call this
   only after calculating the _final_ URI (e.g. after appending a "/" to the
   name of a local directory if necessary)! */
{ const char *uri, *post1;
  tHashIndex i;
  tResourceCacheEntry* entry;
  tResourceProtocol protocol;
  tResourceRequestFlags rrf;

  if ( (request->action != rraLoad) || (request->flags & rrfDownload) )
  { /* We're not "allowed" to look into the cache for this request, e.g.
       because the user wants to _re_-load the resource. */
    goto out;
  }

  /* Look into the RAM cache */

  uri = request->uri;
  i = resource_uri2hashindex(uri);
  entry = rc_head[i];
  if (entry == NULL) goto try_disk; /* (optimizing for the most likely case) */

  protocol = request->protocol;
  rrf = request->flags;
  post1 = null2empty(request->post);

  while (entry != NULL)
  { tResource* resource = entry->resource;
    if (resource->protocol != protocol) goto next_entry; /* different prot.s */
    if ( ((rrf & rrfPost) != 0) != ((resource->flags & rfPost) != 0) )
    { goto next_entry; } /* different methods */
    if (strcmp(resource->uri, uri)) goto next_entry; /* different URIs */
    if (strcmp(post1, null2empty(resource->post))) goto next_entry;
    /* CHECKME: also require identical authorization? */

    /* found it */
    entry->refcount++;
    attach_resource_to_request(request, resource);
#if CONFIG_DEBUG
    { char* spfbuf;
      my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf,
        "found resource in RAM cache (%p,%s)\n", resource, uri);
      debugmsg(spfbuf);
      my_spf_cleanup(debugstrbuf, spfbuf);
    }
#endif
    return(resource);
    next_entry:
    entry = entry->next;
  }

  /* Look into the disk cache */

  try_disk: {}

  /* Nothing found... */

  out:
  return(NULL);
}

static one_caller void start_save_as(tResourceRequest* request,
  const char* filename)
{ tResource* resource = request->resource;
  struct stat statbuf;
  tSaveAs* save_as;
  char *content, *spfbuf;
  tBoolean may_overwrite = cond2boolean(request->flags & rrfMayOverwrite);
  int fd, cflags = O_CREAT | O_TRUNC | O_WRONLY;
  if (!may_overwrite) cflags |= O_EXCL;
#if CONFIG_DEBUG
  my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf, "start_save_as(%p,%p,%s)\n",
    request, resource, filename);
  debugmsg(spfbuf);
  my_spf_cleanup(debugstrbuf, spfbuf);
#endif
  fd = my_create(filename, cflags, S_IRUSR | S_IWUSR);
  if (fd < 0)
  { tResource* reply;
    char* reply_content;
    size_t replysize;
    bad:
    my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("Error when trying to save the document \"%s\" to the file \"%s\": %s (%d)"), resource->uri, filename, my_strerror(errno), errno);
    build_reply:
    replysize = strlen(spfbuf);
    reply_content = contentblock_create(replysize, falsE);
    my_memcpy(CONTENTBLOCK_DATA(reply_content), spfbuf, replysize);
    my_spf_cleanup(strbuf, spfbuf);
    CONTENTBLOCK_USEDSIZE(reply_content) = replysize;
    reply = resource_create(request, reply_content, replysize, rckPlainText,
      rsComplete);
    reply->flags |= rfCannotReload;
    request->action = rraLoad; /* CHECKME: that's quite dirty... */
    request->flags |= rrfResourceChanged;
    reply->bytecount = replysize;
    push_to_main_req(request, 0);
    return;
  }
  if (my_fstat(fd, &statbuf) != 0)
  { int e; close_bad: e = errno; my_close(fd); errno = e; goto bad; }
  if (!S_ISREG(statbuf.st_mode))
  { if (S_ISDIR(statbuf.st_mode)) errno = EISDIR;
    goto close_bad;
  }
  make_fd_cloexec(fd);

  /* the fd is okay, prepare a save_as record */
  save_as = (tSaveAs*) __memory_allocate(sizeof(tSaveAs));
  save_as->fd = fd;
  save_as->next = resource->save_as;
  resource->save_as = save_as;

  /* save the resource as far as we already got it */
  content = resource->content;
  while (content != NULL)
  { size_t usedsize = CONTENTBLOCK_USEDSIZE(content);
    if (usedsize > 0) (void)my_write(fd, CONTENTBLOCK_DATA(content), usedsize);
      /* IMPLEMENTME: check for errors, remove record, alert user? */
      /* IMPROVEME: use a non-blocking algorithm? */
    content = CONTENTBLOCK_NEXT(content);
  }

  /* user feedback */
  my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("Started saving the document \"%s\" to the file \"%s\"."), resource->uri, filename);
  goto build_reply;
}

static void do_save_as(tResource* resource, const char* src, size_t size)
{ tSaveAs* save_as = resource->save_as;
  while (save_as != NULL)
  { (void) my_write(save_as->fd, src, size);
      /* IMPLEMENTME: check for errors, remove record, alert user? */
      /* IMPROVEME: use a non-blocking algorithm? */
    save_as = save_as->next;
  }
}

static void stop_save_as(/*@notnull@*/ tResource* resource)
{ tSaveAs* save_as;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "stop_save_as(%p,%p)\n", resource,
    ( (resource != NULL) ? (resource->save_as) : (NULL) ));
  debugmsg(debugstrbuf);
#endif
  save_as = resource->save_as;
  while (save_as != NULL)
  { tSaveAs* next = save_as->next;
    my_close(save_as->fd);
    memory_deallocate(save_as);
    save_as = next;
  }
  resource->save_as = NULL;
}

static char* download_buf = NULL;
static size_t download_buf_size;

static void resource_provide_room(tResource* resource, /*@out@*/ char** dest,
  /*@out@*/ size_t* size, size_t desired_size, unsigned char flags)
/* <desired_size> and <flags> are only relevant if resource->lastcontent isn't
   used: "&1": desired_size is exact; "&2": it's parameter for ocbs();
   otherwise it's minimum */
{ const tBoolean is_download = cond2boolean(resource->flags & rfDownload);
  char* lastcontent;
  size_t s;

  if ( (!is_download) && ( (lastcontent = resource->lastcontent) != NULL ) )
  { size_t usable = CONTENTBLOCK_USABLESIZE(lastcontent),
      used = CONTENTBLOCK_USEDSIZE(lastcontent);
    if (used < usable) /* there's some free room left */
    { *dest = CONTENTBLOCK_DATA(lastcontent) + used;
      *size = usable - used;
      return;
    }
  }

  if ( (flags & 1) && (desired_size > 0) )
  { /* ("desired_size > 0" "should" always be true if "flags & 1"...) */
    s = desired_size;
  }
  else if (flags & 2) s = optimal_contentblocksize(desired_size);
  else
  { size_t ocbs = optimal_contentblocksize(0);
    s = MAX(desired_size, ocbs);
  }

  if (is_download)
  { if (download_buf == NULL) download_buf_size = 4096;
    if ( (download_buf_size < s) && (download_buf_size < HIGHEST_OCBS) )
    { while ( (download_buf_size < s) && (download_buf_size < HIGHEST_OCBS) )
        download_buf_size <<= 1;
      dealloc(download_buf);
    }
    if (download_buf == NULL)
      download_buf = __memory_allocate(download_buf_size);
    *dest = download_buf; *size = download_buf_size;
  }
  else
  { char* block = resource_append_new_contentblock(resource, s);
    *dest = CONTENTBLOCK_DATA(block); *size = CONTENTBLOCK_USABLESIZE(block);
  }
}

static tBoolean got_activity;
static void resource_record(tResource* resource, const char* data, size_t size)
{ if (resource->flags & rfDownload)
  { ssize_t err = my_write(resource->download_fd, data, size);
    if (err < 0) resource_set_error(resource, conn_err2error(errno));
    /* FIXME: handle <err> further, stop the downloading! */
    if ( (data == download_buf) && (size >= download_buf_size) )
    { /* looks like the server was fast enough to fill our buffer completely,
         so a larger buffer might increase throughput */
      if (download_buf_size < HIGHEST_OCBS)
      { download_buf_size <<= 1; memory_deallocate(download_buf);
        download_buf = __memory_allocate(download_buf_size);
      }
    }
  }
  else
  { CONTENTBLOCK_USEDSIZE(resource->lastcontent) += size;
    do_save_as(resource, data, size);
  }
  resource->bytecount += size;
  if (resource->flags & rfActivityWatched) got_activity = truE;
#if CONFIG_DEBUG
  resource->version++;
#endif
}

static void resource_collect(tResource* resource, const char* src, size_t size)
{ if (resource->flags & rfDownload) resource_record(resource, src, size);
  else /* might have to record it in smaller pieces */
  { size_t available_size, copy_size;
    tBoolean did_loop = falsE;
    char* dest;
    loop:
    resource_provide_room(resource, &dest, &available_size, size, 0);
    copy_size = MIN(size, available_size);
    my_memcpy(dest, src, copy_size);
    src += copy_size; size -= copy_size;
    resource_record(resource, dest, copy_size);
    if ( (size > 0) && (!did_loop) ) { did_loop = truE; goto loop; }
#if CONFIG_DEBUG
    if (size > 0) debugmsg("BUG: resource_collect()\n"); /* "can't happen" */
#endif
  }
}

static my_inline void resource_collect_str(tResource* resource,
  const char* str)
{ resource_collect(resource, str, strlen(str));
}

static void resource_collect_title2(tResource* resource, const char* h,
  const char* b)
/* builds an HTML document intro with title strings for head and body parts */
{ char buf[STRBUF_SIZE];
  char* spfbuf;
  my_spf(buf, STRBUF_SIZE, &spfbuf, "<html><head><title>%s</title></head>\n"
    "<body>\n<h3 align=\"center\"><b>%s</b></h3>\n", h, b);
  resource_collect_str(resource, spfbuf);
  my_spf_cleanup(buf, spfbuf);
}

static my_inline void resource_collect_title(tResource* resource,
  const char* title)
{ resource_collect_title2(resource, title, title);
}

static void stop_resource(tResource* resource)
{ tBoolean did = falsE;
  if (resource->cconn != NULL) { conn_remove(&(resource->cconn)); did = truE; }
  if (resource->dconn != NULL) { conn_remove(&(resource->dconn)); did = truE; }
  if (resource->save_as != NULL) { stop_save_as(resource); did = truE; }
  if ( (did) && (resource->state != rsError) ) resource->state = rsStopped;
}

#if USE_RPSD
static void rpsd_remove(tResource* resource)
{ tRpsdGeneric* rpsd = resource->rpsd;
  if (rpsd != NULL)
  { tRpsdRemover remover = rpsd->remover;
    if (remover != NULL) (remover)(rpsd);
    memory_deallocate(rpsd); resource->rpsd = NULL;
  }
}
#endif

static one_caller void drop_resource(tResource* resource)
{ tResourceCacheEntry* entry = resource->cache_entry;
  tHashIndex i;

#if CONFIG_DEBUG
  { char* spfbuf;
    my_spf(debugstrbuf, STRBUF_SIZE, &spfbuf, "drop_resource(%p,%d,%s)\n",
      resource, entry->refcount, resource->uri);
    debugmsg(spfbuf);
    my_spf_cleanup(debugstrbuf, spfbuf);
  }
#endif

  if (entry->refcount <= 0) debugmsg("BUG: bad resource cache refcount\n");
  else entry->refcount--;
  if (entry->refcount > 0) return; /* done (resource is still in use) */

  /* Stop any ongoing IO for this resource */

  stop_resource(resource);

  /* Deallocate the resource */

  debugmsg("deallocating resource\n");
  if (resource->flags & rfNeedUnmap)
  { my_munmap(CONTENTBLOCK_DATA(resource->content),
      resource->nominal_contentlength);
    /* The following is necessary! (Otherwise contentblocklist_deallocate()
       will crash when trying to free the <data> - been there:-) */
    __CONTENTBLOCK_DATAPTR(resource->content) = NULL;
  }
  i = resource_uri2hashindex(resource->uri);
  memory_deallocate(resource->uri);
  __dealloc(resource->path); __dealloc(resource->query);
  __dealloc(resource->post); __dealloc(resource->username);
  __dealloc(resource->password);
  __dealloc(resource->reqlist);

  if (resource->aenum > 0) /* deallocate active elements */
  { tActiveElementBase* aebase = resource->aebase;
    tActiveElementNumber _ae;
    for (_ae = 0; _ae < resource->aenum; _ae++)
      deallocate_one_aebase(&(aebase[_ae]));
    __dealloc(resource->aebase);
  }

  if (resource->hfnum > 0) /* deallocate forms */
  { tHtmlForm* form = resource->form;
    tHtmlFormNumber count;
    for (count = 0; count < resource->hfnum; count++)
      __dealloc(form[count].action_uri);
    __dealloc(resource->form);
  }

  if (resource->tree != NULL) deallocate_resource_tree(resource);
  if (resource->content != NULL)
  {
#if CONFIG_DEBUG
    sprint_safe(debugstrbuf, "deallocating contents of resource %p\n",
      resource);
    debugmsg(debugstrbuf);
#endif
    contentblocklist_deallocate(resource->content);
  }
#if USE_RPSD
  rpsd_remove(resource);
#endif
  object_mark_vanished(resource->id);
  memory_deallocate(resource);

  /* Deallocate the cache entry */

  if (rc_head[i] == entry) rc_head[i] = entry->next;
  if (entry->prev != NULL) entry->prev->next = entry->next;
  if (entry->next != NULL) entry->next->prev = entry->prev;
  memory_deallocate(entry);
}


/* Some stuff for HTTP/about resources */

#if OPTION_COOKIES
/* KISS */
#include "cookie.c"
#else
#define cookie_collect(resource) (strEmpty)
#define cookie_collect_cleanup(str) do { } while (0)
#endif

/* begin-autogenerated */
my_enum1 enum
{ hhpsDontCare = 0, hhpsConnection = 1, hhpsContentLength = 2,
  hhpsContentType = 3, hhpsLocation = 4, hhpsServer = 5, hhpsSetCookie2 = 6,
  hhpsSetCookie = 7, hhpsStatus = 8, hhpsTransferEncoding = 9
} my_enum2(unsigned char) tHttpHeaderParsingState;
#define MAX_HHPS (9)

static const char* const strHhps[MAX_HHPS + 1] =
{ strA /*don't care*/, "connection:", "content-length:", "content-type:",
  "location:", "server:", "set-cookie2:", "set-cookie:", "status:",
  "transfer-encoding:"
};

static const unsigned char hhpslen[MAX_HHPS + 1] =
{ 1, 11, 15, 13, 9, 7, 12, 11, 7, 18 };

#if CONFIG_ABOUT & 4
static one_caller void prepare_about_ctconfig(void)
{ sprint_safe(strbuf, "OPTION_TEXTMODEMOUSE = %d\n<br>OPTION_I18N = %d\n<br>OPTION_CED = %d\n<br>OPTION_COOKIES = %d\n<br>OPTION_NEWS = %d\n<br>OPTION_LOCAL_CGI = %d\n<br>OPTION_EXECEXT = %d\n<br>OPTION_TLS = %d\n<br>OPTION_IPV6 = %d\n<br>OPTION_THREADING = %d", OPTION_TEXTMODEMOUSE, OPTION_I18N, OPTION_CED, OPTION_COOKIES, OPTION_NEWS, OPTION_LOCAL_CGI, OPTION_EXECEXT, OPTION_TLS, OPTION_IPV6, OPTION_THREADING);
}
#endif
/* end-autogenerated */


/* "about:" resources */

static void hostinfo2uistr(const tCachedHostInformation* hostinfo,
  /*@out@*/ char* buf, size_t size)
/* converts a network address to a UI string and puts that into <buf> */
{ const tSockaddr* addr;
  int address_family;
  buf[0] = '\0';
  if (!(hostinfo->flags & chifAddressValid)) return;
  addr = &(hostinfo->addrinfo); address_family = hostinfo->address_family;
  /* Try the library function with the lower bug probability first... */
#if HAVE_INET_NTOP
  if (address_family == AF_INET)
  { const struct sockaddr_in* a = (const struct sockaddr_in*) addr;
    if ( (inet_ntop(address_family, (const void*) (&(a->sin_addr.s_addr)),
          buf, size) != NULL) && (buf[0] != '\0') )
    { return; }
  }
#if USE_IPv6
  else if (address_family == AF_INET6)
  { const struct sockaddr_in6* a = (const struct sockaddr_in6*) addr;
    if ( (inet_ntop(address_family, (const void*) (&(a->sin6_addr.s6_addr)),
          buf, size) != NULL) && (buf[0] != '\0') )
    { return; }
  }
#endif /* #if USE_IPv6 */
#endif /* #if HAVE_INET_NTOP */
#if HAVE_GETNAMEINFO
  if ( (getnameinfo((const struct sockaddr*) addr, hostinfo->addrinfolen,
        buf, size, NULL, 0, NI_NUMERICHOST) == 0) && (buf[0] != '\0') )
  { return; }
#endif
}

tBoolean resource_ui_conn_ip(const tResource* resource, char* buf, size_t size)
/* produces a user interface string containing an IP address related to the
   <resource> if appropriate */
{ tBoolean retval = falsE;
  const tHostPortProtInfo* hppi;
  const tCachedHostInformation* hostinfo;
  if ( (resource->state == rsConnecting) &&
       ( (hppi = resource->actual_hppi) != NULL ) &&
       ( (hostinfo = hppi->host) != NULL ) )
  { hostinfo2uistr(hostinfo, buf, size);
    if (buf[0] != '\0') retval = truE;
  }
  /* IMPLEMENTME: append portnumber; check ->dconn for FTP-like protocols? */
  return(retval);
}

static one_caller void fetch_about(tResourceRequest* request)
{ const char* path;
  if (resource_lookup(request) != NULL) return; /* found in cache, done */
  path = request->path;

  if (!strcmp(path, strRetawq))
  { tResource* resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH,
      rckHtml, rsComplete);
    char* spfbuf;

    /* introductory text */

#if CONFIG_TG == TG_CONSOLE
#define WHAT ", console"
#elif CONFIG_TG == TG_NCURSES
#define WHAT ", ncurses"
#elif CONFIG_TG == TG_XCURSES
#define WHAT ", xcurses"
#elif TGC_IS_CURSES
#define WHAT ", curses"
#elif CONFIG_TG == TG_X
#define WHAT ", X Window System"
#elif CONFIG_TG == TG_GTK
#define WHAT ", GTK"
#else
#define WHAT ""
#endif

    my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<html><head><title>About retawq</title></head>\n<body>\n<h3 align=\"center\"><b>retawq %s (%s mode%s)</b></h3>\n<p>The web browser <b>retawq</b> is released <b>without any warranty</b>. The project home page is <a href=\"http://retawq.sourceforge.net/\">retawq.sourceforge.net</a>.</p>\n"), RETAWQ_VERSION, _(strTG), WHAT); /* what the h[ae]ck... :-) */
    resource_collect_str(resource, spfbuf);
    my_spf_cleanup(strbuf, spfbuf);

#undef WHAT

    /* support text */

#if OPTION_I18N || CONFIG_JAVASCRIPT || CONFIG_CSS || OPTION_TLS
    { static const char strCommaSpace[] = ", ";
      tBoolean is_first = truE;
      resource_collect_str(resource,
        _("<p>This program contains support for: "));
#define ADDSUPP(str) \
  do \
  { if (is_first) is_first = falsE; \
    else resource_collect_str(resource, strCommaSpace); \
    resource_collect_str(resource, str); \
  } while (0)
#if OPTION_I18N
      ADDSUPP("i18n");
#endif
#if CONFIG_JAVASCRIPT
      ADDSUPP("Javascript");
#endif
#if CONFIG_CSS
      ADDSUPP("CSS");
#endif
#if OPTION_TLS
      ADDSUPP(strTLS);
#endif
#undef ADDSUPP
      resource_collect_str(resource, ".</p>\n");
    }
#endif

    /* UI library text */

    spfbuf = NULL;

#if CONFIG_TG == TG_X

    if (xws_display != NULL)
    { my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<p>Currently using <a href=\"http://www.x.org/\">X Window System</a> library version %d.%d (vendor: %s, release %d) on display \"%s\".</p>\n"), ProtocolVersion(xws_display), ProtocolRevision(xws_display), ServerVendor(xws_display), VendorRelease(xws_display), DisplayString(xws_display));
    }

#elif CONFIG_TG == TG_GTK

    my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<p>Currently using <a href=\"http://www.gtk.org/\">GTK</a> library version %d.%d.%d on display \"%s\".</p>\n"), gtk_major_version, gtk_minor_version, gtk_micro_version, gdk_get_display()); /* gdk_display_name */

#elif (CONFIG_TG == TG_NCURSES) && defined(NCURSES_VERSION)

/* work-around for ncurses silliness... */
#if (defined(NCURSES_VERSION_MAJOR)) && (NCURSES_VERSION_MAJOR >= 5)
#define my_curses_version curses_version()
#else
#define my_curses_version NCURSES_VERSION
#endif

    my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<p>Currently using <a href=\"http://www.gnu.org/directory/ncurses.html\">ncurses</a> library version \"%s\".</p>\n"), my_curses_version);

#endif

    if (spfbuf != NULL)
    { resource_collect_str(resource, spfbuf);
      my_spf_cleanup(strbuf, spfbuf);
    }

    /* TLS/libgcrypt library text */

#if OPTION_TLS == TLS_GNUTLS
    { const char* v = gnutls_check_version(NULL);
      if (v == NULL) v = _(strUnknown);
      my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<p>Currently using <a href=\"http://www.gnutls.org/\">GnuTLS</a> library version %s.%s</p>\n"), v, tls_version_warning(v));
      resource_collect_str(resource, spfbuf);
      my_spf_cleanup(strbuf, spfbuf);
    }
#elif OPTION_TLS == TLS_OPENSSL
    { const char *v = SSLeay_version(SSLEAY_VERSION), *v2 = null2empty(v),
        *v3 = ((strneqcase(v2, "openssl ", 8)) ? v2 + 8 : v2);
      my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<p>Currently using <a href=\"http://www.openssl.org/\">OpenSSL</a> library version \"%s\".%s</p>\n"), v2, tls_version_warning(v3));
      resource_collect_str(resource, spfbuf);
      my_spf_cleanup(strbuf, spfbuf);
    }
#elif OPTION_TLS == TLS_BUILTIN
    { const char *v = gcry_check_version(NULL), *v2 = null2empty(v);
      my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<p>Currently using <a href=\"http://www.gnupg.org/\">libgcrypt</a> library version %s.%s</p>\n"), v2, tls_version_warning(v2));
      resource_collect_str(resource, spfbuf);
      my_spf_cleanup(strbuf, spfbuf);
    }
#endif

    /* initial messages */

    if (initial_messages != NULL)
    { resource_collect_str(resource, _("<p>Start messages:"));
      resource_collect_str(resource, initial_messages);
      resource_collect_str(resource, "</p>\n");
    }

    /* finish */

    resource_collect_str(resource, strEndBodyHtml);
  }
#if CONFIG_ABOUT & 1
  else if (!strcmp(path, "activity")) /* connection activity list */
  { tResource* resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH,
      rckHtml, rsComplete);
    tConnTableIndex i;
    tBoolean is_first = truE;
    char schemebuf[MAXSCHEMESTRSIZE];
    resource->flags |= rfActivityUnwatchable;
    resource_collect_title(resource, _("Activity"));
    for (i = 0; i < conn_table_num; i++)
    { const tConnection* conn = conn_table[i];
      const tHostPortProtInfo* hppi = conn->hppi;
      const tResource* res;
      const char* info = NULL;
      char* spfbuf;
      tBoolean must_cleanup_info;
      /* IMPLEMENTME: reasons for not actually listing a connection? */
      /* list this connection */
      is_first = must_cleanup_info = falsE;
#if NEED_DISSOLVING
      if (conn->flags & cnfDissolving)
      { info = unconstify_or_("disconnecting");
        /* technically wrong notion, but users will understand it better than
           "dissolving" or "shutting down" :-) */
        goto finish;
      }
#endif
      res = conn2resource(conn);
      if (res != NULL)
      { size_t bytecount = res->bytecount;
        char bytecountbuf[200];
        if (bytecount < 2) bytecountbuf[0] = '\0';
        else
        { sprint_safe(bytecountbuf, "%s%d %s", strSpacedDash, bytecount,
            _(strBytes));
        }
        my_spf(strbuf2, STRBUF_SIZE, &spfbuf, strPercsPercs, res->uri,
          bytecountbuf);
        info = my_spf_use(spfbuf); must_cleanup_info = truE;
      }

      if (hppi != NULL)
      { /* IMPLEMENTME: show hostinfo! */
      }

#if OPTION_TLS
      { tTlsSession session = conn->tls_session;
        if (session != NULL) { /* IMPLEMENTME: show TLS info (cipher etc.) */ }
      }
#endif

      finish:
      rp2scheme(conn->protocol, schemebuf);
      my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("<p>fd #%d, scheme %s%s%s</p>"),
        conn->fd, schemebuf, ( (info != NULL) ? strSpacedDash : strEmpty ),
        null2empty(info));
      resource_collect_str(resource, spfbuf);
      if (must_cleanup_info) my_spf_cleanup(strbuf2, info);
      my_spf_cleanup(strbuf, spfbuf);
    }
    if (is_first) resource_collect_str(resource, _("<p>none</p>"));
    resource_collect_str(resource, strNewlineEndBodyHtml);
  }
#endif /* #if CONFIG_ABOUT & 1 */
#if CONFIG_ABOUT & 2
  else if (!strcmp(path, "hostinfo")) /* host information cache excerpt */
  { /* This exists mostly for debugging purposes, but... */
    tResource* resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH,
      rckHtml, rsComplete);
    tBoolean is_first_nonempty_bucket = truE;
    tHashIndex i;
    resource_collect_title(resource,
      _("Contents of the host information cache"));

    for (i = 0; i < HASHTABLESIZE_CHI; i++)
    { const tCachedHostInformation* hostinfo = chi_head[i];
      tBoolean is_first_in_bucket;
      if (hostinfo == NULL) continue; /* bucket empty */
      is_first_in_bucket = truE;
      do
      { const tHostPortProtInfo* hppi = hostinfo->hppi;
        char* spfbuf;
        if (is_first_in_bucket)
        { my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("\n<p><b>Bucket #%d</b></p>"),
            i);
          resource_collect_str(resource, spfbuf);
          my_spf_cleanup(strbuf, spfbuf);
          is_first_in_bucket = falsE;
        }
        resource_collect_str(resource, strNewlineBr);

        hostinfo2uistr(hostinfo, strbuf2, STRBUF_SIZE);
        my_spf(strbuf, STRBUF_SIZE, &spfbuf,
          _("\n<p>name: %s; IP address: %s; flags: %d"), hostinfo->hostname,
          ( (strbuf2[0] != '\0') ? strbuf2 : _(strUnknown) ), hostinfo->flags);
        resource_collect_str(resource, spfbuf);
        my_spf_cleanup(strbuf, spfbuf);

#if 0
        while (hppi != NULL)
        { tResourceProtocol protocol = hppi->protocol;
          const char *username, *password, *swid = hppi->software_id;
          char schemebuf[MAXSCHEMESTRSIZE], *username3, *swid3;
          tHostPortProtInfoFlags hppif = hppi->flags;
          rp2scheme(protocol, schemebuf);
          if (username != NULL)
          { const char* username2 = htmlify(username);
            my_spf(NULL, 0, &username3, _("; username: \"%s\""), username2);
            htmlify_cleanup(username, username2);
          }
          if (password != NULL) password = _("; password: (given)");
          if (swid != NULL)
          { const char* swid2 = htmlify(swid);
            my_spf(strbuf2, STRBUF_SIZE, &swid3, _("; server: \"%s\""), swid2);
            /* CHECKME: hide POP3 timestamps here? */
            htmlify_cleanup(swid, swid2);
          }
          my_spf(strbuf, STRBUF_SIZE, &spfbuf,
            _("\n<br>port %d: scheme %s%s%s%s%s; flags: %d"),
            ntohs(hppi->portnumber), schemebuf, (username ? username3 :
            strEmpty), null2empty(password), (swid ? swid3 : strEmpty),
            ( (hppif & hppifCannotHttp11) ? _("; not HTTP/1.1") :
              strEmpty ), hppif);
          resource_collect_str(resource, spfbuf);
          my_spf_cleanup(strbuf, spfbuf);
          if (username != NULL) memory_deallocate(username3);
          if (swid != NULL) my_spf_cleanup(strbuf2, swid3);
          hppi = hppi->next;
        }
#endif
#if OPTION_COOKIES
        if (hostinfo->cookies != NULL)
        { const char* buf = cookie_reviewlist(hostinfo);
          if (buf != NULL)
          { resource_collect_str(resource, buf); memory_deallocate(buf); }
        }
#endif
        resource_collect_str(resource, "</p>\n");
        hostinfo = hostinfo->next;
      } while (hostinfo != NULL);
      is_first_nonempty_bucket = falsE;
    }

    if (is_first_nonempty_bucket) /* found nothing at all */
      resource_collect_str(resource, _("<p>No information known yet.</p>\n"));
    resource_collect_str(resource, strEndBodyHtml);
  }
#endif /* #if CONFIG_ABOUT & 2 */
#if CONFIG_ABOUT & 4
  else if (!strcmp(path, "ctconfig")) /* compile-time configuration */
  { tResource* resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH,
      rckHtml, rsComplete);
    resource_collect_title(resource,
      _("retawq Compile-Time Configuration Options"));
    prepare_about_ctconfig(); resource_collect_str(resource, strbuf);
    resource_collect_str(resource, strNewlineEndBodyHtml);
  }
#endif /* #if CONFIG_ABOUT & 4 */
  else
  { resource_request_set_error(request, reUri);
  }
}


/* "local:" resources */

typedef struct
{
#if CONFIG_LOCALDIR > 1
  struct stat statbuf;
#endif
  const char *key, *value;
#if CONFIG_LOCALDIR > 1
  tBoolean statbuf_is_good;
#endif
} tDirSorterElement;

#if CONFIG_LOCALDIR > 1

static const char strQuerySort[] = "?sort=";
static const char* current_localdirsort;

static one_caller char* local_query2sorting(const char* query)
{ char* retval = NULL;
  enum { buflen = 20 };
  static char buf[buflen];
  if (query != NULL)
  { size_t len = strlen(strQuerySort) - 1, slen;
    if (my_strstr(query, strQuerySort + 1))
    { const char *sorting = query + len, *end, *end_max;
      if (*sorting == '\0') goto out;
      end = my_strchr(sorting, '&');
      if (end == NULL) end = sorting + strlen(sorting);
      end_max = sorting + buflen - 5;
      if (end > end_max) end = end_max;
      slen = end - sorting;
      my_memcpy(buf, sorting, slen);
      buf[slen] = '\0';
      check_localdirsort(buf);
      if (buf[0] != '\0') retval = buf;
    }
  }
  out:
  return(retval);
}

static int mode2num(mode_t mode)
/* converts a file type (file/directory/...) to a number for sorting */
{ int retval;
  if (S_ISDIR(mode)) retval = 1;
  else if (S_ISREG(mode)) retval = 2;
  else retval = 3;
  return(retval);
}

static one_caller /*@observer@*/ /*@null@*/ const char*
  path2format(const char* path)
{ const char* retval = NULL;
  const tConfigLocaldirformat* ldf = config.ldf;
  while (ldf != NULL)
  { const char* pattern = ldf->path_pattern;
    size_t len;
    if ( (pattern == NULL) || (pattern[0] == '\0') ) goto nxt;
    len = strlen(pattern);
    if ( ( (pattern[len - 1] == '*') && (!strncmp(pattern, path, len - 1)) ) ||
         (!strcmp(pattern, path)) )
    { retval = ldf->format;
      if ( (retval != NULL) && (retval[0] == '\0') ) retval = NULL;
      break;
    }
    nxt:
    ldf = ldf->next;
  }
  return(retval);
}

#endif /* #if CONFIG_LOCALDIR > 1 */

static int directory_sorter(const void* _a, const void* _b)
{ const tDirSorterElement *a = (const tDirSorterElement*) _a,
    *b = (const tDirSorterElement*) _b;
#if CONFIG_LOCALDIR > 1
  const struct stat *as, *bs;
  const char* sorting;
  char sortcrit;

  if (*current_localdirsort == '\0') goto out; /* the most likely case */
  if ( (!(a->statbuf_is_good)) || (!(b->statbuf_is_good)) ) goto out;
  as = &(a->statbuf); bs = &(b->statbuf);
  sorting = current_localdirsort;
  while ( (sortcrit = *sorting++) != '\0' )
  { int val;
    tBoolean reverse;
    if (my_islower(sortcrit)) reverse = falsE;
    else { sortcrit = my_tolower(sortcrit); reverse = truE; }
    switch (sortcrit)
    { case 'g': val = my_numcmp(as->st_gid, bs->st_gid); break;
      case 'i': val = my_strcasecmp(a->key, b->key); break;
      case 'm': val = my_numcmp(as->st_mtime, bs->st_mtime); break;
      case 'n': val = strcmp(a->key, b->key); break;
      case 's': val = my_numcmp(as->st_size, bs->st_size); break;
      case 't': val = mode2num(as->st_mode) - mode2num(bs->st_mode); break;
      case 'u': val = my_numcmp(as->st_uid, bs->st_uid); break;
      default: val = 0; break; /* "should not happen" */
    }
    if (val != 0) /* found an "applicable" sorting criterium */
    { if (reverse) val = -val;
      return(val);
    }
  }
  return(0); /* no criterium applied */
  out: {}
#endif
  return(strcmp(a->key, b->key)); /* just sort by names */
}

#if CONFIG_LOCALDIR > 0
static my_inline char inodemode2str(mode_t mode)
{ if (S_ISREG(mode)) return(config.char_file);
  else if (S_ISDIR(mode)) return(config.char_dir);
  else return('?');
}
static char dirperms_buf[10];
#define __dirperms(pr, pw, px) \
  *p++ = ( (mode & pr) ? ('r') : ('-') ); \
  *p++ = ( (mode & pw) ? ('w') : ('-') ); \
  *p++ = ( (mode & px) ? ('x') : ('-') );
static one_caller void calc_dirperms(mode_t mode)
{ char* p = dirperms_buf;
  __dirperms(S_IRUSR, S_IWUSR, S_IXUSR)
  __dirperms(S_IRGRP, S_IWGRP, S_IXGRP)
  __dirperms(S_IROTH, S_IWOTH, S_IXOTH)
  *p = '\0';
}
#endif /* #if CONFIG_LOCALDIR > 0 */

static /*@observer@*/ const char strParentDirectory[] = N_("Parent directory");

static one_caller tBoolean fetch_local_directory(tResourceRequest* request)
/* returns whether it worked */
{ const char *path = request->path, *htmlpath, *spfbuf2;
#if CONFIG_LOCALDIR > 1
  const char* format;
#endif
  char *spfbuf, *temp;
  const struct dirent* entry;
  unsigned int element_count, element_maxcount, i;
  /*@relnull@*/ tDirSorterElement* sorter_base;
  tResource* resource;
  DIR* dir = opendir(path);
  if (dir == NULL) return(falsE);

  /* Read the directory contents */

#if CONFIG_LOCALDIR > 1
  format = path2format(path);
#endif
  sorter_base = NULL;
  element_count = element_maxcount = 0;
  while ( (entry = readdir(dir)) != NULL )
  { const char *name = entry->d_name, *key, *htmlkey;
    if (name == NULL) continue; /* "should not happen", library bug */
    if (name[0] == '.')
    { char c = name[1];
      if ( (c == '\0') || ( (c == '.') && (name[2] == '\0') ) )
      { continue; } /* we aren't interested in "." and ".." */
    }

    if (element_maxcount <= element_count)
    { element_maxcount += 20;
      sorter_base = memory_reallocate(sorter_base, element_maxcount *
        sizeof(tDirSorterElement));
    }
    key = my_strdup(name); htmlkey = htmlify(key);
    sorter_base[element_count].key = key;

#if CONFIG_LOCALDIR > 0
    /* The user wants more information about the contents of the directory. */
    { int err;
      struct stat statbuf;
      my_spf(strbuf, STRBUF_SIZE, &spfbuf, strPercsPercs, path, key);
      err = my_stat(spfbuf, &statbuf);
      my_spf_cleanup(strbuf, spfbuf);
      if (err != 0)
      { /* very unlikely... */
#if CONFIG_LOCALDIR > 1
        sorter_base[element_count].statbuf_is_good = falsE;
#endif
        goto just_the_name;
      }
      else
      { mode_t mode = statbuf.st_mode;
        tBoolean is_dir = cond2boolean(S_ISDIR(mode));
        const char* slash = (is_dir ? strSlash : strEmpty);
        if (!S_ISREG(mode)) spfbuf2 = strEmpty;
        else
        { size_t bytes = (size_t) statbuf.st_size;
          my_spf(strbuf2, STRBUF_SIZE, &temp, " (%d %s)",localized_size(bytes),
            ( (bytes == 1) ? _(strByte) : _(strBytes)));
          spfbuf2 = my_spf_use(temp);
        }
        calc_dirperms(mode);
        my_spf(NULL, 0, &spfbuf,
          "<br>%c %s <a href=\"local:%s%s%s\">%s%s</a>%s\n",
          inodemode2str(mode), dirperms_buf, path, key, slash, htmlkey,
          slash, spfbuf2);
        if (spfbuf2 != strEmpty) my_spf_cleanup(strbuf2, spfbuf2);
#if CONFIG_LOCALDIR > 1
        sorter_base[element_count].statbuf = statbuf;
        sorter_base[element_count].statbuf_is_good = truE;
#endif
      }
    }
    if (falsE) /* ugly (-: */
#endif /* #if CONFIG_LOCALDIR > 0 */
    { just_the_name:
      my_spf(NULL, 0, &spfbuf, "<br><a href=\"local:%s%s\">%s</a>\n",
        path, key, htmlkey);
    }
    sorter_base[element_count].value = my_spf_use(spfbuf);

    element_count++;
    htmlify_cleanup(key, htmlkey);
  }
  (void) closedir(dir);

  /* Put the directory contents into an HTML page (sorted) */

  resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH, rckHtml,
    rsComplete);

  htmlpath = htmlify(path);
  my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("Contents of local directory \"%s\""),
    htmlpath);
  htmlify_cleanup(path, htmlpath);
  resource_collect_title(resource, spfbuf);
  my_spf_cleanup(strbuf, spfbuf);

  if (calc_parentpath(path, strbuf2, STRBUF_SIZE))
  { my_spf(strbuf, STRBUF_SIZE, &spfbuf,
      "<p><a href=\"local:%s\">%s</a></p>\n", strbuf2, _(strParentDirectory));
    resource_collect_str(resource, spfbuf);
    my_spf_cleanup(strbuf, spfbuf);
  }

  if (element_count <= 0)
  { resource_collect_str(resource, _("This directory is empty."));
  }
  else
  {
#if CONFIG_LOCALDIR > 1
    const char* clds = local_query2sorting(request->query);
    if (clds == NULL) /* no special sorting for this URI given */
    { tConfigLocaldirsort* lds = config.lds;
      while (lds != NULL)
      { const char* pattern = lds->path_pattern;
        size_t len;
        if ( (pattern == NULL) || (pattern[0] == '\0') ) goto nxt;
        len = strlen(pattern);
        if ( ( (pattern[len - 1] == '*') && (!strncmp(pattern,path,len-1)) ) ||
             (!strcmp(pattern, path)) )
        { clds = lds->sorting; break; }
        nxt:
        lds = lds->next;
      }
    }
    if (clds == NULL) clds = strEmpty; /* sort by name (default) */
    else if (*clds == '_') clds = NULL; /* disable all sorting */
    if (clds != NULL)
    { current_localdirsort = clds;
#else
    {
#endif
      qsort(sorter_base, element_count, sizeof(tDirSorterElement),
        directory_sorter);
    }
    for (i = 0; i < element_count; i++)
    { resource_collect_str(resource, sorter_base[i].value);
      memory_deallocate(sorter_base[i].key);
      memory_deallocate(sorter_base[i].value);
    }
  }

  resource_collect_str(resource, strEndBodyHtml);
  __dealloc(sorter_base);
  return(truE);
}

static one_caller void build_local_uri(tResourceRequest* request)
{ const char* path = request->path;
  char* spfbuf;
#if CONFIG_LOCALDIR > 1
  const char* query = request->query;
  if (query != NULL)
  { my_spf(NULL, 0, &spfbuf, "local:%s%s%s", path, strQuerySort, query);
  }
  else
#endif
  { my_spf(NULL, 0, &spfbuf, "local:%s", path);
  }
  __dealloc(request->uri); request->uri = my_spf_use(spfbuf);
}

static one_caller void fetch_local(tResourceRequest* request)
{ char* path = request->path;
  struct stat statbuf;
  mode_t mode;

  /* Check whether the thing really exists and whether it's a regular file or
     a directory */

  if (my_stat(path, &statbuf) != 0)
  { /* The thing doesn't exist (any longer) or is otherwise inaccessible, but
       maybe we have its old contents in cache. */
    try_cache:
    if (resource_lookup(request) != NULL) return; /* found in cache, done */
    bad: resource_request_set_error(request, reFile); return;
  }

  mode = statbuf.st_mode;
  if (S_ISDIR(mode)) /* it's a directory */
  { const size_t len = strlen(path);
    if ( ( (len > 0) && (path[len - 1] != '/') ) || (len == 0) )
    { request->path = path = memory_reallocate(path, len + 1 + 1);
      strcat(path, strSlash);
      build_local_uri(request);
    }
    if (resource_lookup(request) != NULL) return; /* found in cache, done */
    if (!fetch_local_directory(request)) goto bad;
  }
  else if (S_ISREG(mode)) /* it's a regular file */
  { tResource* resource;
    char* content;
    void* buf;
    size_t size;
    tBoolean need_unmap SHUT_UP_COMPILER(falsE);
    if (request->flags & rrfDownload) /* can't "download" local files */
    { resource_request_set_error(request, reUri); return; }
    if (resource_lookup(request) != NULL) return; /* found in cache, done */
    switch (my_mmap_file_readonly(path, &buf, &size))
    { case 0: goto bad; /*@notreached@*/ break;
      case 1: buf = memory_allocate(1); size = 0; need_unmap = falsE; break;
      case 2: need_unmap = truE; break;
    }
    content = contentblock_create(size, truE);
    CONTENTBLOCK_USEDSIZE(content) = size;
    __CONTENTBLOCK_DATAPTR(content) = buf;
    resource = resource_create(request, content, size, filesuffix2rck(path),
      rsComplete);
    if (need_unmap) resource->flags |= rfNeedUnmap;
  }
  else
  { /* The thing isn't a regular file or directory now, but maybe there was
       one at this position earlier, which we might have in cache. */
    goto try_cache;
  }
}


/* Local CGI resources */

#if OPTION_LOCAL_CGI

#if HAVE_SETENV
#define __my_setenv(var, val) (void) setenv(var, val, 1)
#define my_setenv(var, val) __my_setenv(var, val)
#elif HAVE_PUTENV
#define __my_setenv(var, val) (void) putenv(var "=" val)  /* fast and */
static void my_setenv(const char* var, const char* val)   /* slow variant */
{ char* spfbuf;
  my_spf(strbuf, STRBUF_SIZE, &spfbuf, "%s=%s", var, val);
  putenv(spfbuf);
  my_spf_cleanup(strbuf, spfbuf);
}
#else
#error Cannot use local CGI - your libc has neither setenv() nor putenv(). Disable the compile-time configuration option OPTION_LOCAL_CGI!
/* IMPLEMENTME? This is only executed in child processes after fork(), so there
   are no race conditions possible if we change environ manually... Or simply
   setup a minimal, CGI-specific environment and use execle(). */
#endif /* #if HAVE_SETENV/HAVE_PUTENV */

static one_caller uid_t my_geteuid(void)
{ static tBoolean done = falsE;
  static uid_t euid;
  if (!done) { euid = geteuid(); done = truE; } /* (calculate it only once) */
  return(euid);
}

static one_caller gid_t my_getegid(void)
{ static tBoolean done = falsE;
  static gid_t egid;
  if (!done) { egid = getegid(); done = truE; } /* (calculate it only once) */
  return(egid);
}

/* prototypes */
static void http_read(tResource*);
static void http_rpsd_prepare(tResource*);

static void local_cgi_callback(tConnection* conn, tConnCbEventKind ccek)
{ tResource* resource = (tResource*) (conn->data);
  if (conn->x == 0) /* reading the script's reply */
  { switch (ccek)
    { case ccekRead: http_read(resource); break;
      default: is_buggy: conn_bug(conn, ccek); break;
    }
  }
  else /* writing the post data */
  { switch (ccek)
    { case ccekWrite:
        if (conn_write_writedata(conn) != 1)
        { /* error or done - just stop writing, don't stop reading */
          conn_remove(&(resource->dconn));
        }
        break;
      default: goto is_buggy; /*@notreached@*/ break;
    }
  }
}

static one_caller tBoolean start_local_cgi(tResourceRequest* request)
/* starts a local CGI script if allowed/possible; returns want-it-back info */
{ const tConfigLocalCgi* clc = config.local_cgi;
  tResourceError error;
  tBoolean is_post;
  const char *path, *post SHUT_UP_COMPILER(NULL);
  size_t pathlen, postlen SHUT_UP_COMPILER(0);
  struct stat statbuf;
  mode_t mode;
  int fd_pair[2], fd_post[2];
  pid_t pid;

  /* First of all, check whether the configuration allows it */

  if (clc == NULL) goto forbidden; /* (optimizing for the most likely case) */
  path = request->path; pathlen = strlen(path);
  while (clc != NULL)
  { const char* pattern = clc->path_pattern;
    size_t len;
    if ( (!strcmp(path, pattern)) ||
         ( ( (len = strlen(pattern)) > 0 ) && (pattern[len - 1] == '*') &&
           (pathlen >= len) && (!strncmp(path, pattern, len - 1)) ) )
    { if (clc->flags & clcfAllowed) goto allowed;
      break;
    }
    clc = clc->next;
  }
  forbidden:
  error = reConfigForbids;
  failed:
  resource_request_set_error(request, error);
  return(falsE);

  allowed:

  /* Check cache */

  if (resource_lookup(request) != NULL) return(falsE); /*found in cache, done*/

  /* Check whether the thing actually exists, is a regular file, executable */

  if (my_stat(path, &statbuf) != 0) { bad_file: error = reFile; goto failed; }
  mode = statbuf.st_mode;
  if (!S_ISREG(mode)) goto bad_file;
  /* simple pre-check without all the get...id() calls: */
  if (!(mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
  { cant_exec: error = reExec; goto failed; }
  /* more thorough check: */
  if (! (( (mode & S_IXOTH) ||
         ( (mode & S_IXUSR) && (statbuf.st_uid == my_geteuid()) ) ||
         ( (mode & S_IXGRP) && (statbuf.st_gid == my_getegid()) ) )) )
    goto cant_exec;

  /* Execute the script */

  if (pipe(fd_pair) != 0) { pipe_failed: error = rePipe; goto failed; }
  if (!is_fd_selectable(fd_pair[0]))
  { tmofd: my_close(fd_pair[0]); my_close(fd_pair[1]);
    error = reTmofd; goto failed;
  }
  is_post = cond2boolean(request->flags & rrfPost);
  if (is_post)
  { if (pipe(fd_post) != 0)
    { my_close(fd_pair[0]); my_close(fd_pair[1]); goto pipe_failed; }
    if (!is_fd_selectable(fd_post[1]))
    { my_close(fd_post[0]); my_close(fd_post[1]); goto tmofd; }
    post = null2empty(request->post);
    postlen = strlen(post);
  }

  pid = fork();
  if (pid < 0) /* failed */
  { error = reFork;
    my_close(fd_pair[0]); my_close(fd_pair[1]);
    if (is_post) { my_close(fd_post[0]); my_close(fd_post[1]); }
    goto failed;
  }
  else if (pid == 0) /* child process */
  { char *spfbuf, *spfbuf2;
    const char *query, *temp;
    reset_signals();
    query = request->query;
    my_close(0); my_close(fd_pair[0]);
    if (!may_use_fd2()) my_close(2);
    if (fd_pair[1] != 1)
    { if (!my_dup2(fd_pair[1], 1))
      { dup2_failed:
        my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("retawq: local CGI error: can't setup file descriptors (error #%d, %s)\n"), errno, my_strerror(errno));
        if (may_use_fd2()) my_write_str(2, spfbuf);
        debugmsg(spfbuf);
        goto exit_child;
      }
      my_close(fd_pair[1]);
    }
    if (is_post)
    { my_close(fd_post[1]);
      if (fd_post[0] != 0)
      { if (!my_dup2(fd_post[0], 0)) goto dup2_failed;
        my_close(fd_post[0]);
      }
      sprint_safe(strbuf, strPercd, postlen);
      my_setenv("CONTENT_LENGTH", strbuf);
      my_setenv("CONTENT_TYPE", strAxwfu);
    }
    __my_setenv("GATEWAY_INTERFACE", "CGI/1.1");
      my_setenv("HTTP_USER_AGENT", config.user_agent);
    if ( (query != NULL) && (query[0] != '\0') )
      my_setenv("QUERY_STRING", query);
      my_setenv("REMOTE_HOST", strLocalhost);
    __my_setenv("REMOTE_ADDR", "127.0.0.1");
      my_setenv("REQUEST_METHOD", is_post ? strPost : strGet);
      my_setenv("SCRIPT_NAME", path);
      my_setenv("SERVER_NAME", strLocalhost);
    __my_setenv("SERVER_PORT", "80"); /* (just a dummy value) */
    __my_setenv("SERVER_PROTOCOL", "HTTP/1.0"); /* CHECKME! */
      my_setenv("SERVER_SOFTWARE", strSoftwareId);

    temp = my_strrchr(path, '/');
    if (temp != NULL) /* "should" be true */
    { size_t len = temp - path + 1, size = len + 1;
      char* ptr = ( (size <= STRBUF_SIZE) ? strbuf : __memory_allocate(size) );
      my_memcpy(ptr, path, len); ptr[len] = '\0';
      (void) chdir(ptr);
      if (ptr != strbuf) memory_deallocate(ptr);
    }
    (void) execl(path, path, NULL);

    /* If we get here, execl() failed, so we send a little error text to the
       resource handler and exit the child process: */
    my_spf(strbuf, STRBUF_SIZE, &spfbuf,
      _("Local CGI error: can't execute \"%s\" (error #%d, %s)%s"), path,errno,
      my_strerror(errno), ( (errno == ENOENT) ? _("\n\n(Note: an error like \"file not found\" might mean \"script interpreter not found\" here.)") : strEmpty ));
    my_spf(strbuf2, STRBUF_SIZE, &spfbuf2,
     "Status: 500\r\nContent-Type: text/plain\r\nContent-Length: %d\r\n\r\n%s",
     strlen(spfbuf), spfbuf);
    my_write_str(1, spfbuf2);
    exit_child:
    _exit(1);
  }
  else /* parent process */
  { tResource* resource;
    tConnection* conn;
    make_fd_cloexec(fd_pair[0]); my_close(fd_pair[1]);
    resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH,
      rckUnknown, rsReading);
    http_rpsd_prepare(resource);
    conn = resource->cconn = conn_create(fd_pair[0], local_cgi_callback,
      resource, truE, NULL, resource->protocol);
    conn->flags |= cnfConnected; conn_set_read(conn, 2);
    if (is_post)
    { my_close(fd_post[0]);
      if (postlen <= 0) { my_close(fd_post[1]); } /* nothing to post */
      else
      { make_fd_cloexec(fd_post[1]);
        (void) make_fd_nonblocking(fd_post[1]);
          /* (just trying; we won't let it all fail if only this fails...) */
        conn = resource->dconn = conn_create(fd_post[1], local_cgi_callback,
          resource, truE, NULL, resource->protocol);
        conn->flags |= cnfConnected; conn->x = 1; conn_set_write(conn, 2);
        conn_set_writestr(conn, my_strdup(post));
      }
      resource->post = my_strdup(post); resource->flags |= rfPost;
    }
  }
  return(truE);
}

#endif /* #if OPTION_LOCAL_CGI */


/* Execution of external programs: shell */

#if OPTION_EXECEXT & 1

static one_caller void execext_shell_read(tConnection* conn)
{ tResource* resource = (tResource*) (conn->data); /* always != NULL here */
  char* dest;
  size_t size;
  int err;
  resource_provide_room(resource, &dest, &size, resource->bytecount, 2);
  err = my_read(conn->fd, dest, size);
  if (err <= 0) /* error or EOF */
  { conn_remove(&(resource->cconn));
    if (resource->dconn != NULL) conn_remove(&(resource->dconn));
    if (err == 0) resource->state = rsComplete;
    else resource_set_error(resource, reServerClosed);
    push_to_main_res(resource, 0);
  }
  else
  { resource_record(resource, dest, err);
    push_to_main_res(resource, 1);
  }
}

static void execext_shell_callback(tConnection* conn, tConnCbEventKind ccek)
{ tResource* resource = conn2resource(conn);
  if (conn->x == 0) /* reading the command's output */
  { switch (ccek)
    { case ccekRead: execext_shell_read(conn); break;
      default: is_buggy: conn_bug(conn, ccek); break;
    }
  }
  else /* writing the command's input */
  { switch (ccek)
    { case ccekWrite:
        if (conn_write_writedata(conn) != 1)
        { /* error or done - just stop writing, don't stop reading */
          if (resource != NULL) conn_remove(&(resource->dconn));
          else conn_remove(&conn);
        }
        break;
      default: goto is_buggy; /*@notreached@*/ break;
    }
  }
}

static one_caller void start_shell(tExecextShellData* data)
{ tResourceRequest* request = data->request;
  const char *command = data->command, *writedata = data->writedata;
  tExecextShellFlags esf = data->esf;
  tResourceError error;
  const tBoolean do_read = cond2boolean(esf & (esfReadStdout | esfReadStderr)),
    do_write = cond2boolean(writedata != NULL);
  int fd_read[2], fd_write[2]; /* "read"/"write" from parent's point of view */
  unsigned char _cleanup = 0; /* what to cleanup on failure */
  pid_t pid;

  if (do_write)
  { if (pipe(fd_write) != 0)
    { pipe_failed: error = rePipe;
      failed:
      if (request != NULL)
      { resource_request_set_error(request, error);
        push_to_main_req(request, 0);
      }
      if (_cleanup & 1) { my_close(fd_write[0]); my_close(fd_write[1]); }
      if (_cleanup & 2) { my_close(fd_read[0]); my_close(fd_read[1]); }
      return;
    }
    _cleanup |= 1;
    if (!is_fd_selectable(fd_write[1])) { tmofd: error = reTmofd; goto failed;}
  }
  if (do_read)
  { if (pipe(fd_read) != 0) goto pipe_failed;
    _cleanup |= 2;
    if (!is_fd_selectable(fd_read[0])) goto tmofd;
  }

  pid = fork();
  if (pid < 0) { error = reFork; goto failed; }
  else if (pid == 0) /* child process */
  { const char *path, *errstr;
    char *spfbuf, *spfbuf2, *param[20];
    unsigned char count = 0;
    int err;
    if (config.flags & cfExecextShellCustom)
    { tExecextParam *list = config.execext_shell;
      while (list != NULL) { param[count++] = list->param; list = list->next; }
    }
    else
    { param[count++] = unconstify("/bin/sh");
      param[count++] = unconstify("-c");
    }
    param[count++] = unconstify(command); param[count] = NULL;
    if (do_read) my_close(fd_read[0]);
    if (esf & esfReadStdout)
    { if (fd_read[1] != 1)
      { if (!my_dup2(fd_read[1], 1))
        { dup2_failed: err = errno; errstr = _("can't setup file descriptors");
          goto child_error;
        }
      }
    }
    else my_close(1);
    if (esf & esfReadStderr)
    { if ( (fd_read[1] != 2) && (!my_dup2(fd_read[1], 2)) ) goto dup2_failed; }
    else my_close(2);
    if (do_write)
    { my_close(fd_write[1]);
      if ( (fd_write[0] != 0) && (!my_dup2(fd_write[0], 0)) ) goto dup2_failed;
    }
    else my_close(0);
    path = param[0];
    (void) execv(path, param);

    /* If we get here, execv() failed, so we send a little error text to the
       resource handler and exit the child process: */
    err = errno;
    my_spf(strbuf2, STRBUF_SIZE, &spfbuf2, _("can't execute \"%s\""), path);
    errstr = spfbuf2;
    child_error:
    my_spf(strbuf, STRBUF_SIZE, &spfbuf,
      _("retawq: execext-shell error: %s (error #%d, %s)"), errstr, err,
      my_strerror(err));
    if (esf & esfReadStderr) my_write_str(fd_read[1], spfbuf);
    else if (may_use_fd2()) my_write_str(2, spfbuf);
    debugmsg(spfbuf);
    _exit(1);
  }
  else /* parent process */
  { tResource* resource = NULL;
    tConnection *rconn = NULL, *wconn = NULL;
    if (do_read)
    { my_close(fd_read[1]); make_fd_cloexec(fd_read[0]);
      (void) make_fd_nonblocking(fd_read[0]); /* (just trying) */
      resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH,
        ( (esf & esfEnforceHtml) ? rckHtml : rckUnknown ), rsReading);
      resource->flags |= rfCannotReload;
      rconn = conn_create(fd_read[0], execext_shell_callback, resource, truE,
        NULL, rpExecextShell);
      rconn->flags |= cnfConnected; conn_set_read(rconn, 2);
    }
    if (do_write)
    { my_close(fd_write[0]); make_fd_cloexec(fd_write[1]);
      (void) make_fd_nonblocking(fd_write[1]); /* (just trying) */
      wconn = conn_create(fd_write[1], execext_shell_callback, resource,
        cond2boolean(resource != NULL), NULL, rpExecextShell);
      wconn->flags |= cnfConnected; wconn->x = 1; conn_set_write(wconn, 2);
      conn_set_writedata(wconn, writedata, data->writedata_size);
      data->writedata = NULL; /* detach */
#if CONFIG_DEBUG
      wconn->flags |= cnfDebugNodata;
#endif
    }
    if (resource != NULL) { resource->cconn = rconn; resource->dconn = wconn; }
  }
}

#endif /* #if OPTION_EXECEXT & 1 */


/* Non-local resources: HTTP */

#define IS_HTTP_WHITESPACE(ch) ( ((ch) == ' ') || ((ch) == '\t') )

static void skip_http_whitespace(const char** _ptr)
/* skips an HTTP header whitespace sequence at the current position */
{ const char* ptr = *_ptr;
  while (IS_HTTP_WHITESPACE(*ptr)) ptr++;
  *_ptr = ptr;
}

static one_caller void skip_http_line(const char** _ptr)
/* skips the current HTTP header line */
{ const char* ptr = *_ptr;
  while (1)
  { const char ch = *ptr++;
    if (ch == '\n') break;
    else if (ch == '\0') { ptr--; break; } /* "should not happen" */
  }
  /* FIXME: we don't look for a complete CRLF sequence; this is nice for
     local CGI scripts, but otherwise... */
  *_ptr = ptr;
}

#define http_close(res) conn_remove(&((res)->cconn))

static void parse_http_chunks(tResource* resource, const char* src,
  size_t count)
/* Welcome to the World's Worst Chunk-O-Mania! (Now you know the _real_
   meaning of "WWW" and ".com":-) A pseudo-code algorithm for this is available
   in RFC2616, 19.4.6, but of course we do it differently... */
{ size_t chunksize = resource->chunksize_left, usable_left, usedsize, add;
  char *deststart, *dest SHUT_UP_COMPILER(NULL);
  tChunkParsingState cps = resource->cps;
  tBoolean do_push, want_it_back;

  usable_left = usedsize = 0; do_push = falsE; want_it_back = truE;
  while (count-- > 0)
  { const char ch = *src++;
    switch (cps)
    { case cpsData: /* the most likely case */
        if (chunksize > 0) /* it's a "real data" character, store it */
        { if (usable_left <= 0) /* must get more room */
          { if (usedsize > 0)
            { resource_record(resource, deststart, usedsize); usedsize = 0; }
            resource_provide_room(resource, &deststart, &usable_left,
              resource->bytecount, 2);
            dest = deststart;
          }
          *dest++ = ch; usedsize++; usable_left--; chunksize--; do_push = truE;
        }
        else if (ch == '\r') cps = cpsCrlf2;
        else
        { bad_response:
          resource_set_error(resource, reResponse);
          goto stopnpush;
        }
        break;
      case cpsSize: /* parse chunk-size value */
        if (my_isdigit(ch))
        { add = (size_t) (ch - '0');
          calc:
          chunksize = 16 * chunksize + add;
          if (chunksize > 100 * 1024 * 1024) goto bad_response;
        }
        else if ( (ch >= 'a') && (ch <= 'f') )
        { add = (size_t) (ch - 'a' + 10); goto calc; }
        else if ( (ch >= 'A') && (ch <= 'F') )
        { add = (size_t) (ch - 'A' + 10); goto calc; }
        else if (ch == '\r') cps = cpsCrlf1;
        else
        { /* The server sent optional additional information (called
             "chunk-extension" in RFC2616, 3.6.1) we don't care about. */
          cps = cpsSkipSizeline;
        }
        break;
      case cpsCrlf1: /* reached end of chunk-size line */
        if (ch == '\n')
        { if (chunksize == 0) cps = cpsSkipTrailerC;
          else cps = cpsData;
#if CONFIG_DEBUG
          sprint_safe(debugstrbuf, "expecting HTTP chunk of size %d\n",
            chunksize);
          debugmsg(debugstrbuf);
#endif
        }
        else goto bad_response;
        break;
      case cpsSkipSizeline:
        if (ch == '\r') cps = cpsCrlf1;
        /* "else": skip further characters */
        break;
      case cpsCr2: /* reached end of chunk */
        if (ch == '\r') cps = cpsCrlf2;
        else goto bad_response;
        break;
      case cpsCrlf2: /* reached end of chunk */
        if (ch == '\n') cps = cpsSize;
        else goto bad_response;
        break;
      case cpsSkipTrailerA:
        if (ch == '\r') cps = cpsSkipTrailerB;
        break;
      case cpsSkipTrailerB:
        if (ch == '\n') cps = cpsSkipTrailerC;
        else cps = cpsSkipTrailerA;
        break;
      case cpsSkipTrailerC:
        if (ch == '\r') cps = cpsSkipTrailerD;
        else cps = cpsSkipTrailerA;
        break;
      case cpsSkipTrailerD:
        if (ch == '\n') /* finished receiving this resource */
        { debugmsg("finished chunked document\n");
          resource->state = rsComplete;
          stopnpush: http_close(resource);
#if OPTION_LOCAL_CGI
          if (resource->dconn != NULL) conn_remove(&(resource->dconn));
#endif
          do_push = truE; want_it_back = falsE; goto out;
        }
        else cps = cpsSkipTrailerA;
        break;
    }
  }
  out:
  if (usedsize > 0) resource_record(resource, deststart, usedsize);
  resource->cps = cps; resource->chunksize_left = chunksize;
  if (do_push) push_to_main_res(resource, boolean2bool(want_it_back));
}

static one_caller tMbsIndex do_lookup_hhps(const char* str)
{ my_binary_search(0, MAX_HHPS, strneqcase3(str, strHhps[idx], hhpslen[idx]),
    return(idx))
  /* (case-insensitivity of header field names: RFC2616, 4.2) */
}

static one_caller tHttpHeaderParsingState lookup_hhps(const char** _ptr)
{ const char* ptr = *_ptr;
  tMbsIndex idx = do_lookup_hhps(ptr);
  if (idx < 0) idx = 0; /* hhpsDontCare */
  else if (idx > 0) { *_ptr += hhpslen[idx]; skip_http_whitespace(_ptr); }
  return((tHttpHeaderParsingState) idx);
}

static one_caller tBoolean finish_http_header(tResource* resource,
  const char* header)
/* returns whether it worked */
{ const char* temp;
  int http_major, http_minor, http_status, value;
  tHttpHeaderParsingState hhps;
  tBoolean cannot_http11, ignore_contentlength, is_chunked,
    can_keep_alive SHUT_UP_COMPILER(falsE);
#if OPTION_COOKIES
  const tBoolean accept_cookies = cookie_accept_any(resource);
  const char* cookie_start = NULL;
#define handle_cookie \
  if (cookie_start != NULL) \
  { cookie_handle_text(resource, cookie_start, temp - 1, \
      cond2boolean(hhps == hhpsSetCookie2)); \
    cookie_start = NULL; \
  }
#else
#define handle_cookie /* nothing */
#endif
#if OPTION_LOCAL_CGI
  const tBoolean is_local_cgi = cond2boolean(resource->protocol == rpLocalCgi);
#endif
  char ch;

  /* Parse the HTTP header (RFC2616); at first we check the HTTP version
     (RFC2145) and status code. */

  temp = header; hhps = hhpsDontCare;
  ignore_contentlength = is_chunked = falsE;
#if OPTION_LOCAL_CGI
  if (is_local_cgi)
  { resource->server_status_code = 200; /* assume the script worked nicely */
    cannot_http11 = truE; /* CHECKME! */
    goto header_loop_cgientry; /* local scripts don't send an "HTTP/.." line */
  }
#endif
  if (!strneqcase(temp, "http/", 5))
  { bad: resource_set_error(resource, reResponse); return(falsE); }
  my_atoi(temp + 5, &http_major, &temp, 99);
  if (http_major > 1) goto bad; /* beyond our capabilities; blame server :-) */
  if (*temp != '.') goto bad;
  my_atoi(temp + 1, &http_minor, &temp, 99);
  skip_http_whitespace(&temp);
  my_atoi(temp, &http_status, &temp, 999);
  if ( (http_status < 100) || (http_status > 599) ) goto bad;
  resource->server_status_code = (tServerStatusCode) http_status;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "HTTP header: %d,%d,%d\n", http_major, http_minor,
    http_status);
  debugmsg(debugstrbuf);
#endif

  if ( (http_major < 1) || ( (http_major == 1) && (http_minor < 1) ) )
  { resource->actual_hppi->flags |= hppifCannotHttp11; cannot_http11 = truE; }
  else { resource->flags |= rfHttp11; cannot_http11 = falsE; }

  /* Set keep-alive default according to RFC2616, 8.1.2.1 par. 4: */
  if ( (http_major > 1) || ( (http_major == 1) && (http_minor >= 1) ) )
    can_keep_alive = truE;
  else can_keep_alive = falsE;

  header_loop:
  skip_http_line(&temp);
#if OPTION_LOCAL_CGI
  header_loop_cgientry:
#endif
  ch = *temp;
  if ( (ch == '\r') || (ch == '\0') ) goto finish_header; /* end of header */
  else if (IS_HTTP_WHITESPACE(ch))
  { /* next line of a multi-line header field (RFC2616, 4.2) */
    skip_http_whitespace(&temp);
  }
  else
  { handle_cookie
    hhps = lookup_hhps(&temp);
  }

  switch (hhps)
  {case hhpsDontCare: /* the most likely case first */
    break;
   case hhpsContentLength:
    if (ignore_contentlength) hhps = hhpsDontCare;
    else
    { my_atoi(temp, &value, &temp, 1000 * 1024 * 1024);
      resource->nominal_contentlength = value;
    }
    break;
   case hhpsContentType:
    /* (case-insensitivity of media types: RFC2616, 3.7) */
    if (strneqcase(temp, "text/html", 9)) resource->kind = rckHtml;
    else if (strneqcase(temp, "text/plain", 10)) resource->kind = rckPlainText;
    else resource->kind = rckUnknown;
    break;
   case hhpsTransferEncoding:
    ignore_contentlength = truE; /* RFC2616, 4.4(.3|end) plus HTTP errata */
    if ( (!cannot_http11) && (strneqcase(temp, "chunked", 7)) )
      is_chunked = truE;
    break;
   case hhpsConnection:
    if (strneqcase(temp, "keep-alive", 10)) can_keep_alive = truE;
    else if (strneqcase(temp, strClose, 5)) can_keep_alive = falsE;
    break;
   case hhpsLocation:
    { size_t len = 0;
      const char* ptr = temp;
      __dealloc(resource->redirection);
      while ( (*ptr != '\0') && (*ptr != '\r') ) { ptr++; len++; }
      if (len > 0) resource->redirection = my_strndup(temp, len);
      else resource->redirection = NULL;
    }
    break;
   case hhpsServer: hppi_set_software_id(resource->textual_hppi, temp); break;
#if OPTION_COOKIES
   case hhpsSetCookie: case hhpsSetCookie2:
    if (accept_cookies) cookie_start = temp;
    break;
#endif
#if OPTION_LOCAL_CGI
   case hhpsStatus:
    if (is_local_cgi)
    { int i;
      my_atoi(temp, &i, &temp, 999);
      if ( (i < 100) || (i > 599) ) goto bad;
      resource->server_status_code = (tServerStatusCode) i;
    }
    break;
#endif
  }
  goto header_loop;

  finish_header:
  handle_cookie
  if (ignore_contentlength)
    resource->nominal_contentlength = UNKNOWN_CONTENTLENGTH;
  if (resource->kind == rckUnknown)
    resource->kind = filesuffix2rck(resource->path);
  if (is_chunked)
  { resource->flags |= rfHttpChunked;
    resource->chunksize_left = 0;
    resource->cps = cpsSize;
  }
  if ( (resource->nominal_contentlength == UNKNOWN_CONTENTLENGTH) ||
       (!can_keep_alive) )
    resource->cconn->flags |= cnfDontReuse;
  debugmsg("finished HTTP header:\n"); debugmsg(header); debugmsg(strNewline);
  return(truE);
#undef handle_cookie
}

static one_caller void http_calc_version(tResource* resource,
  /*@out@*/ const char** _version)
{ const char* version = NULL;
  const tConfigProtocolVersion* v = config.http_version;

  if (v != NULL)
  { const char* hostname = resource2textual_host(resource)->hostname;
    tPortnumber resport = resource->textual_portnumber;
    while (v != NULL)
    { tPortnumber hp = v->hosts_portnumber;
      if ( (hp == 0) || (hp == resport) )
      { const char* pattern = v->hosts_pattern;
        if ( (pattern != NULL) && (my_pattern_matcher(pattern, hostname)) )
        { version = v->protstr; break; } /* (might be NULL, no problem) */
      }
      v = v->next;
    }
  }

  if (version == NULL)
  { tBoolean cannot_http11 =
      cond2boolean(resource->actual_hppi->flags & hppifCannotHttp11);
    version = (cannot_http11 ? str1dot0 : str1dot1);
  }
  *_version = version;
}

static one_caller void http_setup_reqstr(tResource* resource)
{ const tConfigProxy* proxy = resource->proxy;
  tResourceFlags rf = resource->flags;
  tPortnumber portnumber = ntohs(resource->textual_portnumber);
    /* (host byte order!) */
  const tBoolean use_post = cond2boolean(rf & rfPost);
  const char *uri = resource->uri, *path = resource->path, *spfbuf_uri,
    *query = resource->query, *post = (use_post ? null2empty(resource->post) :
    strEmpty), *hostname = resource2textual_host(resource)->hostname,
    *auth_user, *auth_pass, *cookies = cookie_collect(resource), *http_version,
    *spfbuf_proxyauth = strEmpty;
  const tBoolean use_query = cond2boolean((query!=NULL) && (query[0]!='\0')),
    use_proxy = cond2boolean((proxy != NULL) && (proxy->proxy_hostname!=NULL));
  char *spfbuf;

  http_calc_version(resource, &http_version);

  if (use_post)
  { sprint_safe(strbuf2, "Content-Type: %s\r\nContent-Length: %d\r\n",
      strAxwfu, strlen(post));
    my_strdedup(resource->post, post);
    resource->flags |= rfPost; rf = resource->flags;
  }

  if ( (use_proxy) && ( (auth_user = proxy->auth_user) != NULL ) &&
       ( (auth_pass = proxy->auth_pass) != NULL ) )
  { char* spfbuf_enc;
    const char* enc;
    my_spf(strbuf, STRBUF_SIZE, &spfbuf_enc, strPercsColonPercs, auth_user,
      auth_pass);
    enc = base64_encode(spfbuf_enc);
    my_spf_cleanup(strbuf, spfbuf_enc);
    my_spf(NULL, 0, &spfbuf, "Proxy-Authorization: Basic %s\r\n", enc);
    memory_deallocate(enc);
    spfbuf_proxyauth = my_spf_use(spfbuf);
  }

  if (use_proxy) spfbuf_uri = uri; /* RFC2616, 5.1.2 */
  else if (!use_query) spfbuf_uri = path; /* most likely case */
  else
  { /* RFC2616 (3.2.2) plus HTTP/1.1 erratum "URI includes query" */
    my_spf(NULL, 0, &spfbuf, "%s?%s", path, query);
    spfbuf_uri = my_spf_use(spfbuf);
  }
  my_spf(NULL, 0, &spfbuf,
/*A*/ "%s %s HTTP/%s\r\n"
/*B*/ "Host: %s:%d\r\n"
/*C*/ "User-Agent: %s\r\n"
/*D*/ "%s" /* proxy authentication if configured */
/*E*/ "%s" /* cache-control stuff if desired */
/*F*/ "%s" /* cookies if any */
/*G*/ "%s" /* content-type/-length for POST method */
      "Accept: */*\r\n"
/*H*/ "Accept-Language: %s\r\n"
      "Accept-Charset: iso-8859-1\r\n"
/*I*/ "\r\n%s", /* end of header; POST data if any */

/*A*/ (use_post ? strPost : strGet), spfbuf_uri, http_version,
/*B*/ hostname, portnumber,
/*C*/ config.user_agent,
/*D*/ spfbuf_proxyauth,
/*E*/ ( (rf & rfIsEnforced) ? /* (RFC2616, 14.32 and 14.9) */
       "Pragma: no-cache\r\nCache-Control: no-cache\r\n" : strEmpty ),
/*F*/ cookies,
/*G*/ (use_post ? strbuf2 : strEmpty),
/*H*/ config.languages,
/*I*/ post);

#if OPTION_TLS
  if ( (resource->protocol == rpHttps) && (proxy != NULL) )
  { char* spfbuf2;
    my_spf(NULL, 0, &spfbuf2,
      "CONNECT %s:%d HTTP/%s\r\nHost: %s:%d\r\n%s\r\n%s", /* RFC2817, 5.2 */
      hostname, portnumber, http_version, hostname, portnumber,
      spfbuf_proxyauth, spfbuf);
    memory_deallocate(spfbuf);
    spfbuf = my_spf_use(spfbuf2);
  }
#endif

  conn_set_writestr(resource->cconn, my_spf_use(spfbuf));
  my_spf_cleanup(strEmpty, spfbuf_proxyauth);
  if ( (spfbuf_uri != uri) && (spfbuf_uri != path) )
    memory_deallocate(spfbuf_uri);
  cookie_collect_cleanup(cookies);
}

typedef struct
{ tRpsdRemover remover;
  char* header; /* for the HTTP header */
  size_t usable, used;
} tHttpRpsd;

static void http_rpsd_remove(tRpsdGeneric* _rpsd)
{ tHttpRpsd* rpsd = (tHttpRpsd*) _rpsd;
  __dealloc(rpsd->header);
}

static void http_rpsd_prepare(tResource* resource)
{ resource->rpsd = memory_allocate(sizeof(tHttpRpsd));
  resource->rpsd->remover = http_rpsd_remove;
}

#if CONFIG_DEBUG
static void http_read_debug(const tResource* resource, const char* what,
  int err, const char* buf)
{ sprint_safe(debugstrbuf, "http_read(%p, %s): %d,%d,%d\n", resource, what,
    err, errno, cond2bool(resource->flags & rfHttpChunked));
  debugmsg(debugstrbuf);
  if (err > 0)
  { __debugmsg(buf, err);
    if (buf[err - 1] != '\n') debugmsg(strNewline);
  }
}
#else
#define http_read_debug(a, b, c, d) do { } while (0)
#endif

static void http_read(tResource* resource)
/* This function is used for rpHttp, rpHttps and rpLocalCgi. */
{ tConnection* conn = resource->cconn;
  int err;
  tResourceFlags rf;
  tBoolean is_chunked, want_it_back = truE;
  tServerStatusCode ssc;
  tTlHeaderState ths;
  char* dest;
  size_t size, ncl;

  ths = resource->tlheaderstate;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "http_read(%p): %d\n", resource, ths);
  debugmsg(debugstrbuf);
#endif
  if (ths >= 0) /* reading the header */
  { tHttpRpsd* rpsd = (tHttpRpsd*) resource->rpsd;
    char *header = rpsd->header, *temp;
    size_t usable = rpsd->usable, used = rpsd->used, count;
    if (usable <= used)
    { if (usable >= HIGHEST_OCBS / 2) /* big header - attack? */
      { resource_set_error(resource, reResponse); goto do_stop_reading; }
      else if (usable <= 0) usable = 1024;
      else usable <<= 1;
      rpsd->usable = usable;
      rpsd->header = header = memory_reallocate(header, usable);
    }
    temp = header + used;
    conn_read(err, conn, temp, usable - used, return);
    http_read_debug(resource, "header", err, temp);
    if (err <= 0) goto stop_reading;
    rpsd->used += err; count = err;
    while (count-- > 0)
    { const char ch = *temp++;
      switch (ths)
      { case 0: if (ch == '\r') { ths = 1; } break;
        case 1:
          if (ch == '\n') ths = 2;
          else if (ch == '\r') { /* ths = 1; */ }
          else ths = 0;
          break;
        case 2: ths = ( (ch == '\r') ? 3 : 0 ); break;
        case 3:
          if (ch == '\n') /* found the end of the HTTP header */
          { *(temp - 1) = '\0'; /* terminate the tough text trickily :-) */
            if (!finish_http_header(resource, header)) goto do_stop_reading;
            ssc = resource->server_status_code;
            if ( (ssc >= 100) && (ssc <= 199) ) /* informational header */
            { rpsd->used -= (temp - header);
              if (count > 0) { my_memcpy(header, temp, count); temp = header; }
              ths = 0; goto next_char;
            }
            resource->state = rsReading; ths = -1; /* done with headerisms */
            if (count > 0)
            { if (resource->flags & rfHttpChunked)
              { parse_http_chunks(resource, temp, count);
                resource->tlheaderstate = ths;
                return;
              }
              else resource_collect(resource, temp, count);
            }
            goto header_done;
          }
          else if (ch == '\r') ths = 1;
          else ths = 0;
          break;
      }
      next_char: {}
    }
    header_done: resource->tlheaderstate = ths; goto prefinish;
  }

  rf = resource->flags;
  is_chunked = cond2boolean(rf & rfHttpChunked);
  if ( (is_chunked) && (resource->cps != cpsData) )
  { /* only read a few bytes, they "should" mean the next chunk-size */
    char temp[15];
    conn_read(err, conn, temp, sizeof(temp), return);
    http_read_debug(resource, "chunky", err, temp);
    if (err > 0) { parse_http_chunks(resource, temp, err); return; }
    else /* error or EOF */
    { stop_reading:
      if (err == 0) resource->state = rsComplete;
      else resource_set_error(resource, reServerClosed);
      do_stop_reading: want_it_back = falsE; http_close(resource);
#if OPTION_LOCAL_CGI
      if (resource->dconn != NULL) conn_remove(&(resource->dconn));
#endif
    }
    goto finish;
  }

  { size_t desired_size, bytecount = resource->bytecount;
    unsigned char rprflags;
    ncl = resource->nominal_contentlength;
    if ( (ncl != UNKNOWN_CONTENTLENGTH) && (bytecount < ncl) )
    { /* We know how much to expect, so we try (more or less) to slurp it all
         at once in order to avoid unnecessary read() calls. */
      ncl -= bytecount; /* subtract what was already read */
      if (ncl > HIGHEST_OCBS) ncl = HIGHEST_OCBS; /* (just a "sane" limit) */
      desired_size = ncl; rprflags = 0;
    }
    else { desired_size = bytecount; rprflags = 2; }
    resource_provide_room(resource, &dest, &size, desired_size, rprflags);
  }

  if (is_chunked)
  { size_t max = resource->chunksize_left;
    if (size > max) size = max; /* read at most until end of chunk */
  }
  conn_read(err, conn, dest, size, return);
  http_read_debug(resource, "data", err, dest);
  if (err <= 0) goto stop_reading;
  resource_record(resource, dest, err);
  if (is_chunked)
  { resource->chunksize_left -= err;
    if (resource->chunksize_left == 0) resource->cps = cpsCr2;
  }
#if CONFIG_DEBUG
  resource->version++;
#endif

  prefinish:
  ncl = resource->nominal_contentlength;
  if ( (ncl != UNKNOWN_CONTENTLENGTH) && (resource->bytecount >= ncl) )
  { /* The server sent us all the announced data, so we're done: */
    err = 0; goto stop_reading;
  }
  finish:
  if ( (!want_it_back) || (resource->state == rsReading) )
    push_to_main_res(resource, boolean2bool(want_it_back));
}

static void http_callback(tConnection* conn, tConnCbEventKind ccek)
{ tResource* resource = (tResource*) (conn->data);
  switch (ccek)
  { case ccekStart:
      conn_set_write(conn, 2);
      if (conn->flags & cnfConnected) goto do_setup;
      break;
    case ccekConnectWorked:
      conn->flags |= cnfConnected; resource->handshake = rchConnected;
      do_setup:
      http_rpsd_prepare(resource); http_setup_reqstr(resource);
#if OPTION_TLS
      if ( (resource->protocol == rpHttps) && (resource->proxy == NULL) )
      { tBoolean got_session = tls_session_init(conn,
          resource2textual_host(resource)->hostname);
        if (!got_session)
        { resource_set_error(resource, reTls); goto stopnpush; }
      }
#endif
      push_to_main_res(resource, 1);
      break;
    case ccekConnectFailed:
      resource_set_error(resource, connect_err2error(conn_err));
      stopnpush: http_close(resource); push_to_main_res(resource, 0); break;
    case ccekRead:
#if OPTION_TLS
      if (conn->flags & cnfTlsHandshaking)
      { tdh: if (tls_do_handshaking(conn) == 0) goto stopnpush;
        break;
      }
#endif
      http_read(resource); break;
    case ccekWrite:
#if OPTION_TLS
      if (conn->flags & cnfTlsHandshaking) goto tdh;
#endif
      switch (conn_write_writedata(conn))
      { case 0:
          resource_set_error(resource, reServerClosed); goto stopnpush;
          /*@notreached@*/ break;
        case 1: break; /* keep writing */
        case 2: conn_set_read(conn, 2); break; /* all written, start reading */
      }
      break;
    default: conn_bug(conn, ccek); break;
  }
}


/* Non-local resources: FTP */

#if CONFIG_FTP

my_enum1 enum
{ frfNone = 0, frfSuffsnip = 0x01, frfDidDirstart = 0x02
} my_enum2(unsigned char) tFtpRpsdFlags;

typedef struct
{ tRpsdRemover remover;
  const char* reply_snippet; /* e.g. part of EPSV reply */
#if CONFIG_FTPDIR > 0
  char* text; /* buffer for FTP directory listing beautification */
  size_t usable, used;
#endif
#if OPTION_TLS
  tFtpTlsMethod ftm;
#endif
  tFtpRpsdFlags flags;
} tFtpRpsd;

static void ftp_rpsd_remove(tRpsdGeneric* _rpsd)
{ tFtpRpsd* rpsd = (tFtpRpsd*) _rpsd;
  __dealloc(rpsd->reply_snippet);
#if CONFIG_FTPDIR > 0
  __dealloc(rpsd->text);
#endif
}

#if CONFIG_FTPDIR > 0

static one_caller void ftp_data_handle_dirline(tResource* resource, char* text)
/* adds one line to an FTP directory listing */
{ const char *filename, *htmlfilename, *info, *htmlinfo, *path;
  char *spfbuf, *ws;
  tBoolean found_ws, need_pathslash;
  size_t pathlen;

  if (strneqcase(text, "total", 5)) /* game... */
  { /* Some very funny servers send a "total <number>" line (probably from a
       dumb "ls -l" shell command) at the beginning of the directory listing,
       and the number shouldn't become a link. */
    const char* temp = text + 5;
    if ( (*temp == ' ') || (*temp == '\t') ) /* ...set... */
    { while ( (*temp == ' ') || (*temp == '\t') ) temp++;
      if (my_isdigit(*temp))
      { while (my_isdigit(*temp)) temp++;
        if (*temp == '\0') /* ...and match */
        { my_spf(strbuf, STRBUF_SIZE, &spfbuf, "<br>%s\n", text);
          goto do_collect;
        }
      }
    }
  }

  found_ws = falsE; ws = text + strlen(text) - 1;
  while (ws >= text)
  { char ch = *ws;
    if ( (ch == ' ') || (ch == '\t') ) { found_ws = truE; break; }
    ws--;
  }

  if (found_ws)
  { /* The server sent us some text before the actual file/dir name; we don't
       care what it is (most likely some access rights like "rw-r--r--"), we
       just make sure it goes _before_ the link: */
    *ws++ = '\0'; filename = ws; info = text;
  }
  else { filename = text; info = strEmpty; }

  htmlfilename = htmlify(filename); htmlinfo = htmlify(info);
  path = resource->path; pathlen = strlen(path);
  need_pathslash = cond2boolean((pathlen <= 0) || (path[pathlen - 1] != '/'));
  my_spf(strbuf, STRBUF_SIZE, &spfbuf,
    "<br>%s%s<a href=\"%s://%s:%d%s%s%s\">%s</a>\n", htmlinfo,
    ( (*htmlinfo != '\0') ? strSpace : strEmpty ),
    rp_data[resource->protocol].scheme,
    resource2textual_host(resource)->hostname,
    ntohs(resource->textual_portnumber), path, (need_pathslash ? strSlash :
    strEmpty), filename, htmlfilename);
  htmlify_cleanup(filename, htmlfilename); htmlify_cleanup(info, htmlinfo);
  do_collect:
  resource_collect_str(resource, spfbuf);
  my_spf_cleanup(strbuf, spfbuf);
}

#endif /* #if CONFIG_FTPDIR > 0 */

#if CONFIG_DEBUG
static void debug_ftp_data_llread(tConnection* conn, int result,
  const char* debugstr)
{ sprint_safe(debugstrbuf, "ftp_data_llread(%s): fd=%d,result=%d\n",
    debugstr, conn->fd, result);
  debugmsg(debugstrbuf);
}
#else
#define debug_ftp_data_llread(conn, result, debugstr) do { } while (0)
#endif

#define ftp_data_llread(result, conn, buf, bufsize, debugstr) /* low-level */ \
  do \
  { conn_read(result, conn, buf, bufsize, return(3)); \
    debug_ftp_data_llread(conn, result, debugstr); \
  } while (0)

static void ftp_data_dirstart(tResource* resource)
{ const char* path = resource->path;
  char* spfbuf;
  my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("Contents of FTP directory \"%s\""),
    path);
#if CONFIG_FTPDIR > 0
  resource->kind = rckHtml;
  resource_collect_title(resource, spfbuf);
#else
  resource_collect_str(resource, spfbuf);
  resource_collect_str(resource, "\n\n");
#endif
  my_spf_cleanup(strbuf, spfbuf);
#if CONFIG_FTPDIR > 0
  if (calc_parentpath(path, strbuf2, STRBUF_SIZE))
  { my_spf(strbuf, STRBUF_SIZE, &spfbuf,
      "<p><a href=\"%s://%s:%d%s\">%s</a></p>\n",
      rp_data[resource->protocol].scheme,
      resource2textual_host(resource)->hostname,
      ntohs(resource->textual_portnumber), strbuf2, _(strParentDirectory));
    resource_collect_str(resource, spfbuf);
    my_spf_cleanup(strbuf, spfbuf);
  }
#endif
}

static one_caller unsigned char ftp_data_read(tResource* resource)
/* return value: 0=error, 1=EOF, 2=proceed, 3=interrupted (TLS only) */
{ tFtpRpsd* rpsd = (tFtpRpsd*) resource->rpsd;
  const tBoolean is_dir = cond2boolean(resource->flags & rfFtpDir),
    did_dirstart = cond2boolean(rpsd->flags & frfDidDirstart);
  tConnection* conn = resource->dconn;
  int err;
#if CONFIG_FTPDIR > 0
  if (is_dir)
  { char *text, *temp, *remainder, *end;
    size_t usable, used;
    tBoolean near_r;

    /* read */
    text = rpsd->text; usable = rpsd->usable; used = rpsd->used;
    if (usable <= used)
    { usable += 4096; rpsd->usable = usable;
      text = rpsd->text = memory_reallocate(text, usable);
    }
    ftp_data_llread(err, conn, text + used, usable - used, "dir");
    if (err <= 0) /* error or EOF */
    { if (rpsd->flags & frfDidDirstart)
        resource_collect_str(resource, strEndBodyHtml);
      goto stop_reading;
    }
    if (!did_dirstart)
    { ftp_data_dirstart(resource); rpsd->flags |= frfDidDirstart; }
    used += err; rpsd->used = used;

    /* look for CRLF */
    temp = remainder = text; end = text + used - 1; near_r = falsE;
    while (temp <= end)
    { const char ch = *temp++;
      if (ch == '\r') near_r = truE;
      else
      { if ( (ch == '\n') && (near_r) ) /* found a CRLF */
        { *(temp - 2) = '\0'; /* overwrites the "\r" */
          ftp_data_handle_dirline(resource, remainder);
          remainder = temp;
        }
        near_r = falsE;
      }
    }

    /* finish */
    if (remainder > text) /* did something */
    { if (remainder > end) rpsd->used = 0; /* did everything */
      else
      { const char* src = remainder;
        char* dest = text;
        while (src <= end) *dest++ = *src++;
        rpsd->used = rpsd->used - (remainder - text);
      }
    }
  }
  else /* isn't a directory */
#endif /* #if CONFIG_FTPDIR > 0 */
  { char* dest;
    size_t size;
    resource_provide_room(resource, &dest, &size, resource->bytecount, 2);
    ftp_data_llread(err, conn, dest, size, "file");
    if (err <= 0)
    { stop_reading:
      if (err == 0) { resource->state = rsComplete; return(1); }
      else { resource_set_error(resource, conn_err2error(errno)); return(0); }
    }
    else
    { if ( (is_dir) && (!did_dirstart) )
      { ftp_data_dirstart(resource); rpsd->flags |= frfDidDirstart; }
      resource_record(resource, dest, err);
    }
  }
  return(2); /* all fine */
}

static void ftp_data_callback(tConnection* conn, tConnCbEventKind ccek)
{ tResource* resource = (tResource*) (conn->data);
  switch (ccek)
  { case ccekStart:
      if (conn->flags & cnfConnected) goto do_setup;
      else conn_set_write(conn, 2); /* must "wait" until connected */
      break;
    case ccekConnectWorked:
      conn->flags |= cnfConnected;
      do_setup:
#if OPTION_TLS
      if ( (resource->protocol == rpFtps) &&
           (!(resource->flags & rfFtpsDataClear)) )
      { tBoolean got_session = tls_session_init(conn,
          resource2textual_host(resource)->hostname);
        if (!got_session)
        { resource_set_error(resource, reTls); goto stopnpush; }
      }
      else
#endif
      { conn_set_read(conn, 2); }
      resource->state = rsReading; push_to_main_res(resource, 1);
      break;
    case ccekConnectFailed:
      resource_set_error(resource, connect_err2error(conn_err));
      stopnpush:
      conn_remove(&(resource->dconn));
      conn_remove(&(resource->cconn)); /* CHECKME! */
      push_to_main_res(resource, 0);
      break;
#if OPTION_TLS
    case ccekWrite:
      if (!conn_using_tls(conn)) goto is_buggy;
      /* "else": */ /*@fallthrough@*/
#endif
    case ccekRead:
#if OPTION_TLS
      if (conn->flags & cnfTlsHandshaking)
      { if (tls_do_handshaking(conn) == 0) goto stopnpush;
        break;
      }
#endif
      switch (ftp_data_read(resource))
      { case 1:
          resource->server_status_code = 226; /* CHECKME: that's dirty! */
          /*@fallthrough@*/
        case 0: goto stopnpush; /*@notreached@*/ break;
        case 2: push_to_main_res(resource, 1); break;
        /* case 3: TLS operation was interrupted, just proceed */
      }
      break;
    default:
#if OPTION_TLS
      is_buggy:
#endif
      conn_bug(conn, ccek); break;
  }
}

static const char* ftp_prepare_snippet(const tResource* resource)
{ const char *retval = NULL, /* assume failure */
    *ptr = ((tFtpRpsd*)(resource->rpsd))->reply_snippet;
  unsigned char count;
  if (ptr == NULL) goto out;
  for (count = 0; count <= 2; count++)
  { const char ch = *ptr++;
    if (!my_isdigit(ch)) goto out; /* "can't happen" */
  }
  if ( (*ptr != ' ') && (*ptr != '-') ) goto out; /* "can't happen" */
  ptr++; retval = ptr;
  out:
  return(retval);
}

static one_caller tBoolean ftp_setup_dconn(tResource* resource,
  /*@out@*/ tResourceError* error)
/* returns whether it worked */
{ tCachedHostInformation* hostinfo = resource2actual_host(resource);
  const char *ptr = ftp_prepare_snippet(resource), *ptr2, *path;
  char ch;
  tConnection* conn;
  tBoolean is_epsv = cond2boolean(resource->flags & rfFtpTryingEpsv),
    is_connected;
  tSockaddr addr;
  size_t addrlen;
  int data_fd, err;

  if (ptr == NULL) { bad_ftp: *error = reResponse; failed: return(falsE); }
  ptr2 = my_strchr(ptr, '(');
  if (ptr2 != NULL)
  { ptr2++;
    if ( (*ptr2 != '\0') && ( (is_epsv || my_isdigit(*ptr2)) ) )
    { ptr = ptr2; goto snipfin; }
  }
  if (is_epsv) goto bad_ftp; /* invalid server reply format */
  while ( (ch = *ptr++) != '\0' )
  { if (my_isdigit(ch)) { ptr--; goto snipfin; } }
  goto bad_ftp;

  snipfin:
  if (is_epsv)
  { const char dc = *ptr++; /* "delimiting character" */
    int _portnumber;
    if ( (*ptr != dc) || (*(ptr + 1) != dc) ) goto bad_ftp;
    ptr += 2; if (!my_isdigit(*ptr)) goto bad_ftp;
    my_atoi(ptr, &_portnumber, &ptr, 99999);
#if CONFIG_DEBUG
    sprint_safe(debugstrbuf, "ftp_setup_dconn(): port %d\n", _portnumber);
    debugmsg(debugstrbuf);
#endif
    if ( (_portnumber < 0) || (_portnumber > 65535) || (*ptr != dc) )
      goto bad_ftp;
    addr = hostinfo->addrinfo;
    set_portnumber(&addr, hostinfo->address_family,
      htons((tPortnumber) _portnumber));
    addrlen = hostinfo->addrinfolen;
  }
  else /* used PASV */
  { int a, b, c, d, p, q;
    unsigned int hostlong;
    unsigned short int hostshort;
    PREPARE_SSCANF_FORMAT(format, 30, "%d,%d,%d,%d,%d,%d")
    if (sscanf(ptr, format, &a, &b, &c, &d, &p, &q) != 6) goto bad_ftp;
    if ( (a < 0) || (a > 255) || (b < 0) || (b > 255) || (c < 0) || (c > 255)
      || (d < 0) || (d > 255) || (p < 0) || (p > 255) || (q < 0) || (q > 255) )
    { goto bad_ftp; }
    hostlong = (a << 24) | (b << 16) | (c << 8) | (d);
    hostshort = (p << 8) | (q);
    my_memclr_var(addr);
    ((struct sockaddr_in*)&addr)->sin_family = AF_INET;
    ((struct sockaddr_in*)&addr)->sin_addr.s_addr = htonl(hostlong);
    set_portnumber(&addr, AF_INET, htons(hostshort));
    addrlen = sizeof(struct sockaddr_in);
  }

  data_fd = create_socket(hostinfo->address_family);
  if (data_fd < 0)
  { *error = ( (data_fd == -2) ? reTmofd : reSocket ); goto failed; }

  err = my_do_connect(data_fd, &addr, addrlen);
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "FTP data connect: %d,%d,%d,%d\n", data_fd, err,
    errno, ((struct sockaddr_in*)&addr)->sin_family);
  debugmsg(debugstrbuf);
#endif
  if (err == 0) is_connected = truE; /* unlikely when non-blocking */
  else if (errno == EINPROGRESS) is_connected = falsE;
  else { *error = connect_err2error(errno); my_close(data_fd); goto failed; }

  if ((path = resource->path) != NULL) resource->kind = filesuffix2rck(path);
  if (resource->dconn != NULL) /* "should not happen" */
    conn_remove(&(resource->dconn));
  conn = resource->dconn = conn_create(data_fd, ftp_data_callback, resource,
    truE, NULL, resource->protocol);
  if (is_connected) conn->flags |= cnfConnected;
  conn_callback(conn, ccekStart);
  return(truE);
}

/* File Transfer Protocol server status code equality check; RFC1123 says: "A
   User-FTP SHOULD generally use only the highest-order digit of a 3-digit
   reply code for making a procedural decision, to prevent difficulties when a
   Server-FTP uses non-standard reply codes." CHECKME: should we really? */
#if 1
#define FTPSSCEQ(value1, value2) ((value1) == (value2))
#else
#define FTPSSCEQ(value1, value2) (((value1) / 100) == ((value2) / 100))
#endif

#if CONFIG_USER_QUERY

static my_inline tBoolean resource_ask_username(tResource* resource,
  tUserQueryCallback callback, unsigned char flags)
{ return(resource_ask_anything(resource, callback, mifUsername, flags));
}

static my_inline tBoolean resource_ask_password(tResource* resource,
  tUserQueryCallback callback, unsigned char flags)
{ return(resource_ask_anything(resource, callback, mifPassword, flags));
}

static void ftp_user_callback(tUserQuery*); /* prototype */

#endif

#define FCPNR(x) do { retval = (x); goto out; } while (0)

static unsigned char ftp_control_prepare_next(tResource* resource,
  tRequestContactHandshake next, char** _spfbuf)
/* prepares the next (login-related) FTP handshaking step; return value: 0=bug,
   1=login failure, 2=suspended, 3=switch to writing */
{ unsigned char retval = 0;
  const char *temp;
  resource->handshake = next;
  switch (next)
  { case rchFtpUser:
      if ( (temp = resource->username) != NULL )
      { my_spf(NULL, 0, _spfbuf, strNetcmdUser, temp); stw: FCPNR(3); }
#if CONFIG_USER_QUERY
      else if (resource_ask_username(resource, ftp_user_callback, 0))
      { susp: FCPNR(2); }
#endif
      else { login_failure: FCPNR(1); }
      /*@notreached@*/ break;
    case rchFtpPassword:
      if ( (temp = resource->password) != NULL)
      { my_spf(NULL, 0, _spfbuf, strNetcmdPass, temp); goto stw; }
#if CONFIG_USER_QUERY
      else if (resource_ask_password(resource, ftp_user_callback, 0))
        goto susp;
#endif
      else goto login_failure;
      /*@notreached@*/ break;
    default: FCPNR(0); /*@notreached@*/ break;
  }
  out:
  return(retval);
}

#undef FCPNR

#if OPTION_TLS

static one_caller tFtpTlsMethod calc_ftp_tls_method(const tResource* resource)
{ const tConfigFtpTlsMethod* cftm = config.ftp_tls_method;
  tFtpTlsMethod retval = ftmAutodetect;
  if (cftm != NULL)
  { const char* hostname = resource2textual_host(resource)->hostname;
    tPortnumber resport = resource->textual_portnumber;
    while (cftm != NULL)
    { tPortnumber hp = cftm->hosts_portnumber;
      if ( ( (hp == 0) || (hp == resport) ) &&
           (my_pattern_matcher(cftm->hosts_pattern, hostname)) )
      { retval = cftm->ftm; break; }
      cftm = cftm->next;
    }
  }
  if ( (retval == ftmAutodetect) &&
       (resource->textual_hppi->portnumber == htons(990)) )
    retval = ftmTls; /* (earliest autodetection:-) */
  return(retval);
}

#define ftps_ftm(resource) (((tFtpRpsd*)((resource)->rpsd))->ftm)

static my_inline tBoolean ftp_control_tls_session_init(tConnection* conn,
  tResource* resource)
{ return(tls_session_init(conn, resource2textual_host(resource)->hostname));
}

#if CONFIG_USER_QUERY
static void ftps_dataclear_callback(tUserQuery*); /* prototype */
#endif

#endif /* #if OPTION_TLS */

static unsigned char ftp_control_handshake(tResource* resource,
  tResourceError* _ftp_error)
/* return value: 0=error, 1=continue reading, 2=switched to writing,
   3=complete, 4=suspended */
{ tRequestContactHandshake handshake = resource->handshake, next = rchUnknown;
  tServerStatusCode status = resource->server_status_code;
  tConnection* conn = resource->cconn;
  char* spfbuf;

  if (resource->flags & rfFtpShakyAny)
  { tResourceFlags rf = resource->flags;
    resource->flags &= ~rfFtpShakyAny;
    if (rf & rfFtpShakyNext) { next = handshake; goto do_shaky_next; }
#if OPTION_TLS
    else if (rf & rfFtpsShakyProt) goto do_shaky_prot;
#endif
    else goto hs_bug;
  }

  switch (handshake)
  { case rchConnected:
      if (FTPSSCEQ(status, 220)) /* "server ready", "awaiting input" */
      { const char* ptr = ftp_prepare_snippet(resource);
        if (ptr != NULL) hppi_set_software_id(resource->actual_hppi, ptr);
        conn_set_dissolver(conn, quitcmd_dissolver);
#if OPTION_TLS
        if (resource->protocol == rpFtps)
        { switch (ftps_ftm(resource))
          { case ftmAutodetect: case ftmAuthTls: case ftmAuthTlsDataclear:
              handshake = rchFtpTlsAuthTls; spfbuf = my_strdup("AUTH TLS\r\n");
              goto switch_to_writing; /*@notreached@*/ break;
            case ftmAuthSsl:
              prepare_auth_ssl:
              handshake = rchFtpTlsAuthSsl; spfbuf = my_strdup("AUTH SSL\r\n");
              goto switch_to_writing; /*@notreached@*/ break;
          }
        }
#endif
        next = rchFtpUser;
      }
      break;

#if OPTION_TLS
    case rchFtpTlsAuthTls:
      if (FTPSSCEQ(status, 234))
      { tBoolean got_session;
        spfbuf = my_strdup("PBSZ 0\r\n");
        conn_set_writestr(conn, my_spf_use(spfbuf));
        got_session = ftp_control_tls_session_init(conn, resource);
        if (!got_session) { *_ftp_error = reTls; goto failed; }
        resource->handshake = rchFtpTlsPbsz;
        return(1); /* not "goto switch_to_writing;" here, this is special */
      }
      else if ( (status / 100 == 5) && (ftps_ftm(resource) == ftmAutodetect) )
      { /* The server didn't like "AUTH TLS", but maybe likes "AUTH SSL". */
        goto prepare_auth_ssl;
      }
      break;
    case rchFtpTlsAuthSsl:
      if ( (FTPSSCEQ(status, 234)) || (FTPSSCEQ(status, 334)) )
      { /* (The 334 is technically wrong, but some buggy servers use(d) it.) */
        tBoolean got_session;
        conn->flags |= cnfWantToWrite; /* CHECKME: that's dirty! */
        got_session = ftp_control_tls_session_init(conn, resource);
        if (!got_session) { *_ftp_error = reTls; goto failed; }
        return(1); /* not "goto switch_to_writing;" here, this is special */
      }
      break;
    case rchFtpTlsPbsz:
      if (FTPSSCEQ(status, 200))
      { tBoolean data_clear;
        do_shaky_prot:
        data_clear = cond2boolean(ftps_ftm(resource) == ftmAuthTlsDataclear);
        handshake = rchFtpTlsProt;
        my_spf(NULL, 0, &spfbuf, "PROT %c\r\n", (data_clear ? 'C' : 'P'));
        if (data_clear) resource->flags |= rfFtpsDataClear;
        goto switch_to_writing;
      }
      break;
    case rchFtpTlsProt:
      if (FTPSSCEQ(status, 200)) next = rchFtpUser;
#if CONFIG_USER_QUERY
      else if ( ( (status == 534) || (status == 504) || (status == 431) ) &&
        (ftps_ftm(resource) == ftmAutodetect) &&
        (resource_ask_anything(resource, ftps_dataclear_callback,
            mifFtpsDataclear, 0)) )
      { goto susp; }
#endif
      break;
#endif

    case rchFtpUser:
      if ( (FTPSSCEQ(status, 230))
#if OPTION_TLS
          || ( (resource->protocol == rpFtps) && (FTPSSCEQ(status, 232)) )
#endif
         ) /* "no password/account required"; skip "PASS" command */
      { prepare_epsv_pasv:
        /* We always try the more modern EPSV command (RFC2428) first unless
           we definitely know that the server doesn't support it. */
        if (!(resource->actual_hppi->flags & hppifFtpCannotEpsv))
        { resource->flags |= rfFtpTryingEpsv;
          spfbuf = my_strdup("EPSV\r\n"); /* CHECKME: "EPSV ALL"? */
        }
        else { do_pasv: spfbuf = my_strdup("PASV\r\n"); }
        handshake = rchFtpPasv; goto switch_to_writing;
      }
      else if (status == 331) next = rchFtpPassword; /* "password required" */
      else if ( (status == 332) || (status == 532) ) /* "account required" */
        next = rchFtpPassword; /* (see 332/532 comment below) */
      else
      {
#if CONFIG_USER_QUERY
        if ( (status == 530) &&
             (resource_ask_username(resource, ftp_user_callback, 1)) )
        { drp_susp: dealloc(resource->password); susp: return(4); }
#endif
        login_failed: *_ftp_error = reLogin; failed: return(0);
      }
      break;
    case rchFtpPassword:
      if ( (status == 230) || /* "user logged in" */
           (status == 202) )  /* "command was superfluous" */
      { hppi_login_set(resource->textual_hppi, resource->username,
          resource->password);
        goto prepare_epsv_pasv;
      }
      else if ( (status == 332) || (status == 532) ) /* "account required" */
      { /* We don't actually send account information (simply because neither
           account configuration nor FTP upload has been implemented yet),
           but we also don't cause a "login failed" error message. 332 and
           532 often mean "account might be required later, e.g. for upload"
           and not "account is required _now_, or all further operations will
           be refused". So we simply proceed and hope the best. :-) */
        goto prepare_epsv_pasv;
      }
      else
      {
#if CONFIG_USER_QUERY
        if ( ( (status == 530) /* || (status == 421) */ ) &&
             (resource_ask_password(resource, ftp_user_callback, 1)) )
        { /* (Yes, some buggy servers send 421 instead of the correct 530 to
              indicate "login failed"; IMPLEMENTME: must sometimes create a new
              connection and resend the login commands in this case!) */
          goto drp_susp;
        }
#endif
        goto login_failed;
      }
      break;
    case rchFtpPasv:
      /* (General note: we use PASV/EPSV instead of PORT/EPRT due to RFC1579;
          but IMPLEMENTME: fall back to PORT/EPRT when PASV/EPSV fails?) */
      if (resource->flags & rfFtpTryingEpsv)
      { if (FTPSSCEQ(status, 229)) /* "entering passive mode" (EPSV) */
        { const char* path;
          size_t pathlen;
          epsv_pasv_worked:
          if (!ftp_setup_dconn(resource, _ftp_error)) goto failed;
          path = resource->path; pathlen = strlen(path);
          if ( (pathlen > 0) && (path[pathlen - 1] == '/') )
            resource->flags |= rfFtpDir;
          if (!strncmp(path, "/~/", 3)) path += 3;
          my_spf(NULL, 0, &spfbuf, "%s%s%s\r\n",
            ( (resource->flags & rfFtpDir) ? "LIST" : "RETR" ),
            ( (path[0] != '\0') ? strSpace : strEmpty ), path);
          /* IMPLEMENTME: try "MLSD" instead of "LIST" first? */
          handshake = rchFtpRequest;
          goto switch_to_writing;
        }
        else /* "EPSV" failed */
        { resource->actual_hppi->flags |= hppifFtpCannotEpsv;
          resource->flags &= ~rfFtpTryingEpsv;
          goto do_pasv;
        }
      }
      else if (FTPSSCEQ(status, 227)) /* "entering passive mode" (PASV) */
      { goto epsv_pasv_worked; }
      break;
    case rchFtpRequest:
      if (status == 150)
      { /* "opening data connection"; intermediary replies are currently
           handled by ftp_control_read() */
      }
      else if (status == 226)
      { /* We leave the control connection open until the data transfer
           "actually" finished, not until the server thinks it's finished. */
        return(1);
      }
      else if (status / 10 == 55)
      { if (!(resource->flags & rfFtpDir))
        { /* The FTP server says something like "file not found", so we hope
             the thing is available as a _directory_ and try to get that: */
          if (resource->dconn != NULL) conn_remove(&(resource->dconn));
          resource->flags |= rfFtpDir; debugmsg("retrying as FTP dir\n");
          goto prepare_epsv_pasv;
        }
        /* IMPLEMENTME: else { *_ftp_error = re....; goto failed; } */
      }
      break;
    default: /* "should not happen" */
      hs_bug: *_ftp_error = reHandshake; goto failed; /*@notreached@*/ break;
  }
  do_shaky_next:
  if (next == rchUnknown) { *_ftp_error = reResponse; goto failed; }
  else
  { switch (ftp_control_prepare_next(resource, next, &spfbuf))
    { /*u*/ case 0: goto hs_bug; /*@notreached@*/ break;
      /*g*/ case 1: goto login_failed; /*@notreached@*/ break;
#if CONFIG_USER_QUERY
      /*l*/ case 2: goto susp; /*@notreached@*/ break;
#endif
      /*y*/ case 3: goto do_switch_to_writing; /*@notreached@*/ break;
    }
  }

  switch_to_writing:
  resource->handshake = handshake;
  do_switch_to_writing:
  conn_set_writestr(conn, my_spf_use(spfbuf));
  conn_set_write(conn, 2);
#if CONFIG_DEBUG
  debugmsg("prepared new FTP message: "); debugmsg(my_spf_use(spfbuf));
#endif
  return(2);
}

#define FTP_BUFSIZE (4096 + 1)

static void ftp_control_read(tResource* resource)
/* We treat FTP control connections in a special way because we can throw
   almost every byte of input from them directly into the bit bucket. We need
   mostly two things: 1. we wanna know the server status code ("reply code" in
   FTP terminology); 2. we must recognize the end of a (single- or multi-line)
   control message. */
{ char buf[FTP_BUFSIZE], *ptr; /* the bit bucket */
  tTlHeaderState ths = resource->tlheaderstate;
  tRequestContactHandshake handshake;
  tResourceError ftp_error = reServerClosed;
  tConnection* conn = resource->cconn;
  int err, count;

  if (resource->flags & rfFtpShakyAny) { count = 0; ptr = NULL; goto do_shaky;}
  /* CHECKME: the whole handling of the "shaky" stuff is dirty! */

  conn_read(err, conn, buf, FTP_BUFSIZE - 1, return);
  if (err <= 0) /* error (EOF is also an error here because it's unexpected) */
  { stop_reading: resource_set_error(resource, ftp_error);
    do_stop_reading: conn_remove(&(resource->cconn));
    if (resource->dconn != NULL) conn_remove(&(resource->dconn));
      /* When the FTP control connection closes, we also don't want the _data_
         connection any more. But CHECKME: some servers close the control
         connection as soon as the data transfer starts! */
    resource->tlheaderstate = ths; /* CHECKME: set "= 0" instead? */
    push_to_main_res(resource, 0);
    return;
  }

  buf[err] = '\0'; /* for simplicity (e.g. my_strdup()) */
#if CONFIG_DEBUG
  debugmsg("ftp_control_read(): "); debugmsg(buf);
  if (buf[err - 1] != '\n') debugmsg(strNewline);
#endif
  ptr = buf; count = err;

  handshake = resource->handshake;
  if ( (handshake == rchConnected) || (handshake == rchFtpPasv) )
  { /* might need a part (first line, "snippet") of the reply text later... */
    tFtpRpsd* rpsd = (tFtpRpsd*) (resource->rpsd);
    const char *ptrr, *ptrn;
    size_t len;
    if (ths == 0) { dealloc(rpsd->reply_snippet); rpsd->flags &= ~frfSuffsnip;}
    else if (rpsd->flags & frfSuffsnip) goto skipsnip;
    ptrr = my_strchr(buf, '\r'); ptrn = my_strchr(buf, '\n');
    if ( (ptrr == NULL) || ( (ptrn != NULL) && (ptrr > ptrn) ) ) ptrr = ptrn;
    if (ptrr == NULL) len = strlen(buf); /* must store all */
    else { len = ptrr - buf; rpsd->flags |= frfSuffsnip; }
    if (len <= 0) goto skipsnip;
    if (rpsd->reply_snippet == NULL)
      rpsd->reply_snippet = my_strndup(buf, len);
    else
    { size_t rrslen = strlen(rpsd->reply_snippet);
      char* s = __memory_allocate(rrslen + len + 1);
      my_memcpy(s, rpsd->reply_snippet, rrslen);
      my_memcpy(s + rrslen, buf, len);
      s[rrslen + len] = '\0';
      memory_deallocate(rpsd->reply_snippet);
      rpsd->reply_snippet = s;
    }
    skipsnip: {}
  }

  while (count-- > 0)
  { const char ch = *ptr++;
    if (ths < 3) /* handle the server status code at message start */
    { if (ths == 0) resource->server_status_code = 0;
      if (my_isdigit(ch))
      { tServerStatusCode status = resource->server_status_code;
        status = 10 * status + ((tServerStatusCode) (ch - '0'));
        resource->server_status_code = status;
        ths++;
        if ( (ths == 3) && ( (status < 100) || (status > 699) ) )
        { /* (The 6yz reply codes were introduced in RFC2228.) */
          bad_ftp: ftp_error = reResponse; goto stop_reading;
        }
      }
      else goto bad_ftp;
    }
    else if (ths == 3) /* check whether single- or multi-line message */
    { if (ch == ' ') resource->flags |= rfFtpLastLine;
      else if (ch == '-') resource->flags &= ~rfFtpLastLine;
      else goto bad_ftp;
      ths++;
    }
    else if (ths == 4) /* somewhere in a message line */
    { /* (Note: we don't yet check for a complete CRLF, but only for '\n'!) */
      tServerStatusCode status;
      if (ch != '\n') continue; /* not yet at end of message line */
      ths++; /* reached end of message line */
      if (!(resource->flags & rfFtpLastLine)) continue; /* not yet last line */

      /* We've reached and finished the last line; now look what to do: */
      status = resource->server_status_code;
      ths = 0;
#if CONFIG_DEBUG
      sprint_safe(debugstrbuf,
        "found end of FTP control message (fd=%d, rch=%d, ssc=%d)\n",
        conn->fd, resource->handshake, status);
      debugmsg(debugstrbuf);
#endif
      if ( (status / 100) == 1 ) continue; /* intermediary reply */
#if 0
      else if (status == 421) { /* .... */ } /* RFC1123, 4.1.2.11 */
        /* forget it - some buggy servers send 421 when they mean 530... */
#endif
      else
      { do_shaky:
        switch (ftp_control_handshake(resource, &ftp_error))
        { case 0: goto stop_reading; /*@notreached@*/ break;
          case 1: break;
          case 2:
            push_to_main_res(resource, 1); goto out; /*@notreached@*/ break;
          case 3: goto do_stop_reading; /*@notreached@*/ break;
          case 4: goto out; /*@notreached@*/ break; /* suspended */
        }
      }
    }
    else if (ths < 8) /* beginning of a line within a multi-line message */
    { if (my_isdigit(ch))
      { /* still looks like the status code at the beginning of the last line
           of a multi-line message */
        ths++;
      }
      else ths = 4; /* isn't the last line */
    }
    else /* ths == 8 */
    { if (ch == ' ') resource->flags |= rfFtpLastLine; /* _is_ the last line */
      ths = 4;
    }
  }
  out:
  resource->tlheaderstate = ths;
}

#undef FTP_BUFSIZE

#if CONFIG_USER_QUERY

static void ftp_user_callback(tUserQuery* query)
{ tMissingInformationFlags mif = query->mif;
  tResource* resource;
  if (mif & mifObjectVanished) goto out; /* nothing further to do */
  resource = query->resource;
  resource_unsuspend(resource);
  if (mif & mifQueryFailed)
  { do_bug: resource_set_error(resource, reLogin);
    do_push: stop_resource(resource); push_to_main_res(resource, 0); goto out;
  }
  if (mif & mifUserCancelled) goto do_push;
  if (mif & mifUsername) resource->handshake = rchFtpUser;
  else if (mif & mifPassword) resource->handshake = rchFtpPassword;
  else goto do_bug;
  resource->flags |= rfFtpShakyNext;
  ftp_control_read(resource);
  out:
  resource_ask_finish(query);
}

#if OPTION_TLS
static void ftps_dataclear_callback(tUserQuery* query)
{ tMissingInformationFlags mif = query->mif;
  tResource* resource;
  if (mif & mifObjectVanished) goto out; /* nothing further to do */
  resource = query->resource;
  resource_unsuspend(resource);
  if (mif & (mifQueryFailed | mifUserCancelled))
  { stop_resource(resource); push_to_main_res(resource, 0); goto out; }
  ftps_ftm(resource) = ftmAuthTlsDataclear;
  resource->flags |= rfFtpsShakyProt;
  ftp_control_read(resource);
  out:
  resource_ask_finish(query);
}
#endif

#endif /* #if CONFIG_USER_QUERY */

static void ftp_control_callback(tConnection* conn, tConnCbEventKind ccek)
{ tResource* resource = (tResource*) (conn->data);
  switch (ccek)
  { case ccekStart:
      if (conn->flags & cnfConnected) goto do_setup;
      else conn_set_write(conn, 2); /* must "wait" until connected */
      break;
    case ccekConnectWorked:
      conn->flags |= cnfConnected; resource->handshake = rchConnected;
      do_setup:
      resource->rpsd = memory_allocate(sizeof(tFtpRpsd));
      resource->rpsd->remover = ftp_rpsd_remove;
#if OPTION_TLS
      if (resource->protocol == rpFtps)
      { tFtpTlsMethod ftm = ftps_ftm(resource) = calc_ftp_tls_method(resource);
        if (ftm == ftmTls) /* immediate handshake */
        { tBoolean got_session = ftp_control_tls_session_init(conn, resource);
          if (!got_session)
          { resource_set_error(resource, reTls); goto stopnpush; }
          else push_to_main_res(resource, 1);
          break; /* to skip the conn_set_read() call */
        }
      }
#endif
      conn_set_read(conn, 2); /* to get the "server ready" message */
      push_to_main_res(resource, 1);
      break;
    case ccekConnectFailed:
      resource_set_error(resource, connect_err2error(conn_err));
      stopnpush: conn_remove(&(resource->cconn)); push_to_main_res(resource,0);
      break;
    case ccekRead:
#if OPTION_TLS
      if (conn->flags & cnfTlsHandshaking)
      { tdh:
        switch (tls_do_handshaking(conn))
        { case 0: goto stopnpush; /*@notreached@*/ break;
          case 2:
            if (resource->handshake == rchFtpTlsAuthSsl)
            { resource->flags |= rfFtpShakyNext;
              resource->handshake = rchFtpUser;
              ftp_control_read(resource);
              return;
            }
            break;
        }
        break;
      }
#endif
      ftp_control_read(resource); break;
    case ccekWrite:
#if OPTION_TLS
      if (conn->flags & cnfTlsHandshaking) goto tdh;
#endif
      switch (conn_write_writedata(conn))
      { case 0:
          resource_set_error(resource, reServerClosed);
          if (resource->dconn != NULL) conn_remove(&(resource->dconn));
          goto stopnpush;
          /*@notreached@*/ break;
        case 1: break; /* keep writing */
        case 2: conn_set_read(conn, 2); break; /* all written, start reading */
      }
      break;
    default: conn_bug(conn, ccek); break;
  }
}

#endif /* #if CONFIG_FTP */


/* Non-local resources: NNTP */

#if OPTION_NEWS

typedef struct
{ tRpsdRemover remover;
  char* text;
  size_t usable, used;
  tNewsArticleNumber curr_artnum, end_artnum;
} tNntpRpsd;

static void news_rpsd_remove(tRpsdGeneric* _rpsd)
{ tNntpRpsd* rpsd = (tNntpRpsd*) _rpsd;
  __dealloc(rpsd->text);
}

static one_caller void news_rpsd_prepare(tResource* resource)
{ tNntpRpsd* rpsd = (tNntpRpsd*) resource->rpsd;
  if (rpsd == NULL)
  { void* x = memory_allocate(sizeof(tNntpRpsd));
    rpsd = x; resource->rpsd = x;
    rpsd->remover = news_rpsd_remove;
  }
  else { rpsd->used = 0; /* rpsd->curr_artnum = rpsd->end_artnum = 0; */ }
}

typedef struct tNewsArticle
{ struct tNewsArticle* next;
  char *header, *lastheader, *body, *lastbody; /* content blocks */
  char* id;
  tNewsArticleNumber num;
} tNewsArticle;

enum { ngifNone = 0, ngifCannotPost = 0x01 };
typedef unsigned char tNewsGroupInformationFlags;

typedef struct tNewsGroupInformation
{ struct tNewsGroupInformation* next;
  char* name;
  tNewsArticle* articles;
  tNewsArticleNumber estnum, firstnum, lastnum;
    /* estimated number of articles; number of the first and last article */
  tNewsGroupInformationFlags flags;
} tNewsGroupInformation;

static /* __sallocator -- not an "only" reference */ tNewsGroupInformation*
  __callocator news_ngi_create(tCachedHostInformation* hostinfo,
  const char* name)
{ tNewsGroupInformation* retval =
    (tNewsGroupInformation*) memory_allocate(sizeof(tNewsGroupInformation));
  retval->next = hostinfo->ngi;
  retval->name = my_strdup(name);
  hostinfo->ngi = retval;
  return(retval);
}

static tNewsGroupInformation* news_ngi_lookup(tCachedHostInformation*
  hostinfo, const char* desired_name, tBoolean do_create_if_null)
{ tNewsGroupInformation* retval = hostinfo->ngi;
  while (retval != NULL)
  { const char* name = retval->name;
    if ( (name != NULL) && (!strcmp(name, desired_name)) ) break; /* found */
    retval = retval->next;
  }
  if ( (retval == NULL) && (do_create_if_null) )
    retval = news_ngi_create(hostinfo, desired_name);
  return(retval);
}

static void news_resource2path(const tResource* resource,
  /*@out@*/ const char** _path, /*@out@*/ tBoolean* _has_at)
/* prepares the path value of the <resource> a little bit */
{ const char* path = resource->path;
  *_has_at = falsE;
  if (path != NULL)
  { if (*path == '/') path++;
    if (*path == '\0') path = NULL;
    else if (my_strchr(path, '@')) *_has_at = truE;
  }
  *_path = path;
}

static void news_resource2group(const tResource* resource,
  /*@out@*/ const char** _group, /*@out@*/ tBoolean* _must_dealloc)
{ const char* path;
  tBoolean has_at;
  *_group = NULL; *_must_dealloc = falsE;
  news_resource2path(resource, &path, &has_at);
  if ( (path != NULL) && (!has_at) )
  { const char* temp = my_strchr(path, '/');
    if (temp == NULL) *_group = path; /* all group, no message number */
    else if (temp == path) { /* no group given, bad URI - CHECKME! */ }
    else { *_group = my_strndup(path, temp - path); *_must_dealloc = truE; }
  }
}

static void news_resource2article(const tResource* resource,
  /*@out@*/ const char** _article, /*@out@*/ tBoolean* _is_numerical)
{ const char* path;
  tBoolean has_at;
  *_article = NULL; *_is_numerical = falsE;
  news_resource2path(resource, &path, &has_at);
  if (path != NULL)
  { if (has_at) { *_article = path; /* *_is_numerical = falsE; */ }
    else
    { const char* temp = my_strchr(path, '/');
      if ( (temp != NULL) && (temp[1] == '\0') ) /* article string is empty */
        temp = NULL;
      if (temp != NULL) { *_article = temp + 1; *_is_numerical = truE; }
    }
  }
}

static tBoolean news_fetch_next_header(tResource* resource)
/* returns whether there actually is a "next" article header to be fetched */
{ tNntpRpsd* rpsd = (tNntpRpsd*) resource->rpsd;
  tBoolean retval;
  tNewsArticleNumber currnum = --(rpsd->curr_artnum);
  if (currnum >= rpsd->end_artnum)
  { char* spfbuf;
    my_spf(NULL, 0, &spfbuf, "HEAD %d\r\n", currnum);
    conn_set_writestr(resource->cconn, my_spf_use(spfbuf));
    retval = truE;
  }
  else retval = falsE;
  return(retval);
}

#define is_news_whitespace(ch) ( ((ch) == ' ') || ((ch) == '\t') )
  /* (RFC2822, 2.2.2) */

static one_caller unsigned char news_handshake(tResource* resource,
  const char** _valueptr)
/* return value: 0=error, 1=continue reading, 2=start writing, 3=done */
{ tRequestContactHandshake handshake = resource->handshake;
  tServerStatusCode ssc = resource->server_status_code;
  const char *valueptr = *_valueptr, *group, *article;
  char* spfbuf;
  tBoolean is_numerical, must_dealloc;
  switch (handshake)
  { case rchConnected:
      if (ssc == 200) /* "server ready, posting allowed" */
      { prepare_mode_reader:
        if (valueptr != NULL)
        { while (is_news_whitespace(*valueptr)) valueptr++;
          hppi_set_software_id(resource->textual_hppi, valueptr);
        }
        conn_set_dissolver(resource->cconn, quitcmd_dissolver);
        handshake = rchNntpModeReader;
        spfbuf = my_strdup("MODE READER\r\n"); /* be polite (RFC2980, 2.3) */
        switch_to_writing:
        conn_set_writestr(resource->cconn, my_spf_use(spfbuf));
        start_writing: resource->handshake = handshake;
        do_start_writing: return(2);
      }
      else if (ssc == 201) /* "server ready, no posting allowed" */
      { resource->textual_hppi->flags |= hppifNntpCannotPostArticles;
        goto prepare_mode_reader;
      }
      break;
    case rchNntpModeReader:
      /* Don't care about the status code here. If the server didn't
         understand "MODE READER", that's solely the server's problem. */
      news_resource2article(resource, &article, &is_numerical);
      if ( (article != NULL) && (!is_numerical) )
      { /* fetch the given article (header and body) */
        prepare_fetch_article:
        handshake = rchNntpFetchArticle;
        my_spf(NULL, 0, &spfbuf, "ARTICLE %s%s%s\r\n", (is_numerical ? strEmpty
          : strLt), article, (is_numerical ? strEmpty : strGt));
      }
      else
      { news_resource2group(resource, &group, &must_dealloc);
        if (group == NULL)
        { handshake = rchNntpGetGroups; spfbuf = my_strdup("LIST\r\n"); }
        else
        { /* must select a group first, in order to prepare further actions */
          handshake = rchNntpSelectGroup;
          my_spf(NULL, 0, &spfbuf, "GROUP %s\r\n", group);
        }
        if (must_dealloc) memory_deallocate(group);
      }
      goto switch_to_writing; /*@notreached@*/ break;
    case rchNntpGetGroups:
      if (ssc == 215) /* it worked, list of groups follows */
      { const char* hostname = resource2textual_host(resource)->hostname;
        char* spfbuf2;
        resource->kind = rckHtml;
        my_spf(strbuf, STRBUF_SIZE, &spfbuf, _("News groups on %s"), hostname);
        my_spf(strbuf2, STRBUF_SIZE, &spfbuf2,
          _("List of available news groups on the server %s"), hostname);
        resource_collect_title2(resource, spfbuf, spfbuf2);
        my_spf_cleanup(strbuf, spfbuf);
        my_spf_cleanup(strbuf2, spfbuf2);
        continue_reading: ((tNntpRpsd*)(resource->rpsd))->used = 0; return(1);
      }
      break;
    case rchNntpFetchArticle:
      if (ssc == 220) goto continue_reading; /* it worked, article follows */
      break;
    case rchNntpFetchHeader:
      if (ssc == 221) goto continue_reading; /* it worked, header follows */
      else if ( (resource->flags & rfNntpHeaderSeries) &&
                ( (ssc == 423) || (ssc == 430) ) )
      { /* This specific article in a series doesn't exist, but further
           articles might exist. */
        if (news_fetch_next_header(resource)) goto do_start_writing;
        else goto completed;
      }
      break;
    case rchNntpSelectGroup:
      if (ssc == 211) /* selecting the group worked */
      { tNewsArticleNumber estnum, firstnum, lastnum;
        tBoolean need_valuedata;
        news_resource2group(resource, &group, &must_dealloc);
        news_resource2article(resource, &article, &is_numerical);
        need_valuedata = cond2boolean( (article == NULL) || (!is_numerical) );
          /* (The "!is_numerical" is a "should not happen" test here.) */
        if (need_valuedata)
        { /* interpret/store server data */
          if (valueptr == NULL)
          { bad_valueptr: if (must_dealloc) memory_deallocate(group);
            bad_news: resource_set_error(resource, reResponse);
            found_error: return(0);
          }
          while (*valueptr == ' ') { valueptr++; }
          { PREPARE_SSCANF_FORMAT(format, 20, "%d %d %d ")
            if (sscanf(valueptr, format, &estnum, &firstnum, &lastnum) != 3)
              goto bad_valueptr;
          }
          *_valueptr = NULL;
          if (group != NULL) /* "should" be true */
          { tNewsGroupInformation* ngi =
              news_ngi_lookup(resource2textual_host(resource), group, truE);
            ngi->estnum = estnum; ngi->firstnum = firstnum;
            ngi->lastnum = lastnum;
          }
        }
        /* prepare next command */
        if (need_valuedata)
        { const char *groupstr, *hostname;
          char* spfbuf2;
          tNntpRpsd* rpsd;
          prepare_article_index:
          groupstr = ((group != NULL) ? group : _(strUnknown));
          hostname = resource2textual_host(resource)->hostname;
          resource->kind = rckHtml;
          my_spf(strbuf, STRBUF_SIZE, &spfbuf,
            _("News group %s, articles %d-%d, server %s"), groupstr, firstnum,
            lastnum, hostname);
          my_spf(strbuf2, STRBUF_SIZE, &spfbuf2, _("Index of news articles %d-%d in the group %s on the server %s, latest first"),
            firstnum, lastnum, groupstr, hostname);
          resource_collect_title2(resource, spfbuf, spfbuf2);
          my_spf_cleanup(strbuf, spfbuf);
          my_spf_cleanup(strbuf2, spfbuf2);
          rpsd = (tNntpRpsd*) resource->rpsd;
          rpsd->curr_artnum = lastnum + 1;
          rpsd->end_artnum = firstnum;
          resource->flags |= rfNntpHeaderSeries;
          handshake = rchNntpFetchHeader;
          if (must_dealloc) memory_deallocate(group);
          if (news_fetch_next_header(resource)) goto start_writing;
          else { completed: return(3); }
        }
        else
        { const char* temp;
          if ( (article != NULL) && (is_numerical) && (my_isdigit(*article)) &&
               ( (temp = my_strchr(article, '-')) != NULL ) &&
               (my_isdigit(temp[1])) )
          { /* roughly looks like a numerical range of the form "37-42" */
            my_atoi(article, &firstnum, &temp, MY_ATOI_INT_MAX);
            if (*temp == '-')
            { my_atoi(temp + 1, &lastnum, &temp, MY_ATOI_INT_MAX);
              if (*temp == '\0') goto prepare_article_index; /* it is */
            }
          }
          if (must_dealloc) memory_deallocate(group);
          goto prepare_fetch_article;
        }
        /* currently: */ /*@notreached@*/
        goto switch_to_writing;
      }
      break;
    default:
      resource_set_error(resource, reHandshake); goto found_error;
      /*@notreached@*/ break;
  }
  goto bad_news;
}

static const char* __news_lookup_header_line(const char** lines,
  size_t linenum, const char* which)
{ const char* retval = NULL;
  size_t len = strlen(which), count;
  for (count = 0; count < linenum; count++)
  { const char* line = lines[count];
    if (strneqcase(line, which, len))
    { const char* temp = line + len;
      while (is_news_whitespace(*temp)) temp++;
      if (*temp == ':')
      { temp++;
        while (is_news_whitespace(*temp)) temp++;
        if (*temp != '\0') retval = temp;
        break;
      }
    }
  }
  return(retval);
}

#define news_lookup_header_line(which) \
  __news_lookup_header_line(lines, linenum, which)

#define store_header_line \
  if (linenum >= maxlinenum) \
  { maxlinenum += 20; \
    lines = memory_reallocate(lines, maxlinenum * sizeof(char*)); \
  } \
  lines[linenum++] = linestart;

static one_caller void news_handle_article_index(tResource* resource)
/* adds one line to an article index document */
{ tNntpRpsd* rpsd = (tNntpRpsd*) resource->rpsd;
  char ch, *text, *temp, *linestart; /* article header text */
  const char *subject, *htmlsubject, *sender, *htmlsender, *date, *htmldate,
    *bodylinenum, *htmlbodylinenum, *bodylinenote, *group;
  size_t used, linenum, maxlinenum;
  const char** lines;
  char* spfbuf;
  tBoolean near_r, must_dealloc_group;

  if (rpsd == NULL) return; /* "should not happen" */
  text = rpsd->text; used = rpsd->used;
  if ( (text == NULL) || (*text == '\0') || (used <= 0) ) /* no text */
    goto out;

  /* split the header into lines */

  linestart = temp = text; near_r = falsE;
  linenum = maxlinenum = 0; lines = NULL;
  while ( (ch = *temp++) != '\0' )
  { if (ch == '\r') near_r = truE;
    else
    { if ( (ch == '\n') && (near_r) )
      { /* IMPLEMENTME: strip comments (RFC2822, 3.2.3) */
        if (is_news_whitespace(*temp)) /* folded; unfold (RFC2822, 2.2.3) */
        { *(temp - 2) = *(temp - 1) = ' '; /* overwrites the CRLF */
        }
        else
        { *(temp - 2) = '\0'; /* overwrites the "\r" */
          store_header_line
          linestart = temp;
        }
      }
      near_r = falsE;
    }
  }
  store_header_line

  /* interpret the header lines */

  subject = news_lookup_header_line("subject");
  if (subject == NULL) subject = _("(no subject)");
  sender = news_lookup_header_line("from");
  if (sender == NULL)
    sender = news_lookup_header_line("sender"); /* better than nothing */
  date = news_lookup_header_line("date");
  bodylinenum = news_lookup_header_line("lines");
  if (bodylinenum == NULL) bodylinenote = strEmpty;
  else if (!strcmp(bodylinenum, "1")) bodylinenote = _(" line");
  else bodylinenote = _(" lines");

  /* build the HTML line */

  news_resource2group(resource, &group, &must_dealloc_group);
  htmlsubject = htmlify(subject);

#define hify(orig, ified) ified = ( (orig != NULL) ? htmlify(orig) : NULL )
  hify(sender, htmlsender); hify(date, htmldate);
  hify(bodylinenum, htmlbodylinenum);
#undef hify

#define appt(what) \
  ( (what != NULL) ? strSpacedDash : strEmpty ), \
  ( (what != NULL) ? what : strEmpty )
  my_spf(strbuf, STRBUF_SIZE, &spfbuf,
    "<br><a href=\"news://%s/%s/%d\">%s</a>%s%s%s%s%s%s%s\n",
    resource2textual_host(resource)->hostname,
    ( (group != NULL) ? group : "BUG" ),
    ((tNntpRpsd*) (resource->rpsd))->curr_artnum, htmlsubject,
    appt(htmlsender), appt(htmldate), appt(htmlbodylinenum), bodylinenote);
#undef appt

  resource_collect_str(resource, spfbuf);
  my_spf_cleanup(strbuf, spfbuf);

#define hc(orig, ified) if (orig != NULL) htmlify_cleanup(orig, ified)
  hc(subject, htmlsubject); hc(sender, htmlsender); hc(date, htmldate);
  hc(bodylinenum, htmlbodylinenum);
#undef hc

  if (must_dealloc_group) memory_deallocate(group);
  memory_deallocate(lines);
  out:
  rpsd->used = 0;
}

#undef store_header_line
#undef news_lookup_header_line

static one_caller void news_handle_groupline(tResource* resource, char* text)
/* adds one line to a groups index document; the <text> should be of the form
   "group last first p" (RFC977, 3.6.1) */
{ char *pos = my_strchr(text, ' '), *spfbuf;
  const char *info, *name, *htmlname;
  if (pos == NULL) info = NULL; /* server bug */
  else { *pos++ = '\0'; info = pos; }
  name = text; htmlname = htmlify(name);
  my_spf(strbuf, STRBUF_SIZE, &spfbuf,
    "<br><a href=\"news://%s/%s\">%s</a>%s%s\n",
    resource2textual_host(resource)->hostname, name, htmlname,
    ( (info != NULL) ? strSpace : strEmpty ), null2empty(info));
  resource_collect_str(resource, spfbuf);
  my_spf_cleanup(strbuf, spfbuf);
  htmlify_cleanup(name, htmlname);
}

static void news_collect(tResource* resource, const char* src, size_t size)
{ tRequestContactHandshake handshake = resource->handshake;
  if ( (handshake == rchNntpGetGroups) ||
       ( (handshake == rchNntpFetchHeader) &&
         (resource->flags & rfNntpHeaderSeries) ) )
  { tNntpRpsd* rpsd = (tNntpRpsd*) resource->rpsd;
    char* text;
    size_t used, usable, needed;
    if (rpsd == NULL) goto res_coll; /* "should not happen" */
    text = rpsd->text; used = rpsd->used; usable = rpsd->usable;
    needed = zero2one(used) + size;
    if (needed > usable)
    { text = rpsd->text = memory_reallocate(text, needed);
      rpsd->usable = needed;
    }
    my_memcpy(text + used - cond2bool(used > 0), src, size);
    text[needed - 1] = '\0';
    rpsd->used = needed;
    if (handshake == rchNntpGetGroups)
    { /* look for "\r\n" */
      char *temp = text, *remainder = text;
      tBoolean near_r = falsE;
      char ch;
      while ( (ch = *temp++) != '\0' )
      { if (ch == '\r') near_r = truE;
        else
        { if ( (ch == '\n') && (near_r) )
          { size_t len = (temp - remainder) - 2; /* "-2" for the CRLF */
            if (len > 0) /* actually have useful text */
            { remainder[len] = '\0'; /* overwrites the "\r" */
              news_handle_groupline(resource, remainder);
            }
            remainder = temp;
          }
          near_r = falsE;
        }
      }
      if (remainder > text) /* handled some grouplines */
      { if (*remainder == '\0') rpsd->used = 0; /* handled whole text */
        else
        { char* dest = text;
          rpsd->used = rpsd->used - (remainder - text);
          do { ch = *remainder++; *dest++ = ch; } while (ch != '\0');
        }
        push_to_main_res(resource, 1);
      }
    }
  }
  else { res_coll: resource_collect(resource, src, size); }
}

#define NEWS_BUFSIZE (4096 + 1)
#define NEWS_READ_BUFSIZE (2048 - 5)

#define news_read_flush \
  { if (read_buf_used > 0) \
    { news_collect(resource, read_buf, read_buf_used); read_buf_used = 0; } \
  }

#define news_read_append_ch(ch) \
  do \
  { if (read_buf_used >= NEWS_READ_BUFSIZE) news_read_flush \
    read_buf[read_buf_used++] = ch; \
  } while (0)

#define news_read_append(_addr, _count) \
  do \
  { const char* a = _addr; /* (evaluate it only once) */ \
    size_t c = _count; \
    while (c-- > 0) { char x = *a++; news_read_append_ch(x); } \
  } while (0)

static one_caller void news_read(tResource* resource)
/* We treat NNTP connections in a special way because we must filter out the
   control data and construct/store the actual content data. */
{ static const char strDotMarker[] = "\r\n.\r\n";
  char buf[NEWS_BUFSIZE], *ptr; /* temporary raw data buffer */
  const char* valueptr; /* NULL or inside buf[] */
  char read_buf[NEWS_READ_BUFSIZE + 5]; /* temporary content buffer */
  tTlHeaderState ths = resource->tlheaderstate;
  tResourceError news_error = reServerClosed;
  int fd = resource->cconn->fd, err = my_read(fd, buf, NEWS_BUFSIZE - 1);
  size_t count, read_buf_used = 0;
  tBoolean do_push = falsE;

  if (err <= 0) /* error (EOF is also an error here because it's unexpected) */
  { stop_reading:
    resource_set_error(resource, news_error);
    do_stop_reading:
    conn_remove(&(resource->cconn));
    resource->tlheaderstate = ths; /* CHECKME: set "= 0" instead? */
    news_read_flush
    rpsd_remove(resource);
    push_to_main_res(resource, 0);
    return;
  }

  buf[err] = '\0'; /* for simplicity (e.g. sscanf()) */
#if CONFIG_DEBUG
  debugmsg("news_read(): "); debugmsg(buf);
  if (buf[err - 1] != '\n') debugmsg(strNewline);
#endif
  ptr = buf; valueptr = NULL; count = err;
  while (count-- > 0)
  { const char ch = *ptr++;
    switch (ths)
    { case 0:
        resource->server_status_code = 0; /*@fallthrough@*/
      case 1: case 2: /* parsing server status code at message start */
        if (my_isdigit(ch))
        { tServerStatusCode status = resource->server_status_code;
          status = 10 * status + ((tServerStatusCode) (ch - '0'));
          resource->server_status_code = status;
          ths++;
          if (ths == 3)
          { if ( (status < 100) || (status > 599) )
            { bad_news: news_error = reResponse; goto stop_reading; }
            valueptr = ptr;
          }
        }
        else goto bad_news;
        break;
      case 3: /* somewhere in a status response line */
        if (ch == '\r') ths = 4; /* possibly at end of line */
        break;
      case 4:
        if (ch == '\r') { /* don't change ths */ }
        else if (ch != '\n') ths = 3; /* not actually at end of line */
        else /* at end of line; look what to do */
        { ths = 0;
          switch (news_handshake(resource, &valueptr))
          { case 0: goto bad_news; /*@notreached@*/ break;
            case 1: /* read some "content" */
              resource->state = rsReading; ths = 7; break;
            case 2:
              start_writing: conn_set_write(resource->cconn, 2);
              goto out; /*@notreached@*/ break;
            case 3:
              resource_complete:
              resource->state = rsComplete; goto do_stop_reading;
              /*@notreached@*/ break;
          }
          valueptr = NULL;
        }
        break;
      case 5: case 6: case 7: case 8: case 9: /* "content" */
        if (ch == strDotMarker[ths - 5]) /* looks like the end marker */
        { ths++;
          if (ths == 10) /* _is_ a _complete_ end marker */
          { ths = 0; news_read_flush
            if ( (resource->handshake == rchNntpFetchHeader) &&
                 (resource->flags & rfNntpHeaderSeries) )
            { news_handle_article_index(resource);
              if (news_fetch_next_header(resource))
                goto start_writing; /* get the next article header */
            }
            goto resource_complete;
          }
        }
        else if (ths > 5)
        { /* It wasn't actually the end marker, it only looked like the
             beginning of the marker. */
          if ( (ths == 8) && (ch == '.') )
          { /* collapse double-dot to single-dot; how dotty... */
            ths = 7; /* FIXME: "=5" if there wasn't any actual text! */
          }
          news_read_append(strDotMarker, ths - 5);
          if (ch == '\r') ths = 6; /* possibly beginning of end marker */
          else { ths = 5; goto append_ch; } /* back to "normal content" mode */
        }
        else /* "normal content" */
        { append_ch:
          do_push = truE;
          news_read_append_ch(ch);
          if ( (ch == '\n') && (resource->kind == rckHtml) )
            news_read_append(strBr, 4); /* FIXME: do this in ...append_ch()! */
        }
        break;
    } /* switch (ths) */
  } /* while count */
  out:
  resource->tlheaderstate = ths;
  news_read_flush
  if (do_push) push_to_main_res(resource, 1);
}

#undef NEWS_BUFSIZE
#undef NEWS_READ_BUFSIZE

static void news_callback(tConnection* conn, tConnCbEventKind ccek)
{ tResource* resource = (tResource*) (conn->data);
  switch (ccek)
  { case ccekStart:
      if (conn->flags & cnfConnected) goto start_reading;
      else conn_set_write(conn, 2); /* must "wait" until connected */
      break;
    case ccekConnectWorked:
      conn->flags |= cnfConnected; resource->handshake = rchConnected;
      start_reading:
      conn_set_read(conn, 2); /* to get the "server ready" message */
      news_rpsd_prepare(resource); push_to_main_res(resource, 1);
      break;
    case ccekConnectFailed:
      resource_set_error(resource, connect_err2error(conn_err));
      stopnpush: conn_remove(&(resource->cconn)); push_to_main_res(resource,0);
      break;
    case ccekRead: news_read(resource); break;
    case ccekWrite:
      switch (conn_write_writedata(conn))
      { case 0:
          resource_set_error(resource, reServerClosed); goto stopnpush;
          /*@notreached@*/ break;
        case 1: break; /* keep writing */
        case 2: goto start_reading; /*@notreached@*/ break;
      }
      break;
    default: conn_bug(conn, ccek); break;
  }
}

#endif /* #if OPTION_NEWS */


/* Non-local resources: finger */

#if CONFIG_FINGER

static one_caller void finger_read(tResource* resource)
{ int fd = resource->cconn->fd;
  char* dest;
  size_t size;
  int err;
  resource_provide_room(resource, &dest, &size, resource->bytecount, 2);
  err = my_read(fd, dest, size);
  if (err <= 0) /* error or EOF */
  { conn_remove(&(resource->cconn));
    if (err == 0) resource->state = rsComplete;
    else resource_set_error(resource, reServerClosed);
    push_to_main_res(resource, 0);
  }
  else
  { resource_record(resource, dest, err);
    push_to_main_res(resource, 1);
  }
}

static void finger_callback(tConnection* conn, tConnCbEventKind ccek)
{ tResource* resource = (tResource*) (conn->data);
  const char* username;
  switch (ccek)
  { case ccekStart:
      conn_set_write(conn, 2);
      if (conn->flags & cnfConnected) goto set_req;
      break;
    case ccekConnectWorked:
      conn->flags |= cnfConnected; resource->handshake = rchConnected;
      set_req:
      username = resource->username;
      if (username == NULL) conn_set_writestr(conn, my_strdup(strCrlf));
      else
      { char* spfbuf;
        my_spf(NULL, 0, &spfbuf, "%s\r\n", username);
        conn_set_writestr(conn, my_spf_use(spfbuf));
      }
      push_to_main_res(resource, 1);
      break;
    case ccekConnectFailed:
      resource_set_error(resource, connect_err2error(conn_err));
      stopnpush: conn_remove(&(resource->cconn)); push_to_main_res(resource,0);
      break;
    case ccekRead: finger_read(resource); break;
    case ccekWrite:
      switch (conn_write_writedata(conn))
      { case 0:
          resource_set_error(resource, reServerClosed); goto stopnpush;
          /*@notreached@*/ break;
        case 1: break; /* keep writing */
        case 2: /* all written, start reading */
          conn_set_read(conn, 2); resource->state = rsReading;
          push_to_main_res(resource, 1); break;
      }
      break;
    default: conn_bug(conn, ccek); break;
  }
}

#endif /* #if CONFIG_FINGER */


/* Non-local resources: General */

static tBoolean start_request(tResourceRequest* request,
  tCachedHostInformation* hostinfo)
/* starts (prepares) a network request; returns want-it-back info */
{ tResourceProtocol protocol = request->protocol;
  const tConfigProxy* proxy = request->proxy;
  tResourceRequestFlags rrf = request->flags;
  tResource* resource;
  tPortnumber connport, textual_portnumber;
  tConnection* conn;
  tConnCbFunc cb;
  int err, fd;
  tBoolean is_connected = falsE;

  switch (protocol)
  { case rpHttp:
#if OPTION_TLS
    case rpHttps:
#endif
      cb = http_callback; break;
#if CONFIG_FTP
    case rpFtp:
#if OPTION_TLS
    case rpFtps:
#endif
      cb = ftp_control_callback; break;
#endif
#if OPTION_NEWS
    case rpNntp: cb = news_callback; break;
#endif
#if CONFIG_FINGER
    case rpFinger: cb = finger_callback; break;
#endif
    default: /* "should not happen" */
      resource_request_set_error(request, reProtocol);
      failed: return(falsE); /*@notreached@*/ break;
  }

  fd = create_socket(hostinfo->address_family);
  if (fd < 0)
  { resource_request_set_error(request, ((fd == -2) ? reTmofd : reSocket));
    goto failed;
  }

  textual_portnumber = request->textual_portnumber;

  if ( (proxy != NULL) && (proxy->proxy_hostname != NULL) )
    connport = proxy->proxy_portnumber;
  else connport = textual_portnumber;

  err = my_connect(fd, hostinfo, connport);
  if (err == 0) is_connected = truE; /* unlikely when non-blocking */
  else if (errno != EINPROGRESS)
  { resource_request_set_error(request, connect_err2error(errno));
    goto failed;
  }

  /* Everything's fine, prepare for getting the resource */

  resource = resource_create(request, NULL, UNKNOWN_CONTENTLENGTH, rckUnknown,
    rsConnecting);
  resource->textual_portnumber = textual_portnumber;
  resource->proxy = request->proxy;
  resource->actual_hppi = hppi_lookup(hostinfo, connport, protocol, truE);

  { const char* hostname = request->hostname;
    tCachedHostInformation* textual_host = hostinfo_lookup(hostname);
    if (textual_host == NULL) textual_host = hostinfo_create(hostname);
    resource->textual_hppi = hppi_lookup(textual_host, textual_portnumber,
      protocol, truE);
  }

  if (request->username != NULL)
    resource->username = my_strdup(request->username);
  if (request->password != NULL)
    resource->password = my_strdup(request->password);
  if (request->query != NULL) resource->query = my_strdup(request->query);
  if (request->post != NULL) resource->post = my_strdup(request->post);
  if (rrf & rrfPost) resource->flags |= rfPost;
  if (rrf & rrfIsRedirection) resource->flags |= rfIsRedirection;
  if (rrf & rrfIsEmbedded) resource->flags |= rfIsEmbedded;
  if (request->action == rraEnforcedReload) resource->flags |= rfIsEnforced;

  conn = resource->cconn = conn_create(fd, cb, resource, truE,
    resource->actual_hppi, protocol);
  if (!is_connected) { /* resource->handshake = rchConnecting; -- done */ }
  else { conn->flags |= cnfConnected; resource->handshake = rchConnected; }
  conn_callback(conn, ccekStart);
  return(truE);
}


/* Resource handling basics */

#if OPTION_THREADING

static one_caller void handle_dns_thread_message(void)
{ static tDnsLookup* dns_lookup;
  static unsigned char count = 0;
  tCachedHostInformation* hostinfo;
  tResourceRequest** reqlist;
  tReqlistNum num, rlcount;
  const char* hostname;
  int err = my_read(fd_dns2resource_read, ((char*)(&dns_lookup)) + count,
    sizeof(dns_lookup) - count);
  if (err <= 0) return;
  count += err;
  if (count < sizeof(dns_lookup)) return; /* not yet a whole pointer */
  count = 0; /* for the next round */

  hostinfo = dns_lookup->hostinfo;
  hostinfo->flags &= ~chifAddressLookupRunning;
  hostname = hostinfo->hostname;
  reqlist = hostinfo->waiting;
  num = hostinfo->waiting_num;
  if (dns_lookup->flags & dlfFailed)
  { hostinfo->flags |= chifAddressLookupFailed;
    hostinfo->lookupfailuretime = my_time();
    for (rlcount = 0; rlcount < num; rlcount++)
    { tResourceRequest* request = reqlist[rlcount];
      resource_request_set_error(request, reDns);
      push_to_main_req(request, 0);
    }
  }
  else /* the lookup worked */
  { hostinfo->addrinfo = dns_lookup->addrinfo;
    hostinfo->addrinfolen = dns_lookup->addrinfolen;
    hostinfo->address_family = dns_lookup->address_family;
    hostinfo->flags |= chifAddressValid;
    for (rlcount = 0; rlcount < num; rlcount++)
    { tResourceRequest* request = reqlist[rlcount];
      tBoolean want_it_back = start_request(request, hostinfo);
      push_to_main_req(request, boolean2bool(want_it_back));
    }
  }

  reqlist_reset(&(hostinfo->waiting), &(hostinfo->waiting_num),
    &(hostinfo->waiting_maxnum));
  memory_deallocate(dns_lookup->hostname);
  memory_deallocate(dns_lookup);
}

#endif /* #if OPTION_THREADING */

static one_caller void check_connect(tConnection* conn)
{ int fd = conn->fd, sockerr = 0;
  socklen_t dummy = (socklen_t) sizeof(sockerr);
  if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &sockerr, &dummy) != 0)
  { conn_err = errno;
    failed:
#if CONFIG_DEBUG
    sprint_safe(debugstrbuf, "check_connect(): failed for fd #%d; %d,%d\n",
      fd, conn_err, sockerr);
    debugmsg(debugstrbuf);
#endif
    conn_callback(conn, ccekConnectFailed);
    return;
  }
  if (sockerr != 0) { conn_err = sockerr; goto failed; }
  conn_callback(conn, ccekConnectWorked);
}

static one_caller void resource_handle_fds(int selectcount)
{ tConnTableIndex cti;

#if OPTION_THREADING
  if (FD_ISSET(fd_dns2resource_read, &fds_r)) /* got an answer from DNS */
  { handle_dns_thread_message();
    if (--selectcount <= 0) return; /* done */
  }
#endif

  for (cti = 0; cti < conn_table_num; cti++)
    conn_table[cti]->flags &= ~cnfHandled;
  connloop:
  did_conn_remove = falsE;
  for (cti = 0; cti < conn_table_num; cti++)
  { tConnection* conn = conn_table[cti];
    tConnectionFlags flags = conn->flags;
    int fd;
    if (flags & cnfHandled) continue; /* already handled before */
    conn->flags |= cnfHandled;
    fd = conn->fd;
    if ( (FD_ISSET(fd, &fds_r)) || (FD_ISSET(fd, &fds_e)) )
      conn_callback(conn, ccekRead);
    else if (FD_ISSET(fd, &fds_w))
    { if (flags & cnfConnected) conn_callback(conn, ccekWrite);
      else check_connect(conn);
    }
    else continue; /* nothing to do for this connection */
    if (--selectcount <= 0) break;
    if (did_conn_remove) goto connloop; /* IMPROVEME! */
  }
}

static one_caller void initial_request_handler(tResourceRequest* request)
{ tBoolean want_it_back = falsE;
  tResourceProtocol rp = request->protocol;
  switch (rp)
  { case rpLocal: fetch_local(request); break;
    case rpAbout: fetch_about(request); break;
#if OPTION_LOCAL_CGI
    case rpLocalCgi: want_it_back = start_local_cgi(request); break;
#endif
    default:
      if (resource_lookup(request) != NULL) { /* found in cache, done */ }
#if OPTION_POP
      else if (is_poplike(rp)) /* e-mail access is special */
        want_it_back = pop_fetch(request);
#endif
      else /* some network activity might be necessary :-) */
      { tCachedHostInformation* hostinfo;
        tHostAddressLookupResult res = lookup_hostaddress(request, &hostinfo,
          ( (request->action == rraEnforcedReload) ? halfEnforcedReload :
            halfNone ));
        if (res == halrFine) want_it_back = start_request(request, hostinfo);
#if CONFIG_ASYNC_DNS
        else if (res == halrLookupRunning)
        { request->state = rrsDnsLookup;
          request->lookup = hostinfo;
          reqlist_add(&(hostinfo->waiting), &(hostinfo->waiting_num),
            &(hostinfo->waiting_maxnum), request);
          want_it_back = truE;
        }
#endif
        else resource_request_set_error(request, reDns);
      }
      break;
  }
  push_to_main_req(request, boolean2bool(want_it_back));
}

#if CONFIG_ASYNC_DNS
static void stop_request_dns_lookup(tResourceRequest* request)
{ if (request->state == rrsDnsLookup)
  { tCachedHostInformation* hostinfo = request->lookup;
    if (hostinfo != NULL) /* "should" be true */
    { reqlist_remove(&(hostinfo->waiting), &(hostinfo->waiting_num), request);
      request->lookup = NULL;
    }
    request->state = rrsStopped;
  }
}
#else
#define stop_request_dns_lookup(request) do { } while (0)
#endif

static one_caller void handle_request_removal(tResourceRequest* request)
/* main.c is about to remove the request, so we must detach it from all
   internal lists and drop any associated resource. */
{ tResource* resource = request->resource;
  stop_request_dns_lookup(request);
  if (resource != NULL)
  { reqlist_remove(&(resource->reqlist), &(resource->reqlist_num), request);
    drop_resource(resource);
  }
}

static one_caller void stop_request(tResourceRequest* request)
{ tResource* resource = request->resource;
  stop_request_dns_lookup(request);
  if (resource != NULL) stop_resource(resource);
  push_to_main_req(request, 0);
}

void resource_callback(void* data, tMainCallbackKind kind)
/* main.c wants something */
{ tMainCallbackFds* d;
  tResourceRequest* request;
#if CONFIG_DEBUG
  sprint_safe(debugstrbuf, "resource_callback(%p, %d)\n", data, kind);
  debugmsg(debugstrbuf);
#endif
  switch (kind)
  { case mckCalledSelect:
      is_time_valid = falsE; /* select() in main.c probably slept */
      i18n_cleanup  break;
    case mckStart: initial_request_handler((tResourceRequest*) data); break;
    case mckFds:
      d = (tMainCallbackFds*) data;
      got_activity = falsE;
      resource_handle_fds(d->selectcount);
      if ( (got_activity) && (is_environed) ) { /* IMPLEMENTME! */ }
      break;
    case mckRemovingRequest:
      handle_request_removal((tResourceRequest*) data); break;
    case mckStartSaveAs:
      request = (tResourceRequest*) data;
      start_save_as(request, request->uri);
      break;
    case mckStopActivity: stop_request((tResourceRequest*) data); break;
#if CONFIG_USER_QUERY
    case mckUserResponse:
      { tUserQuery* query = (tUserQuery*) data;
        (query->callback)(query);
      }
      break;
#endif
#if OPTION_EXECEXT & 1
    case mckExecextShell: start_shell((tExecextShellData*) data); break;
#endif
  }
}


/* Initialization */

#ifndef IPPORT_HTTP
#define IPPORT_HTTP 80
#endif
#ifndef IPPORT_FTP
#define IPPORT_FTP 21
#endif
#ifndef IPPORT_FINGER
#define IPPORT_FINGER 79
#endif
#ifndef IPPORT_GOPHER
#define IPPORT_GOPHER 70
#endif
#ifndef IPPORT_NNTP
#define IPPORT_NNTP 119
#endif
#ifndef IPPORT_POP3
#define IPPORT_POP3 110
#endif
#ifndef IPPORT_HTTPS
#define IPPORT_HTTPS 443
#endif

static tPortnumber __init __calculate_portnumber(const char* name,
  tPortnumber defaultport)
{ const struct servent* ent = getservbyname(name, strTcp);
  if (ent != NULL) return((tPortnumber) ent->s_port);
  else return((tPortnumber) htons(defaultport));
}

#define calculate_portnumber(rp, _default) \
  __calculate_portnumber(rp_data[rp].scheme, _default)

static void __init setup_pipe(/*@out@*/ int* reader, /*@out@*/ int* writer)
{ int fd_pair[2];
  unsigned char count;
  if (pipe(fd_pair) != 0) fatal_error(errno, _(strResourceError[rePipe]));
  for (count = 0; count <= 1; count++)
  { int fd = fd_pair[count];
    if (!is_fd_selectable(fd)) fatal_tmofd(fd);
    make_fd_cloexec(fd);
  }
  *reader = fd_pair[0]; *writer = fd_pair[1];
}

static /*@observer@*/ const char strCncih[] =
  N_("can't create internal handler");

#if OPTION_THREADING == 1

static void __init my_clone(void* (*fn)(void* dummy))
{ pthread_t dummy_id;
  pthread_attr_t thread_attr;
  int err;
  (void) pthread_attr_init(&thread_attr);
  (void) pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
  err = pthread_create(&dummy_id, &thread_attr, fn, NULL);
  if (err != 0) fatal_error(errno, _(strCncih));
}

#elif OPTION_THREADING == 2

static int __init my_clone(int (*fn)(void* dummy))
{ void* childstack = __memory_allocate(102400);
  int err = clone(fn, childstack + 102400 - 128, CLONE_VM | CLONE_FS |
    CLONE_FILES | CLONE_SIGHAND, NULL);
  /* CHECKME: we don't want CLONE_SIGHAND!? */
  if (err == -1) fatal_error(errno, _(strCncih));
  return(err);
}

#endif /* #if OPTION_THREADING... */

one_caller void __init resource_initialize(void)
/* initializes the resource handling; note that this is executed in the main
   thread - it _creates_ the other threads at the end. */
{ const struct protoent* proto;
#if OPTION_THREADING == 2
  int err;
#endif

  /* Get protocol and port numbers */

  proto = getprotobyname(strTcp);
  if (proto != NULL) ipprotocolnumber_tcp = proto->p_proto;
  /* "else": use the standard value IPPROTO_TCP */

  setservent(1);
  portnumber_http = calculate_portnumber(rpHttp, IPPORT_HTTP);
  portnumber_ftp = calculate_portnumber(__rpFtp, IPPORT_FTP);
  portnumber_finger = calculate_portnumber(__rpFinger, IPPORT_FINGER);
  portnumber_cvs = __calculate_portnumber("cvspserver", 2401);
  portnumber_gopher = calculate_portnumber(rpGopher, IPPORT_GOPHER);
  portnumber_nntp = __calculate_portnumber(strNntp, IPPORT_NNTP);
  portnumber_pop3 = __calculate_portnumber(strPop3, IPPORT_POP3);
  portnumber_https = calculate_portnumber(__rpHttps, IPPORT_HTTPS);
  endservent();

  /* Initialize some data structures */

  my_memclr_arr(chi_head);
#if OPTION_BUILTIN_DNS
  my_memclr_var(dns_data);
#endif

  /* Initialize cookies */

#if OPTION_COOKIES
  cookie_initialize();
#endif

  /* Setup some file descriptors */

#if OPTION_THREADING
  setup_pipe(&fd_dns2resource_read, &fd_dns2resource_write);
  setup_pipe(&fd_resource2dns_read, &fd_resource2dns_write);
  my_FD_SET(fd_dns2resource_read, &fds_read); /* read from DNS thread */
#endif

  setup_pipe(&fd_any2main_read, &fd_any2main_write);
  my_FD_SET(fd_any2main_read, &fds_read);

#if CONFIG_TG == TG_XCURSES
  setup_pipe(&fd_xcurses2main_read, &fd_xcurses2main_write);
#endif
#if TGC_IS_CURSES || (CONFIG_TG == TG_CONSOLE)
  if (is_promptable) my_FD_SET(fd_keyboard_input, &fds_read);
#endif

  /* Setup DNS thread */

#if OPTION_THREADING == 1
  my_clone(dns_handler_thread);
#elif OPTION_THREADING == 2
  pid_main = getpid();
  err = my_clone(dns_handler_thread);
  if (err > 0) pid_dns = err;
#endif
}

void resource_quit(void)
{
#if OPTION_TLS
  tls_deinitialize();
#endif
#if OPTION_THREADING == 2
  if (pid_dns > 0) (void) kill(pid_dns, SIGKILL); /* let's be rude :-) */
#endif
}
